use std::str::FromStr;

/// The section in the ASS file.
#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub enum ASSFileSection {
    ScriptInfo,
    V4Styles,
    Events,
}

/// The events of the [ASSFileSection::Events] section.
#[derive(Debug, Clone, PartialEq, Eq)]
pub struct ASSEvent {
    /// - Marked=0 means the line is not shown as "marked" in SSA.
    /// - Marked=1 means the line is shown as "marked" in SSA.
    pub marked: bool,

    /// Subtitles having different layer number will be ignored during the collusion detection.
    /// Higher numbered layers will be drawn over the lower numbered.
    pub layer: i32,

    /// Start Time of the Event, in 0:00:00:00 format ie. Hrs:Mins:Secs:hundredths. This is the
    /// time elapsed during script playback at which the text will appear onscreen. Note that there
    /// is a single digit for the hours!
    pub start: i32,

    /// End Time of the Event, in 0:00:00:00 format ie. Hrs:Mins:Secs:hundredths. This is the time
    /// elapsed during script playback at which the text will disappear offscreen. Note that there
    /// is a single digit for the hours!
    pub end: i32,

    /// Style name. If it is "Default", then your own *Default style will be subtituted. However,
    /// the Default style used by the script author IS stored in the script even though SSA ignores
    /// it - so if you want to use it, the information is there - you could even change the Name in
    /// the Style definition line, so that it will appear in the list of "script" styles.
    pub style: String,

    /// Character name. This is the name of the character who speaks the dialogue. It is for
    /// information only, to make the script is easier to follow when editing/timing.
    pub name: String,

    /// Transition Effect. This is either empty, or contains information for one of the three
    /// transition effects implemented in SSA v4.x The effect names are case sensitive and must
    /// appear exactly as shown. The effect names do not have quote marks around them.
    /// - "Karaoke" means that the text will be successively highlighted one word at a time.
    ///   Karaoke as an effect type is obsolete.
    /// - "Scroll up;y1;y2;delay[;fadeawayheight]"means that the text/picture will scroll up the
    ///   screen. The parameters after the words "Scroll up" are separated by semicolons. The y1
    ///   and y2 values define a vertical region on the screen in which the text will scroll. The
    ///   values are in pixels, and it doesn't matter which value (top or bottom) comes first. If
    ///   the values are zeroes then the text will scroll up the full height of the screen. The
    ///   delay value can be a number from 1 to 100, and it slows down the speed of the scrolling -
    ///   zero means no delay and the scrolling will be as fast as possible.
    /// - "Banner;delay" means that text will be forced into a single line, regardless of length,
    ///   and scrolled from right to left accross the screen. The delay value can be a number from
    ///   1 to 100, and it slows down the speed of the scrolling - zero means no delay and the
    ///   scrolling will be as fast as possible.
    /// - "Scroll down;y1;y2;delay[;fadeawayheight]"
    /// - "Banner;delay[;lefttoright;fadeawaywidth]" lefttoright 0 or 1. This field is optional.
    ///   Default value is 0 to make it backwards compatible. When delay is greater than 0, moving
    ///   one pixel will take (1000/delay) second. (WARNING: Avery Lee’s "subtitler" plugin reads
    ///   the "Scroll up" effect parameters as delay;y1;y2) fadeawayheight and fadeawaywidth
    ///   parameters can be used to make the scrolling text at the sides transparent.
    pub effect: String,

    /// Subtitle Text. This is the actual text which will be displayed as a subtitle onscreen.
    /// Everything after the 9th comma is treated as the subtitle text, so it can include commas.
    /// The text can include \n codes which is a line break, and can include Style Override control
    /// codes, which appear between braces { }.
    pub text: String,
}

#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
pub enum ScriptInfoKey {
    /// This is a description of the script. If the original author(s) did not provide this
    /// information then <untitled> is automatically substituted
    Title,

    /// The original author(s) of the script. If the original author(s) did not provide this
    /// information then <unknown> is automatically substituted
    OriginalScript,

    /// The original translator of the dialogue. This entry does not appear if no information was
    /// entered by the author. (optional)
    OriginalTranslation,

    /// The original script editor(s), typically whoever took the raw translation and turned it
    /// into idiomatic english and reworded for readability. This entry does not appear if no
    /// information was entered by the author. (optional)
    OriginalEditing,

    /// Whoever timed the original script. This entry does not appear if no information was entered
    /// by the author. (optional)
    OriginalTiming,

    /// Description of where in the video the script should begin playback. This entry does not
    /// appear if no information was entered by the author. (optional)
    SynchPoint,

    /// Names of any other subtitling groups who edited the original script. This entry does not
    /// appear if subsequent editors did not enter the information. (optional)
    ScriptUpdatedBy,

    /// The details of any updates to the original script - made by other subtitling groups. This
    /// entry does not appear if subsequent editors did not enter any information
    UpdateDetails,

    /// This is the SSA script format version eg. "V4.00". It is used by SSA to give a warning if
    /// you are using a version of SSA older than the version that created the script.
    /// ***ASS version is "V4.00+"***
    ScriptType,

    /// This determines how subtitles are moved, when automatically preventing onscreen collisions.
    ///
    /// If the entry says "Normal" then SSA will attempt to position subtitles in the position
    /// specified by the "margins". However, subtitles can be shifted vertically to prevent
    /// onscreen collisions. With "normal" collision prevention, the subtitles will "stack up" one
    /// above the other - but they will always be positioned as close the vertical (bottom) margin
    /// as possible - filling in "gaps" in other subtitles if one large enough is available.
    ///
    /// If the entry says "Reverse" then subtitles will be shifted upwards to make room for
    /// subsequent overlapping subtitles. This means the subtitles can nearly always be read
    /// top-down - but it also means that the first subtitle can appear half way up the screen
    /// before the subsequent overlapping subtitles appear. It can use a lot of screen area.
    Collisions,

    /// This is the height of the screen used by the script's author(s) when playing the script.
    /// SSA v4 will automatically select the nearest enabled setting, if you are using Directdraw
    /// playback
    PlayResY,

    /// This is the width of the screen used by the script's author(s) when playing the script. SSA
    /// will automatically select the nearest enabled, setting if you are using Directdraw
    /// playback
    PlayResX,

    /// This is the colour depth used by the script's author(s) when playing the script. SSA will
    /// automatically select the nearest enabled setting if you are using Directdraw playback.
    PlayDepth,

    /// This is the Timer Speed for the script, as a percentage. eg. "100.0000" is exactly 100%. It
    /// has four digits following the decimal point.
    ///
    /// The timer speed is a time multiplier applied to SSA's clock to stretch or compress the
    /// duration of a script. A speed greater than 100% will reduce the overall duration, and means
    /// that subtitles will progressively appear sooner and sooner. A speed less than 100% will
    /// increase the overall duration of the script means subtitles will progressively appear later
    /// and later (like a positive ramp time).
    ///
    /// The stretching or compressing only occurs during script playback - this value does not
    /// change the actual timings for each event listed in the script.
    ///
    /// Check the SSA user guide if you want to know why "Timer Speed" is more powerful than "Ramp
    /// Time", even though they both achieve the same result.
    Timer,

    /// Defines the default wrapping style:
    /// - 0: smart wrapping, lines are evenly broken
    /// - 1: end-of-line word wrapping, only \N breaks
    /// - 2: no word wrapping, \n \N both breaks
    /// - 3: same as 0, but lower line gets wider.
    WrapStyle,

    /// Undocumented, generated by Aegisub.
    ScaledBorderAndShadow,

    /// Undocumented, generated by Aegisub.
    YCbCrMatrix,
}

impl FromStr for ScriptInfoKey {
    type Err = String;

    #[inline]
    fn from_str(s: &str) -> Result<Self, Self::Err> {
        use ScriptInfoKey::*;
        match s.trim() {
            "Title" => Ok(Title),
            "OriginalTranslation" | "Original Translation" => Ok(OriginalTranslation),
            "OriginalEditing" | "Original Editing" => Ok(OriginalEditing),
            "ScriptUpdatedBy" | "Script Updated By" => Ok(ScriptUpdatedBy),
            "OriginalTiming" | "Original Timing" => Ok(OriginalTiming),
            "OriginalScript" | "Original Script" => Ok(OriginalScript),
            "UpdateDetails" | "Update Details" => Ok(UpdateDetails),
            "YCbCrMatrix" | "YCbCr Matrix" => Ok(YCbCrMatrix),
            "SynchPoint" | "Synch Point" => Ok(SynchPoint),
            "ScriptType" | "Script Type" => Ok(ScriptType),
            "Collisions" => Ok(Collisions),
            "PlayResY" => Ok(PlayResY),
            "PlayResX" => Ok(PlayResX),
            "PlayDepth" => Ok(PlayDepth),
            "Timer" => Ok(Timer),
            "WrapStyle" => Ok(WrapStyle),
            "ScaledBorderAndShadow" => Ok(ScaledBorderAndShadow),
            _ => Err(format!("unknown Script Info key: {s}")),
        }
    }
}

impl FromStr for ASSFileSection {
    type Err = String;

    #[inline]
    fn from_str(s: &str) -> Result<Self, Self::Err> {
        use ASSFileSection::*;
        const TRIM_PAT: &[char] = &['[', ']', ' ', '\t'];
        match s.trim_matches(TRIM_PAT).to_lowercase().as_str() {
            "script info" => Ok(ScriptInfo),
            "events" => Ok(Events),
            "v4+ styles" | "v4 styles" | "v4 styles+" => Ok(V4Styles),
            s => Err(format!("unknown section [{s}]")),
        }
    }
}
