use crate::args::Args;
use anyhow::Context as _;
use clap::CommandFactory as _;
use std::path::PathBuf;
use vvs_lang::{FrontendPipeline, SearchPath, SourceCode};
use vvs_runtime::Runtime;
use vvs_xdg::file::TempFile;

pub struct Settings {
    pub script: Option<PathBuf>,
    pub ass_file: Option<PathBuf>,
    pub options: Option<PathBuf>,
    pub includes: Vec<PathBuf>,
    pub terminal: ratatui::DefaultTerminal,
}

pub fn run(settings: Settings) -> anyhow::Result<()> {
    let script = match settings.script {
        Some(script) => SourceCode::from_path(&script)?,
        None => return Ok(Args::command().print_help()?),
    };
    let _ = script.module().context("expected a valid module name")?;

    let Settings { ass_file, options, includes, .. } = settings;
    let output = FrontendPipeline::new(&script)
        .passes(&[])
        .search(
            &SearchPath::from_iter(includes)
                .with_script_folder(script.folder().context("expected the script to have a parent folder")?),
        )
        .options(options.as_ref())?
        .run()?;

    let mut file = TempFile::new("vvcc")?;
    Runtime::new(move |ctx| {
        Ok(vvs_codegen::lowerer::Lowerer::new(&ctx, &output.main.name)?
            .declare_globals(output.declared_functions())?
            .lower_module(&output.main.name, output.ast)?
            .lower_modules(output.imports)?
            .generate_main(output.main)?
            .finished())
    })?
    .with_output_file(&mut file)
    .with_terminal(settings.terminal)
    .run_on_file(ass_file.context("no subtitle file to run the script on")?)?;
    file.leak();
    Ok(())
}
