use crate::args::Args;
use clap_complete::Shell;
use serde::{Deserialize, Serialize};
use std::path::PathBuf;
use vvs_utils::*;
use vvs_xdg::XDGConfigFileSerde;

#[derive(Debug, Serialize, Deserialize)]
pub struct ConfigKaraMaker {
    #[serde(rename = "name")]
    kara_maker: String,

    email: Option<String>,
}

#[derive(Debug, Default, Serialize, Deserialize)]
pub struct ConfigFile {
    includes: Vec<PathBuf>,

    karamaker: ConfigKaraMaker,
}

#[derive(Debug, Default)]
pub struct Config {
    // From config::ConfigFile
    pub kara_maker: String,

    // From args::Args
    pub script: Option<PathBuf>,
    pub ass_file: Option<PathBuf>,
    pub options: Option<PathBuf>,
    pub manpage: bool,
    pub shell: Option<Shell>,
    pub font_info: Option<Option<PathBuf>>,
    pub verbose: u8,

    // Merge of the two
    pub includes: Vec<PathBuf>,
}

#[derive(Debug, thiserror::Error)]
pub enum ConfigFileError {
    #[error("deserialization error: {0}")]
    Deserialize(toml::de::Error),

    #[error("serialization error: {0}")]
    Serialize(toml::ser::Error),
}

impl XDGConfigFileSerde for ConfigFile {
    type Error = ConfigFileError;

    fn deserialize(input: &str) -> Result<Self, Self::Error> {
        toml::from_str(input).map_err(ConfigFileError::Deserialize)
    }

    fn serialize(&self) -> Result<String, Self::Error> {
        toml::to_string_pretty(self).map_err(ConfigFileError::Serialize)
    }
}

macro_rules! merge {
    ($($what: ident $src: expr => $dest: expr, { $( $field: ident ),+ });+ $(;)?) => {{
        $(merge! { @@internal $what $( $field ),+ : $src => $dest });+;
    }};

    (@@internal override $( $field: ident ),+ : $src: expr => $dest: expr) => {{$( $dest.$field  = $src.$field; )+}};
    (@@internal flg_or   $( $field: ident ),+ : $src: expr => $dest: expr) => {{$( $dest.$field |= $src.$field; )+}};
    (@@internal flg_and  $( $field: ident ),+ : $src: expr => $dest: expr) => {{$( $dest.$field &= $src.$field; )+}};

    (@@internal append   $( $field: ident ),+ : $src: expr => $dest: expr) => {{$( $dest.$field.extend($src.$field); )+}};

    (@@internal max      $( $field: ident ),+ : $src: expr => $dest: expr) => {{$( $dest.$field = max_partial!($src.$field, $dest.$field); )+}};
    (@@internal min      $( $field: ident ),+ : $src: expr => $dest: expr) => {{$( $dest.$field = min_partial!($src.$field, $dest.$field); )+}};

    (@@internal set_if_not $( $field: ident),+ : $src: expr => $dest: expr) => {{$(
        if $dest.$field.is_none() {
            $dest.$field = $src.$field;
        }
    )+}};
}

impl Extend<ConfigFile> for ConfigFile {
    fn extend<T: IntoIterator<Item = ConfigFile>>(&mut self, iter: T) {
        iter.into_iter().for_each(|other| {
            let (km, self_km) = (other.karamaker, &mut self.karamaker);
            merge! {
                append   other => self,    { includes };
                override km    => self_km, { kara_maker };
            }
        });
    }
}

impl Config {
    /// Extends the config from config files.
    pub fn with_config_files(self, iter: impl IntoIterator<Item = ConfigFile>) -> Self {
        iter.into_iter().fold(self, |this, cfg| this.with_config_file(cfg))
    }

    /// Extends the config from config file.
    pub fn with_config_file(mut self, cfg: ConfigFile) -> Self {
        merge! {
            append   cfg           => self, { includes };
            override cfg.karamaker => self, { kara_maker };
        }
        self
    }

    /// Extends the config from command line arguments.
    pub fn with_args(mut self, args: Args) -> Self {
        merge! {
            append     args => self, { includes };
            set_if_not args => self, { script, ass_file, options, shell, font_info };
            max        args => self, { verbose };
            override   args => self, { manpage };
        }
        self
    }

    /// Apply the logger setting and returns self.
    pub fn apply_logger(self) -> Self {
        crate::logger::level(self.verbose);
        self
    }
}

impl Default for ConfigKaraMaker {
    #[inline]
    fn default() -> Self {
        Self { kara_maker: "Viieux".to_string(), email: Default::default() }
    }
}
