#![forbid(unsafe_code)]

//! The VivyScript cli

use anyhow::{Context, Result};
use clap::{CommandFactory as _, Parser as _};
use std::{fs, io, path::Path};
use vvs_cli::{
    args::Args,
    compiler,
    config::{Config, ConfigFile},
};
use vvs_xdg::{XDGConfig, XDGConfigMergedSilent};

fn print_font(n: impl AsRef<Path>, f: &[u8]) -> Result<()> {
    let f = vvs_font::Font::try_from(f).with_context(|| format!("failed to parse font {}", n.as_ref().display()))?;
    let ts = Vec::from_iter(f.font_types().into_iter().map(Into::<String>::into));
    println!("###");
    println!("# Family Name:        {}", f.name().context("failed to get the font name")?);
    println!("# Name(s):            {}", f.family_names().join(", "));
    println!("# Font Type(s):       {}", ts.join(", "));
    println!("# Number of glyph(s): {}", f.number_of_glyphs());
    Ok(())
}

fn main() -> Result<()> {
    vvs_cli::logger::init(None).context("failed to init logger")?;

    let Config { script, ass_file, options, manpage, shell, font_info, includes, .. } = Config::default()
        .with_config_file(
            XDGConfig::<ConfigFile, XDGConfigMergedSilent>::new("vvcc")
                .file("vvcc.toml")
                .read_or_default(),
        )
        .with_args(Args::parse())
        .apply_logger();

    if manpage {
        clap_mangen::Man::new(Args::command())
            .render(&mut io::stdout())
            .context("failed to render manpage for vvcc")
    } else if let Some(shell) = shell {
        clap_complete::generate(shell, &mut Args::command(), Args::command().get_name().to_string(), &mut io::stdout());
        Ok(())
    } else if let Some(font_info) = font_info {
        match font_info {
            Some(p) => print_font(&p, &fs::read(&p).with_context(|| format!("failed to read font: {}", p.display()))?),
            None => vvs_font::embeded_fonts().try_for_each(|(n, f)| print_font(n, f)),
        }
    } else {
        compiler::run(compiler::Settings { script, ass_file, options, includes, terminal: ratatui::init() })?;
        ratatui::restore();
        Ok(())
    }
}
