#![allow(dead_code)]

//! Small utility to store a basic block graph of a function to compute dominance of values.

use hashbrown::HashSet;
use std::{collections::VecDeque, hash::Hash, iter};

#[derive(Debug)]
pub(crate) struct Graph<N: Copy + Clone + PartialEq + Eq + Hash> {
    /// The nodes of the graph.
    content: Vec<N>,

    /// Encodes the edges of the graph. In general we have at most 2 outgoing edges per node (one
    /// for test false, one for test true). We should not encounter a switch, or at least not that
    /// many...
    adjacence: Vec<(usize, usize)>,
}

impl<N: Copy + Clone + PartialEq + Eq + Hash> Default for Graph<N> {
    fn default() -> Self {
        Self { content: Vec::default(), adjacence: Vec::default() }
    }
}

impl<N: Copy + Clone + PartialEq + Eq + Hash> Graph<N> {
    /// Create a new empty graph but with a given capacity.
    pub fn with_capacity(capacity: usize) -> Self {
        Self { content: Vec::with_capacity(capacity), adjacence: Vec::with_capacity(capacity.saturating_mul(2)) }
    }

    /// Insert a new node in the graph. If the node is already present we do nothing.
    pub fn add_node(mut self, node: N) -> Self {
        if !self.content.contains(&node) {
            self.content.push(node);
        }
        self
    }

    /// Insert a list of nodes in the graph.
    pub fn add_nodes(self, nodes: impl IntoIterator<Item = N>) -> Self {
        nodes.into_iter().fold(self, |this, node| this.add_node(node))
    }

    /// Insert an edge in the graph. If the edge already exists we do nothing.
    pub fn add_edge(mut self, from: N, to: N) -> Self {
        let (from, to) = (self.node_index(from), self.node_index(to));
        if !self.adjacence.iter().any(|(ef, et)| *ef == from && *et == to) {
            self.adjacence.push((from, to));
        }
        self
    }

    /// Find the index of a node in the graph or panic!
    fn node_index(&self, node: N) -> usize {
        self.content
            .iter()
            .enumerate()
            .find_map(|(idx, n)| (*n == node).then_some(idx))
            .expect("failed to find node")
    }

    /// Get all the direct next nodes from a given node.
    fn nexts(&self, from: N) -> Vec<N> {
        let node = self.node_index(from);
        self.adjacence
            .iter()
            .filter(|&&(from, _)| (from == node))
            .map(|&(_, to)| self.content[to])
            .collect()
    }

    /// Returns all the nodes in all the simple paths between two nodes. The nodes can be the same,
    /// in this case we return only this node. When computing the the paths we don't take loop edges.
    pub fn any_path_content(&self, from: N, to: N) -> HashSet<N> {
        // Don't ever reallocate.
        let mut visited = VecDeque::with_capacity(self.content.len());
        let mut stack = Vec::with_capacity(self.content.len());
        let mut ret = HashSet::with_capacity(self.content.len());

        // Init the state.
        visited.push_front(from);
        stack.push(self.nexts(from));

        // Just do the search.
        while let Some(children) = stack.last_mut() {
            match children.pop() {
                Some(child) if child == to => {
                    ret.extend(visited.iter().copied().chain(iter::once(to)));
                }
                Some(child) if !visited.contains(&child) => {
                    visited.push_front(child);
                    stack.push(self.nexts(child));
                }
                Some(_visited_child) => {}
                None => {
                    visited.pop_back();
                    stack.pop();
                }
            }
        }

        // We only have to returns the traversed nodes from the paths that found to.
        ret
    }
}
