use crate::{error::*, *};
use ab_glyph::Font as _;
use derive_more::Display;

/// The different types of fonts. To be sure of what you are doing, please use only regular
/// fonts, behaviour of non regular fonts may change depending on the platform if the italic or
/// bold tag is set in the ASS Style.
#[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord, Display)]
#[rustfmt::skip]
pub enum FontType {
    #[display("regular")]    Regular,
    #[display("italic")]     Italic,
    #[display("bold")]       Bold,
    #[display("oblique")]    Oblique,
    #[display("monospaced")] Monospaced,
    #[display("variable")]   Variable,
}

impl From<FontType> for String {
    fn from(value: FontType) -> Self {
        value.to_string()
    }
}

/// Struct used to store informations we need about a font.
#[derive(Debug)]
pub struct Font<'a> {
    font: ab_glyph::FontRef<'a>,
    face: ttf_parser::Face<'a>,
}

impl<'a> TryFrom<&'a [u8]> for Font<'a> {
    type Error = FontCreationError;

    #[inline]
    fn try_from(data: &'a [u8]) -> Result<Self, Self::Error> {
        Ok(Self {
            face: ttf_parser::Face::parse(data, 0).map_err(FontCreationError::TTFParserError)?,
            font: ab_glyph::FontRef::try_from_slice(data).map_err(FontCreationError::ABGlyphError)?,
        })
    }
}

impl Font<'_> {
    /// Get the general name of the font.
    pub fn name(&self) -> Result<String, FontError> {
        (self.face.names().into_iter())
            .find(|name| name.name_id == ttf_parser::name_id::FAMILY && name.is_unicode())
            .and_then(|name| name.to_string())
            .ok_or(FontError::NoName)
    }

    /// Get the family names.
    pub fn family_names(&self) -> Vec<String> {
        let names = self.face.names();
        let filter = [ttf_parser::name_id::POST_SCRIPT_NAME, ttf_parser::name_id::FULL_NAME];
        (names.into_iter())
            .flat_map(|name| (filter.contains(&name.name_id) && name.is_unicode()).then(|| name.to_string()))
            .flatten()
            .collect()
    }

    /// The the font types.
    pub fn font_types(&self) -> Vec<FontType> {
        let face = &self.face;

        let mut ret = [
            face.is_regular().then_some(FontType::Regular),
            face.is_italic().then_some(FontType::Italic),
            face.is_bold().then_some(FontType::Bold),
            face.is_oblique().then_some(FontType::Oblique),
            face.is_monospaced().then_some(FontType::Monospaced),
            face.is_variable().then_some(FontType::Variable),
        ]
        .into_iter()
        .flatten()
        .chain(Some(match self.face.style() {
            ttf_parser::Style::Normal => FontType::Regular,
            ttf_parser::Style::Italic => FontType::Italic,
            ttf_parser::Style::Oblique => FontType::Oblique,
        }))
        .collect::<Vec<_>>();

        ret.sort();
        ret.dedup();
        ret
    }

    /// Get the number of glyphs in the font.
    #[inline]
    pub fn number_of_glyphs(&self) -> i64 {
        self.face.number_of_glyphs() as i64
    }

    /// Outline a glyph with the specified pt size.
    pub fn outline_glyph(&self, pt: f64, glyph: char) -> Result<Rect, FontError> {
        let scaled_glyph = self.font.glyph_id(glyph).with_scale(
            self.font
                .pt_to_px_scale(f64::clamp(pt, f32::MIN.into(), f32::MAX.into()) as f32)
                .ok_or(FontError::FailedToScale(pt, glyph))?,
        );
        let ab_glyph::Rect { min, max } = self
            .font
            .outline_glyph(scaled_glyph)
            .ok_or(FontError::FailedToOutline(pt, glyph))?
            .px_bounds();
        Ok(Rect::new(
            Point { x: min.x.trunc() as i64, y: min.y.trunc() as i64 },
            Point { x: max.x.trunc() as i64, y: max.y.trunc() as i64 },
        ))
    }

    /// Outline a string slice with the specified pt size.
    pub fn outline_str(&self, pt: f64, str: impl AsRef<str>) -> Result<Rect, FontError> {
        let mut rects = str
            .as_ref()
            .chars()
            .map(|glyph| self.outline_glyph(pt, glyph))
            .collect::<Result<Vec<_>, _>>()?
            .into_iter();
        let first = rects.next().ok_or(FontError::EmptyStringToOutline)?;
        Ok(rects.fold(first, Rect::merge))
    }
}
