//! Contains the main code for the compiler. The compiler parses and do checks on the code, do some
//! optimisations and then returns the [ast::Ast] if no errors where found.

use crate::{
    error_report::ErrorReport,
    library::Library,
    passes::*,
    search_path::{SearchPath, DEFAULT_SEARCH_PATH},
    symbol_table::SymbolTable,
    *,
};
use derive_more::Display;
use hashbrown::{HashMap, HashSet};
use std::{borrow::Cow, cell::RefCell, fmt, fs, mem, path::Path, rc::Rc};
use vvs_parser::prelude::{ast::*, *};
use vvs_shortstring::ShortString;

/// Process a program.
pub struct FrontendPipeline<'a> {
    /// Options from a possibly passed file to the compiler.
    options: Option<Rc<OptionTable>>,

    /// The content of the program / main module.
    program: Cow<'a, str>,

    /// List of imported modules.
    imports: HashMap<ShortString, Rc<(Ast, Cow<'static, str>)>>,

    /// The search path.
    search: &'a SearchPath,

    /// Is the current module being compiled the main module?
    is_main: bool,

    /// The include path, upper in the vector is the modules importing the ones down in the vector.
    process_path: Vec<ShortString>,

    /// The library of available types.
    library: Rc<RefCell<Library>>,

    /// List of user defined passes. Will be run in order after we are done with our own passes.
    passes: &'a [&'a dyn UserFrontendPass],
}

/// Record of all the options defined, unused, setted in the program.
pub struct OptionsRecord {
    /// List of all unused declared options.
    unused: Vec<(ShortString, ShortString)>,

    /// All setted options.
    setted: Vec<(ShortString, ShortString, Expression)>,

    /// All declared options.
    declared: Vec<(ShortString, ShortString, TypeInfo, Expression)>,
}

impl fmt::Display for OptionsRecord {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self.setted.iter().map(|(a, b, _)| a.len() + b.len()).max() {
            None | Some(0) => Ok(()),
            Some(len) => self.setted.iter().try_for_each(|(module, name, expression)| {
                writeln!(f, "{module}.{name: >len$} {expression}", len = len - module.len())
            }),
        }
    }
}

/// Result of the front-end. Don't show that directly to the user, will show [FrontendOutput].
struct CompilerResult {
    /// The [Ast] of the main program's module.
    ast: Ast,

    /// List of all imported modules' [Ast]
    imports: HashMap<ShortString, Rc<(Ast, Cow<'static, str>)>>,

    /// The main program.
    main: MainProgram,

    /// Record of all options in the program.
    options: OptionsRecord,

    /// All the defined types.
    library: Rc<RefCell<Library>>,
}

/// The result of the frontend.
pub struct FrontendOutput {
    /// The [Ast] of the main program's module.
    pub ast: Ast,

    /// List of all imported modules' [Ast]
    pub imports: HashMap<ShortString, Ast>,

    /// The main program.
    pub main: MainProgram,

    /// Record of all options in the program.
    pub options: OptionsRecord,

    /// All the defined types.
    pub library: Library,
}

impl FrontendOutput {
    /// Get all the declared functions from the output of the frontend.
    pub fn declared_functions(&self) -> impl Iterator<Item = (&str, ShortString, Vec<&TypeInfo>, &TypeInfo)> {
        fn handle_function<'a>(
            module: &'a str,
            name: ShortString,
            body: &'a FunctionBody,
        ) -> (&'a str, ShortString, Vec<&'a TypeInfo>, &'a TypeInfo) {
            let arguments = body
                .type_specifiers()
                .flat_map(|ty| Some(ty?.type_info()))
                .collect::<Vec<_>>();
            let returns = body
                .return_type()
                .map(TypeSpecifier::type_info)
                .unwrap_or(TypeInfo::default_ref());
            (module, name, arguments, returns)
        }

        fn handle_stmt<'a>(
            module: &'a str,
            stmt: &'a Stmt,
        ) -> Option<(&'a str, ShortString, Vec<&'a TypeInfo>, &'a TypeInfo)> {
            match stmt {
                Stmt::FunctionDeclaration(d) => Some(handle_function(module, d.name().to_string().into(), d.body())),
                Stmt::LocalFunction(d) => Some(handle_function(module, d.name().to_string().into(), d.body())),
                Stmt::JobDeclaration(d) => Some(handle_function(module, d.name().to_string().into(), d.body())),
                _ => None,
            }
        }

        let this = self.ast.nodes().stmts().map(|stmt| (&self.main.name, stmt));
        let others = self.imports.iter().map(|(m, a)| a.nodes().stmts().map(move |s| (m, s)));

        this.chain(others.flatten())
            .flat_map(|(module, stmt)| handle_stmt(module, stmt))
    }
}

impl From<CompilerResult> for FrontendOutput {
    fn from(value: CompilerResult) -> Self {
        let CompilerResult { ast, imports, main, options, library } = value;
        let library = RefCell::into_inner(Rc::into_inner(library).expect("unique owner"));
        let imports = HashMap::from_iter(
            imports
                .into_iter()
                .map(|(module, ptr)| (module, Rc::into_inner(ptr).expect("unique owner").0)),
        );
        Self { ast, main, options, imports, library }
    }
}

impl fmt::Display for FrontendOutput {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let Self { imports, main, options, .. } = self;
        writeln!(f, "imported modules:")?;
        imports.keys().try_for_each(|module| writeln!(f, "\t- {module}"))?;
        writeln!(f, "{options}\n{main}")
    }
}

/// Tells if the frontend failed or not.
#[derive(Debug, Display, Clone, Copy)]
#[display("frontend error")]
pub struct FrontendError;

impl std::error::Error for FrontendError {}

impl From<()> for FrontendError {
    fn from(_: ()) -> Self {
        Self {}
    }
}

#[derive(Default)]
struct ImportAllResult {
    declared: Vec<(ShortString, ShortString, TypeInfo, Expression)>,
    unused: Vec<(ShortString, ShortString)>,
}

impl FromIterator<ImportAllResult> for ImportAllResult {
    fn from_iter<T: IntoIterator<Item = ImportAllResult>>(iter: T) -> Self {
        let mut iter = iter.into_iter();
        let Some(ret) = iter.next() else {
            return Default::default();
        };
        iter.fold(ret, |mut ret, item| {
            let Self { mut declared, unused } = item;
            ret.declared.append(&mut declared);
            ret.unused.retain(|option| unused.contains(option));
            ret
        })
    }
}

impl<'a> FrontendPipeline<'a> {
    /// Create a new pipeline to parse, to check and do black magic.
    pub fn new(source @ SourceCode { code, .. }: &'a SourceCode) -> Self {
        log::debug!(target: "vvs_parser", "run script `{}`", source.module().unwrap());
        Self {
            program: Cow::Borrowed(code.as_ref()),
            options: None,
            imports: Default::default(),
            is_main: true,
            library: Rc::new(RefCell::new(Library::default())),
            process_path: vec![ShortString::new(source.module().unwrap())],
            search: &DEFAULT_SEARCH_PATH,
            passes: &[],
        }
    }

    /// Add passes to do for each module at the end of mendatory passes.
    pub fn passes(self, passes: &'a [&'a dyn UserFrontendPass]) -> Self {
        Self { passes, ..self }
    }

    /// Get the name of the module we are compiling here. We should always have a name, if we parse
    /// the stdin we still have a name for the module: stdin...
    pub fn name(&self) -> &ShortString {
        self.process_path.last().expect("module name")
    }

    /// Set the option file to use for this instance of the program.
    pub fn options(self, options: Option<impl AsRef<Path>>) -> Result<Self, FrontendError> {
        let Some(options) = options else { return Ok(self) };
        let options = options.as_ref();

        let name = options
            .file_name()
            .ok_or_else(|| log::error!("expected a file with an actual name"))?
            .to_str()
            .ok_or_else(|| log::error!("expected files with utf8 names"))?;

        let code = fs::read_to_string(options)
            .map_err(|err| log::error!("failed to read option file '{}': {err}", options.display()))?;

        Ok(ast::OptionTableResult::parse_fallible(&code)
            .into_result()
            .map(|options| Self { options: Some(Rc::new(options)), ..self })
            .map_err(|errors| ErrorReport::new().add_errors(name, &code, errors).report_and_ignore())?)
    }

    /// Set the [SearchPath] to use for this instance of the program.
    pub fn search(self, search: &'a SearchPath) -> Self {
        Self { search, ..self }
    }

    /// Import all modules from the given list and add them to the current pipeline.
    fn import_all(&mut self, import_list: Vec<TokenReference>) -> Result<ImportAllResult, FrontendError> {
        Iterator::collect(import_list.into_iter().map(|import| {
            let name = import.token_type().as_str();
            if self.imports.contains_key(name) {
                return Ok(ImportAllResult::default());
            } else if self.process_path.iter().any(|path| path.as_str() == name) {
                ErrorReport::error(
                    self.name(),
                    &self.program,
                    AstError::from_parts(import.clone(), "recursive indlude of module"),
                );
                return Err(FrontendError);
            }

            let program = self.search.resolve(name).ok_or_else(|| {
                ErrorReport::error(
                    self.name(),
                    &self.program,
                    AstError::from_parts(import.clone(), "failed to find the module to import"),
                )
            })?;

            let CompilerResult {
                ast,
                imports, // Will replace the thing we already took
                options: OptionsRecord { unused, declared, .. },
                ..
            } = FrontendPipeline {
                process_path: self.process_path.clone().with(ShortString::new(name)),
                imports: mem::take(&mut self.imports), // Will be replaced.
                program: match program {
                    Cow::Borrowed(program) => Cow::Borrowed(program),
                    Cow::Owned(ref program) => Cow::Borrowed(program.as_str()),
                },
                options: self.options.clone(),
                library: self.library.clone(),
                is_main: false,
                ..*self
            }
            .process()?;

            debug_assert!({
                let prev = HashSet::<&ShortString>::from_iter(self.imports.keys());
                let news = HashSet::<&ShortString>::from_iter(imports.keys());
                (prev.len() <= news.len()) && (prev.intersection(&news).count() == prev.len())
            });

            let _ = mem::replace(&mut self.imports, imports); // Replace with a more complete import list.
            self.imports.insert(ShortString::new(name), Rc::new((ast, program)));

            Ok(ImportAllResult { declared, unused })
        }))
    }

    /// Process the pipeline. Some returned values may be hidden from the user, or things
    /// transformed a but more for the main result, but here we keep them.
    fn process(mut self) -> Result<CompilerResult, FrontendError> {
        // Parse the source code.
        let name = self.name().clone();
        let ast = AstResult::parse_fallible(&self.program)
            .into_result()
            .map_err(|err| ErrorReport::errors(&name, &self.program, err))?;

        // Import stuff and check vivy specific statements…
        let StmtCheckerResult { setted, main, imported } =
            StmtChecker::new(&ast, (name.as_str(), self.program.as_ref()))
                .is_main(self.is_main)
                .process()?;
        let result = self.import_all(imported)?;
        let ImportAllResult { mut declared, unused } = result;

        // Type checker + replace options in the AST…
        let loc = (name.as_str(), self.program.as_ref());
        let symbols = SymbolTable::try_from(&ast)
            .map_err(|err| ErrorReport::error(&name, &self.program, err))?
            .lock();
        TypeCollector::new(&ast, loc, &mut self.library.borrow_mut()).process()?;
        let opts = TypeChecker::new(&ast, loc, &self.library.borrow(), symbols)
            .process()?
            .options;

        let mut option_table = self.options.map(Rc::unwrap_or_clone).unwrap_or_default();
        option_table
            .section_mut("self")
            .extend(opts.iter().cloned().map(|(name, _, default)| (name, default)));
        declared.extend(opts.into_iter().map(|(n, t, v)| (name.clone(), n, t, v)));

        let (ast, results) = OptsTransform::new(ast, loc, &option_table, &declared).process()?;
        let unused_options = results.unused.without(|opt| unused.contains(opt));

        // User passes + return the result
        Ok(CompilerResult {
            ast: self.passes.iter().try_fold(ast, |prog, pass| {
                log::info!("apply pass {} on module {name}", pass.name());
                pass.process(prog, &name, self.program.as_ref())
            })?,
            imports: self.imports,
            library: self.library,
            options: OptionsRecord { declared, setted, unused: unused_options },
            main: main.unwrap_or_default().with_name(name),
        })
    }

    /// Process the pipeline. Returns the [Ast] or an error. Note that if an error occured, they
    /// will be printed out to the terminal.
    pub fn run(self) -> Result<FrontendOutput, FrontendError> {
        self.process().map(Into::into)
    }
}
