//! Passes to verify and transform the Vivy Script code written by an user.

mod opts_transform;
mod stmt_checker;
mod type_checker;
mod type_collector;

/// Macro used to declare a pass, can be a transform pass or a visit pass (for the role).
/// ```ignore
/// declare_pass! {
///     transform::OptsTransform
///     aux_infos { options: toml::Table }
///     opt_infos { verbose: u8 }
///     results { replaced_count: usize, unused_options: Vec<String> }
/// };
/// impl VisitorMut for OptsTransform {}
/// ```
macro_rules! declare_pass {
    ($(#[$meta: meta])* $role: ident :: $name: ident : $lifetime: lifetime {
            $($field_attr: ident : $field_ty: ty),* $(,)?
        }
        $(aux    { $($info_attr: ident : $info_ty: ty),+$(,)? })?
        $(opt    { $($opts_attr: ident : $opts_ty: ty),+$(,)? })?
        $(result { $($res_attr:  ident : $res_ty:  ty),+$(,)? })?
    ) => {
        $crate::passes::declare_pass! { @struct-decl $(#[$meta])* $role :: $name : $lifetime
            { $(  $field_attr : $field_ty),*   }
            { $($($info_attr  : $info_ty ),+)? }
            { $($($opts_attr  : $opts_ty ),+)? }
            { $($($res_attr   : $res_ty  ),+)? }
        }

        impl <$lifetime> $name <$lifetime> {
            $crate::passes::declare_pass! { @constructor $role :: $name : $lifetime
                { $(  $field_attr : $field_ty),*   }
                { $($($info_attr  : $info_ty ),+)? }
                { $($($opts_attr  : $opts_ty ),+)? }
                { $($($res_attr   : $res_ty  ),+)? }
            }

            /// Add a new ast error to the error list.
            #[allow(dead_code)]
            fn error(&mut self, token: impl Into<vvs_parser::prelude::ast::TokenReference>, error: impl Into<std::borrow::Cow<'static, str>>) {
                self.report.add_errors(self.name, self.source, [vvs_parser::prelude::ast::AstError::from_parts(token.into(), error)]);
            }

            /// Add a new ast warning to the error list.
            #[allow(dead_code)]
            fn warning(&mut self, token: impl Into<vvs_parser::prelude::ast::TokenReference>, error: impl Into<std::borrow::Cow<'static, str>>) {
                self.report.add_warnings(self.name, self.source, [vvs_parser::prelude::ast::AstError::from_parts(token.into(), error)]);
            }

            paste::paste! {
                $($(
                    #[doc = "Set the optional information '" $opts_attr "' for the [" $name "] pass."]
                    #[allow(clippy::wrong_self_convention)]
                    pub fn $opts_attr(self, $opts_attr: $opts_ty) -> Self {
                        Self { $opts_attr: Some($opts_attr), ..self }
                    }
                )+)?

                #[doc = "Apply the [" $name "] pass and get the resulting [vvs_parser::ast::Ast] or the list of the errors."]
                #[allow(unused_imports, unused_variables)]
                pub fn process(mut self) -> $crate::passes::declare_pass!(@pass-returns $role :: $name) {
                    use vvs_parser::prelude::ast::{Visitor as _, VisitorMut as _};
                    let ast = self.program.take().unwrap();
                    let ast = self.visit_ast(ast);
                    (!self.report.report()).then_some(()).ok_or(())?;
                    $crate::passes::declare_pass!(@returns $role :: $name (ast, [< $name Result >] {
                        $($($res_attr: self.$res_attr),+)?
                    }))
                }
            }
        }
    };


    (@constructor $role: ident :: $name: ident : $lifetime: lifetime
        { $($field_attr: ident : $field_ty: ty),* }
        { $($info_attr:  ident : $info_ty:  ty),* }
        { $($opts_attr:  ident : $opts_ty:  ty),* }
        { $($res_attr:   ident : $res_ty:   ty),* }
    ) => {
        paste::paste! {
            #[doc = "Create a new [" $name "] pass."]
            pub fn new(
                program: $crate::passes::declare_pass!(@decl-program $role $lifetime),
                (name, source): (&$lifetime str, &$lifetime str)
                $(, $info_attr: $info_ty)*
            ) -> Self {
                Self {
                    program: Some(program),
                    report: Default::default(),
                    name, source, $($info_attr,)*
                    $($field_attr: Default::default(),)*
                    $($opts_attr:  None,)*
                    $($res_attr:   Default::default(),)*
                }
            }
        }
    };

    (@struct-decl $(#[$meta: meta])* $role: ident :: $name: ident : $lifetime: lifetime
        { $(  $field_attr: ident : $field_ty: ty),*$(,)?   }
        { $($($info_attr:  ident : $info_ty:  ty),+$(,)?)? }
        { $($($opts_attr:  ident : $opts_ty:  ty),+$(,)?)? }
        { $($($res_attr:   ident : $res_ty:   ty),+$(,)?)? }
    ) => {
        $(#[$meta])*
        pub struct $name <$lifetime> {
            program: Option<$crate::passes::declare_pass!(@decl-program $role $lifetime)>,
            name:    &$lifetime str,
            source:  &$lifetime str,
            report:  $crate::error_report::ErrorReport<$lifetime>,
            $(  $field_attr: $field_ty,)*
            $($($info_attr:  $info_ty),+,)?
            $($($opts_attr:  Option<$opts_ty>),+,)?
            $($($res_attr:   $res_ty),+,)?
        }

        paste::paste! {
            #[doc = "The resut structure for the  [" $name "] pass."]
            #[derive(Default)]
            pub struct [< $name Result >]  {
                $($(pub $res_attr: $res_ty),+)?
            }
        }

    };

    (@pass-returns visit     :: $name: ident) => { paste::paste! { Result<[< $name Result >], ()> } };
    (@pass-returns transform :: $name: ident) => { paste::paste! { Result<(vvs_parser::prelude::ast::Ast, [< $name Result >]), ()> } };

    (@returns visit     :: $name: ident ($ast: expr, $result: expr)) => { Ok(($result)) };
    (@returns transform :: $name: ident ($ast: expr, $result: expr)) => { Ok((($ast, $result))) };

    (@decl-program visit     $lifetime: lifetime) => { &$lifetime vvs_parser::prelude::ast::Ast };
    (@decl-program transform $lifetime: lifetime) => { vvs_parser::prelude::ast::Ast };
}
use declare_pass;

pub use self::{opts_transform::*, stmt_checker::*, type_checker::*, type_collector::*};

/// Trait used to provide a way for users to create their own passes.
///
/// We don't want to expose the [declare_pass] macro for simplicity sake.
pub trait UserFrontendPass {
    /// Get the name of the pass. For logging purposes.
    fn name(&self) -> &'static str;

    /// Get the description of the pass. For logging purposes.
    fn description(&self) -> &'static str;

    /// Do the work on a specific module. The order in which the modules are passed is not
    /// guarented. It is better to not store state with interior mutability.
    fn process(
        &self,
        program: vvs_parser::prelude::ast::Ast,
        name: &str,
        source: &str,
    ) -> Result<vvs_parser::prelude::ast::Ast, crate::FrontendError>;
}
