use crate::{library::*, passes::declare_pass, symbol_table::*};
use hashbrown::HashSet;
use std::{borrow::Cow, cell::RefCell, mem};
use vvs_parser::prelude::{ast::*, *};
use vvs_shortstring::ShortString;

declare_pass! {
    /// The type checker is here to verify type corectness of the program and infer types, etc. To
    /// check the types we will need to do custom things...
    visit::TypeChecker: 'a {
        global: RefCell<HashSet<ShortString>>,
    }

    aux {
        library: &'a Library,
        symbols: SymbolTable<'a, TypeInfo, SymbolTableLocked>,
    }

    result {
        options: Vec<(ShortString, TypeInfo, Expression)>,
    }
}

macro_rules! box_error {
    ($tok:expr, $fmt:literal                  $(,)?) => { Box::new(AstError::from_parts($tok.into(), format!($fmt            ))) };
    ($tok:expr, $fmt:literal, $($args:expr),+ $(,)?) => { Box::new(AstError::from_parts($tok.into(), format!($fmt, $($args),+))) };
}

impl Visitor for TypeChecker<'_> {
    fn visit_ast(&mut self, ast: &Ast) {
        let errors = Vec::from_iter(ast.nodes().stmts().flat_map(|stmt| {
            match stmt {
                Stmt::FunctionDeclaration(decl) => self.handle_callable(decl.name(), decl.body(), false).err(),
                Stmt::JobDeclaration(decl) => self.handle_callable(decl.name(), decl.body(), true).err(),
                Stmt::LocalFunction(decl) => self
                    .type_compute()
                    .callable(decl.name(), decl.name().token_type().as_str(), decl.body(), false)
                    .err(),

                Stmt::Assignment(assignments) => self.handle_assignements(assignments).err(),
                Stmt::OptionDecl(decl) => self.handle_option_decl(decl).err(),

                Stmt::FunctionCall(call) => self.handle_top_level_function_call(call).err(),

                Stmt::TypeDeclaration(_) | Stmt::Import(_) => None,
                stmt => unreachable!("unexpected statement {stmt}"),
            }
        }));
        self.report.add_errors(self.name, self.source, errors);
        if !ast.eof().is_kind(TokenKind::Eof) {
            self.error(ast.eof(), "expected eof token here");
        }
    }
}

impl TypeChecker<'_> {
    fn type_compute(&self) -> TypeCompute {
        TypeCompute::new(self.library, &self.global, self.symbols.sub_scope())
    }

    fn handle_callable(
        &mut self,
        name: &FunctionName,
        body: &FunctionBody,
        is_job: bool,
    ) -> Result<(), TypeComputeError> {
        match name.method_name() {
            Some(method) => return Err(box_error!(method, "custom method are not supported for now")),
            None => self
                .type_compute()
                .callable(name.tokens().next().unwrap(), &name.to_string(), body, is_job)?,
        };
        Ok(())
    }

    fn handle_assignements(&mut self, assignments: &Assignment) -> Result<(), TypeComputeError> {
        if let Some(err) = assignments.variables().iter().find_map(|var| match var {
            Var::Expression(var) => match (var.prefix(), var.suffixes().next()) {
                (Prefix::Name(_), Some(Suffix::Index(Index::Dot { .. }))) => None,
                _ => Some(box_error!(var.tokens().next().unwrap(), "expected to set like `module.option = expr`")),
            },
            _ => Some(box_error!(var.tokens().next().unwrap(), "expected to set like `module.option = expr`")),
        }) {
            return Err(err);
        }
        self.type_compute().assignments(assignments)?;
        Ok(())
    }

    fn handle_top_level_function_call(&mut self, call: &FunctionCall) -> Result<(), TypeComputeError> {
        let token = call.tokens().next().unwrap();
        let Prefix::Name(function) = call.prefix() else {
            return Err(box_error!(token, "expected a function name here"));
        };

        let args: Vec<Cow<_>> = match call.suffixes().next() {
            Some(Suffix::Call(Call::AnonymousCall(args))) if call.suffixes().count() == 1 => {
                Vec::from_iter(args.as_arguments().into_iter().map(|field| match field {
                    Cow::Borrowed(field) => Cow::Borrowed(field.into()),
                    Cow::Owned(field) => Cow::Owned(field.into()),
                }))
            }
            _ => return Err(box_error!(token, "expected a simple function call here")),
        };

        match function.token_type().as_str() {
            "assert" => match args.len() {
                1 => self.type_compute().expr_as_type(TypeInfo::number(), &args[0]),
                2 => {
                    self.type_compute().expr_as_type(TypeInfo::number(), &args[0])?;
                    self.type_compute().expr_as_type(TypeInfo::string(), &args[1])?;
                    Ok(())
                }
                n => Err(box_error!(token, "expected 1 or 2 arguments when calling 'assert', got {n}")),
            },
            function @ ("assert_eq" | "assert_neq") => match args.len() {
                2 => {
                    let (ty, _) = self.type_compute().expr(&args[0])?;
                    self.type_compute().expr_as_type(&ty, &args[1])
                }
                3 => {
                    let (ty, _) = self.type_compute().expr(&args[0])?;
                    self.type_compute().expr_as_type(&ty, &args[1])?;
                    self.type_compute().expr_as_type(TypeInfo::string(), &args[2])
                }
                n => Err(box_error!(token, "expected 2 or 3 arguments when calling '{function}', got {n}")),
            },
            function => Err(box_error!(token, "unexpected call to function '{function}', expected 'assert' functions")),
        }
    }

    fn handle_option_decl(&mut self, decl: &OptionDecl) -> Result<(), TypeComputeError> {
        if decl.names().len() != decl.expressions().len() {
            return Err(box_error!(
                decl.option_token(),
                "number of option names differs from the number of default values"
            ));
        }

        let mut names = decl.names().iter();
        if let Some(name) = names.find(|n| !self.global.borrow_mut().insert(n.token_type().as_str().into())) {
            return Err(box_error!(name, "symbol redefinition"));
        }

        let expressions = decl
            .expressions()
            .iter()
            .map(|expr| (expr, self.type_compute().expr(expr)));
        let names = decl
            .names()
            .iter()
            .map(|name| ShortString::new(name.token_type().as_str()));

        let options = names
            .zip(expressions)
            .map(|(name, (expr, res))| res.map(|(ty, _)| (name, expr.clone(), ty)))
            .collect::<Result<Vec<_>, _>>()?;

        self.symbols = options
            .into_iter()
            .fold(mem::take(&mut self.symbols).unlock(), |symbols, (name, expr, ty)| {
                let symbols = symbols
                    .into_set(format!("self.{name}"), ty.clone())
                    .expect("non redefinition of symbol should have been checked")
                    .into_set(format!("{}.{name}", self.name), ty.clone())
                    .expect("non redefinition of symbol should have been checked");
                self.options.push((name, ty, expr));
                symbols
            })
            .lock();

        Ok(())
    }
}

/// The thing that will actually compute types and returns errors in case of problems.
struct TypeCompute<'a> {
    library: &'a Library,
    symbols: SymbolTable<'a, TypeInfo, SymbolTableMutable>,
    globals: &'a RefCell<HashSet<ShortString>>,
    in_loop: bool,
    in_job: bool,
    returns: &'a TypeInfo,
}

type TypeComputeError = Box<AstError>;
type TypeComputeResult<'a> = Result<(TypeInfo, SymbolTable<'a, TypeInfo, SymbolTableMutable>), TypeComputeError>;

fn unsupported<T>(token: &TokenReference, whats: &str) -> Result<T, TypeComputeError> {
    Err(box_error!(token, "{whats} are not supported for now"))
}

impl<'a> TypeCompute<'a> {
    fn new(
        library: &'a Library,
        globals: &'a RefCell<HashSet<ShortString>>,
        symbols: SymbolTable<'a, TypeInfo, SymbolTableMutable>,
    ) -> Self {
        Self { library, symbols, globals, returns: TypeInfo::default_ref(), in_loop: false, in_job: false }
    }

    fn sub(&'a self) -> Self {
        Self { symbols: self.symbols.clone(), ..*self }
    }

    fn set(mut self, symbol: &TokenReference, ty: TypeInfo) -> Result<Self, TypeComputeError> {
        self.symbols.set(symbol.token_type().as_str(), ty).map_err(|var| {
            box_error!(symbol, "{var} was already defined at the root level of the program and is thus a constant")
        })?;
        Ok(self)
    }

    fn in_loop(self) -> Self {
        Self { in_loop: true, ..self }
    }

    fn assignments(self, assignments: &'a Assignment) -> TypeComputeResult<'a> {
        assignments
            .variables()
            .iter()
            .zip(assignments.expressions())
            .try_for_each(|(dest, expr)| self.sub().assignation(dest, expr).map(|_| ()))?;
        Ok((TypeInfo::default(), self.symbols))
    }

    fn check_method_call(
        self,
        record: &TypeRecord,
        method: &MethodCall,
        caller: &TokenReference,
        token: &TokenReference,
    ) -> Result<TypeInfo, TypeComputeError> {
        let method_name = method.name().token_type().as_str();
        let (arguments, returns) = match record
            .method(method_name)
            .ok_or_else(|| box_error!(token, "undefined method '{caller}:{method_name}'"))?
        {
            TypeRecordMethod([this, args @ ..], ret, true) if this == record.type_info() => (args, ret),
            TypeRecordMethod(args, ret, false) => (args, ret),
            TypeRecordMethod([this, ..], _, true) => {
                return Err(box_error!(token, "try to call '{this}:{method_name}' on expression of type '{caller}'"))
            }
            TypeRecordMethod(.., true) => unreachable!(),
        };
        self.check_func_args(arguments, returns, method.args(), token)
    }

    fn check_func_args(
        self,
        arguments: &[TypeInfo],
        returns: &TypeInfo,
        func_args: &FunctionArgs,
        token: &TokenReference,
    ) -> Result<TypeInfo, TypeComputeError> {
        log::error!("what to do in case of table?");

        match arguments {
            // Treat the passed thing as a table construct if needed.
            [TypeInfo::Basic(name)] if name.token_type().as_str() == "table" => todo!(),

            // Treat the thing as an argument list.
            _ => {
                let func_args = func_args.as_arguments();
                func_args
                    .iter()
                    .any(|f| matches!(&**f, Field::ExpressionKey { .. } | Field::NameKey { .. }))
                    .then_some(())
                    .ok_or_else(|| {
                        box_error!(token, "passing named argument to a function that didn't expect a table")
                    })?;

                let args = Vec::from_iter(func_args.into_iter().map(|field| match field {
                    Cow::Borrowed(field) => Cow::Borrowed(Into::<&_>::into(field)),
                    Cow::Owned(field) => Cow::Owned(field.into()),
                }));

                let (arguments_len, args_len) = (arguments.len(), args.len());
                if args_len != arguments_len {
                    Err(box_error!(token, "argument count mismatch, expected {arguments_len}, got {args_len}"))
                } else {
                    args.into_iter()
                        .zip(arguments)
                        .try_for_each(|(expr, ty)| self.sub().expr_as_type(ty, &expr))
                        .map(|_| returns.clone())
                }
            }
        }
    }

    fn expr(self, expression: &'a Expression) -> TypeComputeResult<'a> {
        match expression {
            Expression::Parentheses { expression, .. } => self.expr(expression),
            Expression::FunctionCall(call) => self.suffixes(call.prefix(), call.suffixes()),

            Expression::Number(_) => Ok((TypeInfo::number().clone(), self.symbols)),
            Expression::String(_) => Ok((TypeInfo::string().clone(), self.symbols)),
            Expression::TypeAssertion { .. } => Ok((TypeInfo::default(), self.symbols)),

            Expression::Var(Var::Expression(var)) => self.suffixes(var.prefix(), var.suffixes()),
            Expression::Var(Var::Name(token)) => {
                let ty = self
                    .symbols
                    .get(token.token_type().as_str())
                    .ok_or_else(|| box_error!(token, "undefined variable"))?;
                Ok((ty.clone(), self.symbols))
            }

            Expression::Symbol(token) => match token.token_type() {
                TokenType::Symbol { symbol: Symbol::Nil } => Ok((TypeInfo::default(), self.symbols)),
                TokenType::Symbol { symbol: Symbol::False | Symbol::True } => {
                    Ok((TypeInfo::number().clone(), self.symbols))
                }
                _ => unreachable!(),
            },

            Expression::UnaryOperator { unop, expression } => {
                use {TypeInfo::*, UnOp::*};
                let (ty, symbols) = self.expr(expression)?;
                let is_container =
                    matches!(ty, Array { .. } | Basic(_) | Optional { .. } | Table { .. } | Tuple { .. });
                match unop {
                    Minus(_) | Not(_) | Tilde(_) if ty == *TypeInfo::number() => Ok((ty, symbols)),
                    Hash(_) if is_container => Ok((TypeInfo::number().clone(), symbols)),
                    _ => Err(box_error!(unop.token(), "can't apply this operator to expression of type '{ty}'")),
                }
            }

            Expression::BinaryOperator { lhs, binop, rhs } => {
                use BinOp::*;
                let ty = self.sub().expr(lhs)?.0;
                self.sub().expr_as_type(&ty, rhs)?;
                match binop {
                    TwoEqual(_) | TildeEqual(_) => Ok((TypeInfo::number().clone(), self.symbols)),
                    TwoDots(_) if ty == *TypeInfo::string() => Ok((TypeInfo::number().clone(), self.symbols)),

                    GreaterThan(_) | GreaterThanEqual(_) | LessThan(_) | LessThanEqual(_) | DoubleGreaterThan(_)
                    | DoubleLesserThan(_) | Caret(_) | Pipe(_) | Ampersand(_) | Tilde(_) | Percent(_) | Star(_)
                    | Slash(_) | DoubleSlash(_) | Minus(_) | Plus(_) | And(_) | Or(_)
                        if ty == *TypeInfo::number() =>
                    {
                        Ok((ty, self.symbols))
                    }

                    _ => Err(box_error!(binop.token(), "can't apply this operator to expression of type '{ty}'")),
                }
            }

            Expression::IfExpression(_) => todo!(),
            Expression::TableConstructor(_) => todo!(),

            Expression::Function(_) => unsupported(expression.tokens().next().unwrap(), "lambdas"),
            _ => unreachable!(),
        }
    }

    fn suffixes(self, prefix: &'a Prefix, suffixes: impl Iterator<Item = &'a Suffix> + 'a) -> TypeComputeResult<'a> {
        let prefix = match prefix {
            Prefix::Expression(expression) => self.sub().expr(expression)?.0,
            Prefix::Name(token) => self
                .symbols
                .get(token.token_type().as_str())
                .cloned()
                .ok_or_else(|| box_error!(token, "try to reference an undefined variable or function"))?,
            _ => unreachable!(),
        };

        let ty = suffixes.into_iter().try_fold(prefix, |prefix, suffix| match prefix {
            TypeInfo::Callback { arguments, return_type, .. } => match suffix {
                Suffix::Call(Call::AnonymousCall(call)) => self.sub().check_func_args(
                    &Vec::from_iter(arguments.into_iter().map(TypeArgument::into_type_info)),
                    &return_type,
                    call,
                    suffix.tokens().next().unwrap(),
                ),
                suffix => unreachable!("{suffix:?}"),
            },

            TypeInfo::Module(module) => match suffix {
                Suffix::Index(Index::Dot { name, .. }) => self
                    .symbols
                    .get(format!("{}.{}", module.token_type(), name.token_type()))
                    .cloned()
                    .ok_or_else(|| {
                        let module = module.token_type();
                        box_error!(name, "can't option, function, etc, with this name in module '{module}'")
                    }),
                suffix => Err(box_error!(
                    suffix.tokens().next().unwrap(),
                    "invalid way of getting a function, option, etc, from a module"
                )),
            },

            TypeInfo::Array { type_info, .. } => match suffix {
                Suffix::Index(Index::Dot { name, .. }) => match name.token_type().as_str() {
                    "first" | "last" => Ok(*type_info),
                    field => Err(box_error!(name, "unknown field '{field}' for array")),
                },
                Suffix::Index(Index::Brackets { expression, .. }) => self
                    .sub()
                    .expr_as_type(TypeInfo::number(), expression)
                    .map(|_| *type_info),
                _ => unreachable!(),
            },

            TypeInfo::Basic(type_name) => {
                let suffix_token = suffix.tokens().next().unwrap();
                let record = self
                    .library
                    .get(type_name.token_type().as_str())
                    .ok_or_else(|| box_error!(suffix_token, "try to use undefined type '{type_name}'"))?;
                match suffix {
                    Suffix::Call(Call::MethodCall(method)) => {
                        self.sub().check_method_call(record, method, &type_name, suffix_token)
                    }
                    Suffix::Index(Index::Dot { name, .. }) => {
                        let TypeRecordField(_, field) = record
                            .field(name.token_type().as_str())
                            .ok_or_else(|| box_error!(name, "undefined field for type '{type_name}'"))?;
                        Ok(field.clone())
                    }
                    Suffix::Index(Index::Brackets { expression, .. }) => todo!("{expression}"),
                    suffix => unreachable!("{suffix}"),
                }
            }

            TypeInfo::Tuple { types, .. } => {
                let index = match suffix {
                    Suffix::Index(Index::Dot { name: token, .. })
                    | Suffix::Index(Index::Brackets { expression: Expression::Number(token), .. }) => token,
                    _ => unreachable!(),
                };
                let handle = |str: &str| {
                    str.parse::<usize>().map_err(|err| {
                        box_error!(index, "index into a tuple must be a known integer at compilation time: {err}",)
                    })
                };
                let ty = types.iter().nth(match index.token_type() {
                    TokenType::Identifier { identifier } => match identifier.as_str() {
                        "zero" | "first" => 0,
                        "one" | "second" => 1,
                        "two" => 2,
                        idx => return Err(box_error!(index, "unknown tuple index: '{idx}'")),
                    },
                    TokenType::Number { text } => handle(text.as_str())?,
                    idx @ TokenType::StringLiteral { .. } => handle(idx.as_str())?,
                    _ => todo!(),
                });
                ty.cloned()
                    .ok_or_else(|| box_error!(index, "out of range in tuple: ({types})"))
            }

            TypeInfo::Table { fields, .. } => {
                let field = match suffix {
                    Suffix::Index(Index::Brackets { expression: Expression::String(name), .. })
                    | Suffix::Index(Index::Dot { name, .. }) => name.token_type().as_str(),
                    _ => unreachable!(),
                };
                let field = fields.iter().find_map(|type_field| {
                    let TypeFieldKey::Name(key) = type_field.key() else {
                        unreachable!()
                    };
                    let key = key.token_type().as_str();
                    (field == key).then_some(type_field.value())
                });
                field.cloned().ok_or_else(|| {
                    box_error!(suffix.tokens().next().unwrap(), "failed to find field in {{ {fields} }}")
                })
            }

            TypeInfo::Optional { base, .. } => match suffix {
                Suffix::Call(Call::MethodCall(call)) => match call.args() {
                    FunctionArgs::Parentheses { arguments, .. } if arguments.is_empty() => {
                        match call.name().token_type().as_str() {
                            "is_nil" | "is_value" => Ok(TypeInfo::number().clone()),
                            "value" => Ok(*base),
                            m => Err(box_error!(suffix.tokens().next().unwrap(), "unknown method '{base}::{m}'")),
                        }
                    }
                    _ => Err(box_error!(
                        suffix.tokens().next().unwrap(),
                        "can't pass parguments to '{base}::{}'",
                        call.name(),
                    )),
                },
                _ => Err(box_error!(suffix.tokens().next().unwrap(), "incorrect usage of optional value")),
            },

            _ => unreachable!(),
        })?;

        Ok((ty, self.symbols))
    }

    fn expr_as_type(self, as_ty: &TypeInfo, expression: &'a Expression) -> Result<(), TypeComputeError> {
        let (ty, _) = self.sub().expr(expression)?;
        (ty == *as_ty).then_some(()).ok_or_else(|| {
            box_error!(expression.tokens().next().unwrap(), "the expression should be of type '{as_ty}', got: '{ty}'",)
        })
    }

    fn local_assignment(mut self, assignment: &'a LocalAssignment) -> TypeComputeResult<'a> {
        let variables = assignment.names().iter();
        let variables = variables
            .zip(assignment.type_specifiers().map(|ty| ty.map(TypeSpecifier::type_info)))
            .zip(assignment.expressions())
            .map(|((name, ty), val)| match ty {
                None => Ok((name.token_type().as_str().into(), self.sub().expr(val)?.0)),
                Some(ty) => {
                    self.sub().expr_as_type(ty, val)?;
                    Ok((name.token_type().as_str().into(), ty.clone()))
                }
            });
        self.symbols
            .extend(variables.collect::<Result<Vec<_>, TypeComputeError>>()?)
            .map_err(|variable| {
                box_error!(assignment.local_token(), "can't declare {variable} because it shadows a global variable",)
            })?;
        Ok((TypeInfo::default(), self.symbols))
    }

    fn yields(self, yields: &'a Yield) -> TypeComputeResult<'a> {
        if !self.in_job {
            return Err(box_error!(yields.token(), "yielding values is only possible from jobs"));
        }

        for expr in yields.yields().iter() {
            match (self.sub().expr(expr)?.0, self.returns) {
                (TypeInfo::Basic(expr), TypeInfo::Basic(returns))
                    if matches!(
                        (expr.token_type().as_str(), returns.token_type().as_str()),
                        ("line" | "syllabe" | "lines" | "syllabes", "line") | ("syllabe" | "syllabes", "syllabe")
                    ) => {}
                (expr, returns) => {
                    return Err(box_error!(
                        expr.tokens().next().unwrap(),
                        "can't yield an expression of type '{expr}' as '{returns}'"
                    ))
                }
            }
        }

        Ok((TypeInfo::default(), self.symbols))
    }

    fn compound_assignment(self, assignment: &'a CompoundAssignment) -> TypeComputeResult<'a> {
        let symbols = self.symbols.clone();
        let ty = self.assignation(assignment.lhs(), assignment.rhs())?;
        match assignment.compound_operator() {
            op if op.is_numerical_op() && ty == *TypeInfo::number() => Ok((TypeInfo::default(), symbols)),
            CompoundOp::TwoDotsEqual(_) if ty == *TypeInfo::string() => Ok((TypeInfo::default(), symbols)),
            operator => Err(box_error!(operator.token(), "can't apply this operator to a variable of type '{ty}'")),
        }
    }

    fn assignation(self, variable: &'a Var, expression: &'a Expression) -> Result<TypeInfo, TypeComputeError> {
        match variable {
            Var::Expression(var) => {
                let (ty, _) = self.sub().suffixes(var.prefix(), var.suffixes())?;
                self.sub().expr_as_type(&ty, expression)?;
                Ok(ty)
            }

            Var::Name(variable) => match variable.token_type() {
                TokenType::Identifier { identifier } if self.symbols.is_global(identifier) => {
                    Err(box_error!(variable, "try to assign a global variable"))
                }

                TokenType::Identifier { identifier } => match self.symbols.get(identifier).cloned() {
                    Some(ty) => self.expr_as_type(&ty, expression).map(|_| ty),
                    None => Err(box_error!(variable, "try to assign an undefined variable")),
                },

                _ => unreachable!(),
            },

            _ => unreachable!(),
        }
    }

    fn stmt(mut self, stmt: &'a Stmt) -> TypeComputeResult<'a> {
        match stmt {
            Stmt::CompoundAssignment(assignments) => self.compound_assignment(assignments),
            Stmt::LocalAssignment(assignment) => self.local_assignment(assignment),
            Stmt::Assignment(assignments) => self.assignments(assignments),
            Stmt::FunctionCall(call) => self.suffixes(call.prefix(), call.suffixes()),
            Stmt::Yield(yields) => self.yields(yields),
            Stmt::Do(block) => self.block(block.block()),

            Stmt::If(if_stmt) => {
                self.sub().expr_as_type(TypeInfo::number(), if_stmt.condition())?;
                self.sub().block(if_stmt.block())?;
                for elseif in if_stmt.else_if().into_iter().flatten() {
                    self.sub().expr_as_type(TypeInfo::number(), elseif.condition())?;
                    self.sub().block(elseif.block())?;
                }
                if let Some(block) = if_stmt.else_block() {
                    self.sub().block(block)?;
                }
                Ok((TypeInfo::default(), self.symbols))
            }

            Stmt::Repeat(repeat_stmt) => {
                let symbols = self.sub().block(repeat_stmt.block())?.1;
                TypeCompute { symbols, ..self }
                    .in_loop()
                    .expr_as_type(TypeInfo::number(), repeat_stmt.until())?;
                Ok((TypeInfo::default(), self.symbols))
            }

            Stmt::While(while_stmt) => {
                self.sub().expr_as_type(TypeInfo::number(), while_stmt.condition())?;
                self.sub().in_loop().block(while_stmt.block())?;
                Ok((TypeInfo::default(), self.symbols))
            }

            Stmt::NumericFor(for_stmt) => {
                self.sub().expr_as_type(TypeInfo::number(), for_stmt.start())?;
                self.sub().expr_as_type(TypeInfo::number(), for_stmt.end())?;
                if let Some(step) = for_stmt.step() {
                    self.sub().expr_as_type(TypeInfo::number(), step)?;
                }
                self.sub()
                    .set(for_stmt.index_variable(), TypeInfo::number().clone())?
                    .block(for_stmt.block())?;
                Ok((TypeInfo::default(), self.symbols))
            }

            Stmt::GenericFor(for_stmt) if for_stmt.names().len() == for_stmt.expressions().len() => {
                let symbols = self.symbols.clone();
                self.symbols.extend(for_stmt
                    .type_specifiers()
                    .zip(for_stmt.names().iter().map(|name| ShortString::new(name.token_type().as_str())))
                    .zip(for_stmt.expressions())
                    .map(|((ty, name), expr)| match ty {
                        None => self.sub().expr(expr).map(|(ty, _)| (name, ty)),
                        Some(ty) => self.sub().expr_as_type(ty.type_info(), expr).map(|_| (name, ty.type_info().clone())),
                    })
                    .collect::<Result<Vec<_>, _>>()?)
                    .map_err(|name| box_error!(for_stmt.for_token(), "can't declare a variable with the name '{name}', is it a global?"))?;

                self.sub().in_loop().block(for_stmt.block())?;
                Ok((TypeInfo::default(), symbols))
            }
            Stmt::GenericFor(for_stmt) if for_stmt.expressions().len() == 1 => {
                let (it_type, _) = self.sub().expr(for_stmt.expressions().first_value().expect("oupsy"))?;
                let symbols = self.symbols.clone();
                let names = for_stmt.names();
                let handle = |name| box_error!(for_stmt.for_token(), "can't declare a variable with the name '{name}', is it a global?");

                match for_stmt.names().len() {
                    1 => self.symbols
                        .set(names.first_value().unwrap().token_type().as_str(), it_type)
                        .map_err(handle)?,
                    2 => self.symbols
                        .set(names.first_value().unwrap().token_type().as_str(), TypeInfo::number().clone())
                        .map_err(handle)?
                        .set(names.iter().nth(1).unwrap().token_type().as_str(), it_type)
                        .map_err(handle)?,
                    n => return Err(box_error!(for_stmt.for_token(), "expected one or two variables for this generic for, got {n} of them")),
                };

                self.sub().in_loop().block(for_stmt.block())?;
                Ok((TypeInfo::default(), symbols))
            }

            Stmt::JobDeclaration(decl) => Err(box_error!(decl.function_token(), "nested declaration of job is forbidden")),
            Stmt::GenericFor(for_stmt) => Err(box_error!(for_stmt.for_token(),
                "a generic for loop can either have the same amount of iterators as the expressions, or a single expression"
            )),

            Stmt::FunctionDeclaration(decl) => unsupported(decl.function_token(), "nested declaration of functions"),
            Stmt::LocalFunction(decl) => unsupported(decl.function_token(), "nested declaration of functions"),
            Stmt::TypeDeclaration(_) => unsupported(stmt.tokens().next().unwrap(), "custom types"),

            Stmt::Import(_) | Stmt::Write(_) | Stmt::Main(_) | Stmt::OptionDecl(_) => {
                unreachable!("should have been already checked / can't be inside function and job declarations")
            }
            _ => unreachable!(),
        }
    }

    fn last_stmt(self, stmt: &'a LastStmt) -> TypeComputeResult<'a> {
        match stmt {
            LastStmt::Break(token) | LastStmt::Continue(token) => match self.in_loop {
                true => Ok((TypeInfo::default(), self.symbols)),
                false => Err(box_error!(token, "invalid statement, not inside a loop")),
            },

            LastStmt::Return(returns) if returns.returns().len() >= 2 => {
                unsupported(returns.token(), "tuples to return multiple things")
            }
            LastStmt::Return(returns) if returns.returns().len() == 1 => (self.returns != TypeInfo::default_ref())
                .then(|| self.expr(returns.returns().iter().next().unwrap()))
                .ok_or_else(|| {
                    box_error!(returns.token(), "function or job returns something, but this return statement is empty")
                })?,
            LastStmt::Return(returns) if returns.returns().len() == 1 => (self.returns != TypeInfo::default_ref())
                .then(|| (TypeInfo::default(), self.symbols))
                .ok_or_else(|| {
                    box_error!(
                        returns.token(),
                        "expected to return an expression of type {}, but returns nothing here",
                        self.returns,
                    )
                }),
            _ => unreachable!(),
        }
    }

    fn block(self, block: &'a Block) -> TypeComputeResult<'a> {
        let symbols = block.stmts().try_fold(self.symbols.clone(), |symbols, stmt| {
            Ok::<_, TypeComputeError>(TypeCompute { symbols, ..self }.stmt(stmt)?.1)
        })?;
        Ok(match block.last_stmt() {
            Some(stmt) => (TypeCompute { symbols, ..self }.last_stmt(stmt)?.0, self.symbols),
            None => (TypeInfo::default(), self.symbols),
        })
    }

    fn callable(
        mut self,
        tok: &TokenReference,
        name: &str,
        body: &'a FunctionBody,
        is_job: bool,
    ) -> TypeComputeResult<'a> {
        if !self.globals.borrow_mut().insert(name.into()) {
            return Err(box_error!(tok, "redefinition of global variable, job or function"));
        }

        self.in_job = is_job;
        self.returns = body
            .return_type()
            .map(|specifier| specifier.type_info())
            .unwrap_or(TypeInfo::default_ref());

        if let Some((idx, parameter)) = body
            .parameters()
            .iter()
            .zip(body.type_specifiers().map(|ty| ty.map(|ty| ty.type_info())))
            .enumerate()
            .find_map(|(idx, (parameter, specifier))| match specifier {
                Some(ty) => self
                    .symbols
                    .set(parameter.token_type().as_str(), ty.clone())
                    .is_err()
                    .then_some((idx + 1, parameter)),
                None => Some((idx + 1, parameter)),
            })
        {
            return Err(box_error!(
                parameter.tokens().next().unwrap(),
                "parameter n°{idx} named '{parameter}' has no type specifier or is an ellipse '...' or shadows a global variable",
            ));
        }

        self.block(body.block())
    }
}
