//! Used to search for other files when using the `import` statement in Vivy Script.

use std::{
    borrow::Cow,
    fs,
    path::{Path, PathBuf},
};

/// Search path, can pass the name of a module and will try to return the associated code.
#[derive(Default)]
pub struct SearchPath {
    script_folder: Option<PathBuf>,
    folders: Vec<PathBuf>,
}

/// The extension for vivy script files.
pub const VIVY_SCRIPT_EXTENSION: &str = "vvs";

impl SearchPath {
    /// Create a new search path.
    pub const fn new() -> Self {
        Self { script_folder: None, folders: Vec::new() }
    }

    /// Resolve the search path to get the desired code.
    pub fn resolve(&self, import: &str) -> Option<Cow<'static, str>> {
        self.script_folder.iter().chain(&self.folders).find_map(|folder| {
            let mut path = folder.join(import);
            path.set_extension(VIVY_SCRIPT_EXTENSION);
            Some(Cow::Owned(fs::read_to_string(path).ok()?))
        })
    }

    /// Set the folder in which the script resides.
    pub fn with_script_folder(self, folder: impl AsRef<Path>) -> Self {
        Self { script_folder: Some(folder.as_ref().to_path_buf()), ..self }
    }

    /// Add a folder to the search list.
    pub fn with_include(mut self, folder: PathBuf) -> Self {
        self.folders.push(folder);
        self
    }
}

impl FromIterator<PathBuf> for SearchPath {
    /// Create a search list from a list of folders.
    fn from_iter<T: IntoIterator<Item = PathBuf>>(iter: T) -> Self {
        Self { folders: iter.into_iter().collect(), script_folder: None }
    }
}

/// The default search path, by default we only search the pre-packaged files.
pub(crate) static DEFAULT_SEARCH_PATH: SearchPath = SearchPath::new();
