use hashbrown::HashMap;
use std::{
    marker,
    ops::{self, Deref},
};
use vvs_parser::prelude::{ast::*, *};
use vvs_shortstring::ShortString;

#[derive(Debug, Clone)]
pub struct SymbolTableLocked;

#[derive(Debug, Clone)]
pub struct SymbolTableMutable;

/// The symbol table stores the available symbols with theyr type or whatever.
#[derive(Clone)]
pub struct SymbolTable<'a, T: Clone, State> {
    /// A possible parent for this scope.
    parent: Option<&'a SymbolTable<'a, T, SymbolTableLocked>>,

    /// List of symbols from this scope.
    symbols: HashMap<ShortString, T>,

    /// The state of the table, can be locked or mutable.
    _state: marker::PhantomData<State>,
}

impl<T: Clone, State> Default for SymbolTable<'_, T, State> {
    fn default() -> Self {
        Self { parent: None, symbols: HashMap::new(), _state: marker::PhantomData }
    }
}

impl<'a, T: Clone> SymbolTable<'a, T, SymbolTableLocked> {
    /// Create a new sub-scoped symbol table.
    pub fn sub_scope(&'a self) -> SymbolTable<'a, T, SymbolTableMutable> {
        SymbolTable { parent: Some(self), symbols: Default::default(), _state: marker::PhantomData }
    }

    /// Get the root scope.
    pub fn root(&self) -> &SymbolTable<'a, T, SymbolTableLocked> {
        let mut root = self;
        while let Some(parent) = root.parent {
            root = parent
        }
        root
    }

    /// Unlock the table.
    pub fn unlock(self) -> SymbolTable<'a, T, SymbolTableMutable> {
        let Self { parent, symbols, .. } = self;
        SymbolTable::<'a, T, SymbolTableMutable> { parent, symbols, _state: marker::PhantomData }
    }
}

impl<'a, T: Clone> SymbolTable<'a, T, SymbolTableMutable> {
    /// Lock the table.
    pub fn lock(self) -> SymbolTable<'a, T, SymbolTableLocked> {
        let Self { parent, symbols, .. } = self;
        SymbolTable::<'a, T, SymbolTableLocked> { parent, symbols, _state: marker::PhantomData }
    }

    /// Add a new variable to the current scope. We override symbols (non-UB shadowing of
    /// variables). Note that we can't shadow or set variables from the root scope as it denotes
    /// the global state. In case of error we return the name of the problematic variable.
    pub fn set(&mut self, symbol: impl Into<ShortString> + AsRef<str>, value: T) -> Result<&mut Self, ShortString> {
        match self.parent {
            Some(parent) if parent.root().get(symbol.as_ref()).is_none() => {
                self.symbols.insert(symbol.into(), value);
                Ok(self)
            }
            None => {
                self.symbols.insert(symbol.into(), value);
                Ok(self)
            }
            Some(_) => Err(symbol.into()),
        }
    }

    /// Like [Self::set], but also consume self.
    pub fn into_set(mut self, symbol: impl Into<ShortString> + AsRef<str>, value: T) -> Result<Self, ShortString> {
        match self.parent {
            Some(parent) if parent.root().get(symbol.as_ref()).is_none() => {
                self.symbols.insert(symbol.into(), value);
                Ok(self)
            }
            None => {
                self.symbols.insert(symbol.into(), value);
                Ok(self)
            }
            Some(_) => Err(symbol.into()),
        }
    }

    /// Add variables from an iterator. Same rules as [Self::set].
    pub fn extend(&mut self, iter: impl IntoIterator<Item = (ShortString, T)>) -> Result<&mut Self, ShortString> {
        iter.into_iter()
            .try_fold(self, |this, (symbol, value)| this.set(symbol, value))
    }
}

impl<T: Clone, State> SymbolTable<'_, T, State> {
    /// Get a variable/callback/function from this scope or any parent scope.
    pub fn get(&self, symbol: impl AsRef<str>) -> Option<&T> {
        self.symbols
            .get(symbol.as_ref())
            .or_else(|| self.parent?.get(symbol.as_ref()))
    }

    /// Tells if a variable/callback/function is present in this scope or any parent scope.
    pub fn has(&self, symbol: impl AsRef<str>) -> bool {
        self.get(symbol).is_some()
    }

    /// Tells if a symbol is global or not.
    pub fn is_global(&self, symbol: impl AsRef<str>) -> bool {
        (self.parent.is_none() && self.get(symbol.as_ref()).is_some())
            || self
                .parent
                .map(SymbolTable::root)
                .map_or(false, |root| root.get(symbol).is_some())
    }
}

impl<T: Clone> FromIterator<(ShortString, T)> for SymbolTable<'_, T, SymbolTableMutable> {
    fn from_iter<I: IntoIterator<Item = (ShortString, T)>>(iter: I) -> Self {
        let mut table = Self::default();
        let _ = table.extend(iter);
        table
    }
}

impl<Idx: AsRef<str>, T: Clone + DefaultRef + 'static, State> ops::Index<Idx> for SymbolTable<'_, T, State> {
    type Output = T;

    /// Same as [SymbolTable::get] but we returns a reference to a default value if the variable was not found.
    fn index(&self, index: Idx) -> &Self::Output {
        self.get(index).unwrap_or_else(|| T::default_ref())
    }
}

fn fold_on_stmt<'a>(
    mut symbol: SymbolTable<'a, TypeInfo, SymbolTableMutable>,
    stmt: &'a Stmt,
) -> Result<SymbolTable<'a, TypeInfo, SymbolTableMutable>, Box<AstError>> {
    let handle_decl = |mut symbol: SymbolTable<'a, _, _>, name: &str, body: &FunctionBody| {
        let (arrow, return_type) = body
            .return_type()
            .map(|ty| (ty.punctuation().clone(), Box::new(ty.type_info().clone())))
            .unwrap_or_else(|| {
                let arrow = Token::new(TokenType::Symbol { symbol: Symbol::ThinArrow });
                (TokenReference::new(vec![], arrow, vec![]), Box::new(TypeInfo::default()))
            });

        let arguments = Punctuated::from_iter(body.parameters().pairs().zip(body.type_specifiers()).map(
            |(param, ty)| {
                let punctuation = ty.map(|ty| ty.punctuation()).cloned().unwrap_or_else(|| {
                    let token = Token::new(TokenType::Symbol { symbol: Symbol::TwoDots });
                    TokenReference::new(vec![], token, vec![])
                });
                let name = Some((param.value().clone().deref().clone(), punctuation));
                let type_info = ty.map(TypeSpecifier::type_info).cloned().unwrap_or_default();
                let type_argument = TypeArgument::new(type_info).with_name(name);
                match param.punctuation() {
                    Some(punctuation) => Pair::Punctuated(type_argument, punctuation.clone()),
                    None => Pair::End(type_argument),
                }
            },
        ));

        let parentheses = body.parameters_parentheses().clone();
        let _ = symbol.set(name, TypeInfo::Callback { parentheses, arguments, arrow, return_type });
        Ok(symbol)
    };

    match stmt {
        // Handled by the type collector
        Stmt::FunctionDeclaration(d) if d.name().method_name().is_some() => Ok(symbol),

        // Function declarations errors
        Stmt::FunctionDeclaration(d) if d.name().names().len() != 1 => Err(Box::new(AstError::from_parts(
            d.name().names().first_value().unwrap().clone(),
            "invalid name, expected a single name without the dots",
        ))),
        Stmt::FunctionDeclaration(d) if matches!(d.name().names().first_value(), Some(n) if symbol.has(n.token_type().as_str())) => {
            Err(Box::new(AstError::from_parts(
                d.name().names().first_value().unwrap().clone(),
                "multiple definition of this symbol",
            )))
        }
        Stmt::LocalFunction(local) if symbol.has(local.name().token_type().as_str()) => {
            Err(Box::new(AstError::from_parts(local.name().clone(), "multiple definition of this symbol")))
        }

        // Function declarations
        Stmt::LocalFunction(local) => handle_decl(symbol, local.name().token_type().as_str(), local.body()),
        Stmt::FunctionDeclaration(decl) => {
            handle_decl(symbol, decl.name().names().first_value().unwrap().token_type().as_str(), decl.body())
        }

        // Module import
        Stmt::Import(import) => {
            symbol
                .set(import.name().token_type().as_str(), TypeInfo::Module(import.name().clone()))
                .map_err(|_| Box::new(AstError::from_parts(import.name().clone(), "import shadows other global")))?;
            Ok(symbol)
        }

        // Ignore
        _ => Ok(symbol),
    }
}

impl<'a> TryFrom<&'a Ast> for SymbolTable<'a, TypeInfo, SymbolTableMutable> {
    type Error = AstError;

    /// Create a symbol table from an AST. Note that only imported modules and functions/job
    /// declarations are treated here. We need the type checker to be ran over to verify
    /// correctness of the thingy and add options/constants to the table, because we need to
    /// compute their infered types when it's not explicit…
    fn try_from(ast: &'a Ast) -> Result<Self, Self::Error> {
        let mut sym = Self::default();
        let _ = sym.set(
            ShortString::new_inline("self"),
            TypeInfo::Module(TokenReference::new(
                vec![],
                Token::new(TokenType::Identifier { identifier: ShortString::new_inline("self") }),
                vec![],
            )),
        );
        ast.nodes().stmts().try_fold(sym, fold_on_stmt).map_err(|err| *err)
    }
}
