use crate::prelude::{Context, Module};
use anyhow::{bail, Context as _};
use llvm_sys::{
    error::*,
    orc2::{ee::*, lljit::*, *},
};
use std::{
    ffi::{c_int, c_void, CStr, CString},
    ptr,
};

/// Create a JIT engine.
#[allow(dead_code)]
pub struct JIT {
    jit: *mut LLVMOrcOpaqueLLJIT,
    es: *mut LLVMOrcOpaqueExecutionSession,

    tsctx: *mut LLVMOrcOpaqueThreadSafeContext,

    irtl: *mut LLVMOrcOpaqueIRTransformLayer,
    objl: *mut LLVMOrcOpaqueObjectLayer,
    dylib: *mut LLVMOrcOpaqueJITDylib,
}

macro_rules! llvm_expect {
    (unsafe $err: expr, $msg: literal $(, $arg: expr)* $(,)?) => {{
        unsafe { llvm_expect!($err, $msg $(, $arg)*) }
    }};

    ($err: expr, $msg: literal $(, $arg: expr)* $(,)?) => {{
        let err = $err;
        if !err.is_null() {
            let msg = CStr::from_ptr(LLVMGetErrorMessage(err))
                .to_string_lossy()
                .to_string();
            LLVMConsumeError(err);
            bail!("{}: {}", format_args!($msg, $(, $arg)*), msg)
        }
    }};
}

extern "C" fn orc_sym_filter(_: *mut c_void, sym: LLVMOrcSymbolStringPoolEntryRef) -> c_int {
    eprintln!("toto");
    let Ok(sym) = unsafe { CStr::from_ptr(LLVMOrcSymbolStringPoolEntryStr(sym)) }.to_str() else {
        log::error!("got a symbol which name was not valid utf8 from orc");
        return 0;
    };
    sym.starts_with("VVRT") as c_int
}

impl JIT {
    /// Create a new JIT thingy to execute Vivy Script code.
    fn new() -> anyhow::Result<JIT> {
        if let Err(err) = crate::init::initialize_llvm() {
            // Should fail only on re-init. If a real problem occured, we would have already
            // aborted the program.
            log::debug!("{err}");
        }

        unsafe {
            // Create the jit.
            let mut jit = ptr::null_mut();
            llvm_expect!(LLVMOrcCreateLLJIT(&mut jit, ptr::null_mut()), "failed to create the orc jit");
            log::debug!("got data layout: {}", CStr::from_ptr(LLVMOrcLLJITGetDataLayoutStr(jit)).to_string_lossy());

            // Create things, in this order.
            let es = LLVMOrcLLJITGetExecutionSession(jit);
            let objl = LLVMOrcCreateRTDyldObjectLinkingLayerWithSectionMemoryManager(es);
            let irtl = LLVMOrcLLJITGetIRTransformLayer(jit);
            let dylib = LLVMOrcExecutionSessionCreateBareJITDylib(es, c"<main>".as_ptr());

            // Allow to search for symbols in the current process, only if they start with `VVRT`.
            let mut dg = ptr::null_mut();
            llvm_expect!(
                LLVMOrcCreateDynamicLibrarySearchGeneratorForProcess(
                    &mut dg,
                    LLVMOrcLLJITGetGlobalPrefix(jit),
                    Some(orc_sym_filter),
                    ptr::null_mut()
                ),
                "failed to create an orc dylib search generator for the current process"
            );
            LLVMOrcJITDylibAddGenerator(dylib, dg);

            // Ok.
            Ok(JIT { jit, es, dylib, tsctx: LLVMOrcCreateNewThreadSafeContext(), irtl, objl })
        }
    }

    /// Initialize a jot with a module, that must be created from the passed context.
    pub fn init<'a>(cb: impl FnOnce(Context<'a>) -> anyhow::Result<Module<'a>>) -> anyhow::Result<Self> {
        let jit = JIT::new()?;
        let module = unsafe { cb(Context::from_ptr(LLVMOrcThreadSafeContextGetContext(jit.tsctx), true))?.into_ptr() };
        let tsm = unsafe { LLVMOrcCreateNewThreadSafeModule(module, jit.tsctx) };
        let rc = unsafe { LLVMOrcLLJITAddLLVMIRModule(jit.jit, jit.dylib, tsm) };
        unsafe { llvm_expect!(rc, "failed to add the module into the jit") };
        Ok(jit)
    }

    /// LookUp a symbol defined in the JIT.
    pub fn lookup(&self, symbol: &str) -> anyhow::Result<*const u8> {
        let mut addr = Default::default();
        llvm_expect!(unsafe
            LLVMOrcLLJITLookup(self.jit, &mut addr, CString::new(symbol).with_context(|| "CString::new failed")?.as_ptr()),
            "failed to lookup symbol `{symbol}`"
        );
        Ok(addr as *const u8)
    }
}

impl Drop for JIT {
    fn drop(&mut self) {
        unsafe {
            LLVMOrcDisposeThreadSafeContext(self.tsctx);
            LLVMOrcDisposeLLJIT(self.jit);
            LLVMOrcDisposeObjectLayer(self.objl);
        }
    }
}
