use crate::prelude::*;
use llvm_sys::{core::*, prelude::*, LLVMLinkage, LLVMVisibility};
use std::ffi::{CString, NulError};

crate::bindings::declare! { mut LLVMModuleRef as Module<'a> {
    is_borrowed: bool,
} }

impl<'a> Module<'a> {
    /// Create a new module.
    pub fn new(ctx: &Context<'a>, name: impl AsRef<str>) -> Result<Self, NulError> {
        unsafe {
            Ok(Module::from_ptr(
                LLVMModuleCreateWithNameInContext(CString::new(name.as_ref())?.as_ptr(), ctx.as_ptr()),
                false,
            ))
        }
    }

    /// Add a new function in this module.
    ///
    /// Note that by default all functions are private.
    pub fn add_function(&mut self, name: impl AsRef<str>, ty: FunctionType) -> Result<FunctionBuilder, NulError> {
        let function = unsafe { LLVMAddFunction(self.inner, CString::new(name.as_ref())?.as_ptr(), ty.as_ptr()) };

        // All the functions are private to the module!
        unsafe {
            LLVMSetLinkage(function, LLVMLinkage::LLVMLinkerPrivateLinkage);
            LLVMSetVisibility(function, LLVMVisibility::LLVMHiddenVisibility);
        }

        Ok(unsafe { FunctionBuilder::from_ptr(function) })
    }

    /// Get a function out of the a module as a declaration.
    pub fn get_function(&self, name: impl AsRef<str>) -> Result<FunctionDeclaration, LLVMError> {
        match unsafe { LLVMGetNamedFunction(self.as_ptr(), CString::new(name.as_ref())?.as_ptr()) } {
            ptr if ptr.is_null() => Err(LLVMError::Undefined(name.as_ref().into())),
            ptr => Ok(unsafe { FunctionDeclaration::from_ptr(ptr) }),
        }
    }

    /// Get the name of the module.
    pub fn name(&self) -> &str {
        let mut len = 0;
        std::str::from_utf8(unsafe {
            let c_ptr = LLVMGetModuleIdentifier(self.as_ptr(), &mut len);
            std::slice::from_raw_parts(c_ptr as *const _, len)
        })
        .expect("valid utf8 module name")
    }

    /// Get out the pointer out of the struct, the caller is now responsible of managing the
    /// lifetime of the returned pointer.
    #[must_use]
    pub(super) unsafe fn into_ptr(mut self) -> LLVMModuleRef {
        self.is_borrowed = true;
        self.inner
    }
}

impl Drop for Module<'_> {
    fn drop(&mut self) {
        if !self.is_borrowed {
            unsafe { LLVMDisposeModule(self.inner) }
        }
    }
}
