use llvm_sys::{core::*, prelude::*, LLVMTypeKind};

crate::bindings::declare! { const LLVMTypeRef as Type<'a> }
crate::bindings::declare! { const LLVMTypeRef as FunctionType<'a> }

impl FunctionType<'_> {
    /// Get the returned type of the function.
    pub fn returns(&self) -> Type {
        unsafe { Type::from_ptr(LLVMGetReturnType(self.as_ptr())) }
    }

    /// Get the number of arguments.
    pub fn args_len(&self) -> usize {
        (unsafe { LLVMCountParamTypes(self.as_ptr()) }) as usize
    }

    /// Get the arguments of the function type.
    pub fn args(&self) -> Vec<Type> {
        let params_count = self.args_len();
        let mut params = Vec::with_capacity(params_count);
        params.resize(params_count, std::ptr::null_mut());
        unsafe { LLVMGetParamTypes(self.as_ptr(), params.as_mut_ptr()) };
        params.into_iter().map(|ptr| unsafe { Type::from_ptr(ptr) }).collect()
    }
}

impl<'a> From<FunctionType<'a>> for Type<'a> {
    fn from(value: FunctionType<'a>) -> Self {
        unsafe { Type::from_ptr(value.inner) }
    }
}

impl<'a, const N: usize> From<([Type<'a>; N], Type<'a>)> for FunctionType<'a> {
    fn from((args, returns): ([Type<'a>; N], Type<'a>)) -> Self {
        let (returns, mut args) = (returns.as_ptr(), args.map(|ty| ty.as_ptr()));
        unsafe { Self::from_ptr(LLVMFunctionType(returns, args.as_mut_ptr(), N as u32, 0)) }
    }
}

impl<'a> From<(Vec<Type<'a>>, Type<'a>)> for FunctionType<'a> {
    fn from((mut args, returns): (Vec<Type<'a>>, Type<'a>)) -> Self {
        Self::from((args.as_mut_slice(), returns))
    }
}

impl<'a> From<(&mut [Type<'a>], Type<'a>)> for FunctionType<'a> {
    fn from((args, returns): (&mut [Type<'a>], Type<'a>)) -> Self {
        let count = args.len() as u32;
        let (returns, mut args) = (returns.as_ptr(), args.iter().map(|ty| ty.as_ptr()).collect::<Vec<_>>());
        unsafe { Self::from_ptr(LLVMFunctionType(returns, args.as_mut_ptr(), count, 0)) }
    }
}

macro_rules! type_is {
    ($llvm:ident as $name:ident) => {
        paste::paste! {
            #[doc = "Tells wether the type is " $name " or not."]
            pub fn [< is_$name >](&self) -> bool {
                (unsafe { LLVMGetTypeKind(self.as_ptr()) }) == LLVMTypeKind::$llvm
            }
        }
    };
}

impl Type<'_> {
    type_is! { LLVMIntegerTypeKind  as int   }
    type_is! { LLVMVoidTypeKind     as void  }
    type_is! { LLVMArrayTypeKind    as array }
    type_is! { LLVMFloatTypeKind    as f32   }
    type_is! { LLVMLabelTypeKind    as label }
    type_is! { LLVMPointerTypeKind  as ptr   }
    type_is! { LLVMVectorTypeKind   as vec   }
}
