//! Utilities for ASTs (Abstract Syntax Trees).

use crate::{
    ast::parsers::ParserState,
    node::Node,
    tokenizer::{Position, Symbol, Token, TokenReference, TokenType},
    util::*,
    ShortString,
};
use crate::{traits::DefaultRef, util::display_option};
use derive_more::{Deref, Display, IsVariant};
use serde::{Deserialize, Serialize};
use std::{borrow::Cow, fmt};
use vvs_parser_derive::{Node, Visit};

pub use crate::ast::{
    options::OptionTable,
    parsers::{AstResult, OptionTableResult},
    punctuated::*,
    span::*,
};

mod options;
mod parsers;
mod punctuated;
mod span;
mod update_positions;
mod visitors;

/// A block of statements, such as in if/do/etc block
#[derive(Clone, Debug, Default, Display, PartialEq, Node, Visit, Deserialize, Serialize)]
#[display("{}{}",
    display_optional_punctuated_vec(stmts),
    display_option(last_stmt.as_ref().map(display_optional_punctuated))
)]
pub struct Block {
    stmts: Vec<(Stmt, Option<TokenReference>)>,

    #[serde(skip_serializing_if = "Option::is_none")]
    last_stmt: Option<(LastStmt, Option<TokenReference>)>,
}

impl Block {
    /// Creates an empty block
    pub fn new() -> Self {
        Default::default()
    }

    /// An iterator over the statements in the block, such as `local foo = 1`.
    ///
    /// Note that this does not contain the final statement which can be
    /// attained via [`Block::last_stmt`].
    pub fn stmts(&self) -> impl Iterator<Item = &Stmt> {
        self.stmts.iter().map(|(stmt, _)| stmt)
    }

    /// An iterator over the mutable statements in the block, such as `local foo = 1`.
    ///
    /// Note that this does not contain the final statement which can be
    /// attained via [`Block::last_stmt`].
    #[allow(dead_code)]
    pub(crate) fn stmts_mut(&mut self) -> impl Iterator<Item = &mut Stmt> {
        self.stmts.iter_mut().map(|(stmt, _)| stmt)
    }

    /// An iterator over the statements in the block, including any optional
    /// semicolon token reference present
    pub fn stmts_with_semicolon(&self) -> impl Iterator<Item = &(Stmt, Option<TokenReference>)> {
        self.stmts.iter()
    }

    /// The last statement of the block if one exists, such as `return foo`
    pub fn last_stmt(&self) -> Option<&LastStmt> {
        Some(&self.last_stmt.as_ref()?.0)
    }

    /// The mutable last statement of the block if one exists, such as `return foo`
    pub fn last_stmt_mut(&mut self) -> Option<&mut LastStmt> {
        Some(&mut self.last_stmt.as_mut()?.0)
    }

    /// The last statement of the block if on exists, including any optional semicolon token reference present
    pub fn last_stmt_with_semicolon(&self) -> Option<&(LastStmt, Option<TokenReference>)> {
        self.last_stmt.as_ref()
    }

    /// Returns a new block with the given statements
    /// Takes a vector of statements, followed by an optional semicolon token reference
    pub fn with_stmts(self, stmts: Vec<(Stmt, Option<TokenReference>)>) -> Self {
        Self { stmts, ..self }
    }

    /// Returns a new block with the given last statement, if one is given
    /// Takes an optional last statement, with an optional semicolon
    pub fn with_last_stmt(self, last_stmt: Option<(LastStmt, Option<TokenReference>)>) -> Self {
        Self { last_stmt, ..self }
    }

    pub(crate) fn merge_blocks(&mut self, other: Self) {
        self.stmts.extend(other.stmts);
        self.last_stmt = other.last_stmt;
    }
}

/// The last statement of a [`Block`]
#[derive(Clone, Debug, Display, PartialEq, Node, Visit, Deserialize, Serialize)]
#[non_exhaustive]
pub enum LastStmt {
    /// A `break` statement
    Break(TokenReference),

    /// A continue statement
    Continue(TokenReference),

    /// A `return` statement
    Return(Return),
}

/// A `return` statement
#[derive(Clone, Debug, Display, PartialEq, Node, Visit, Deserialize, Serialize)]
#[display("{token}{returns}")]
pub struct Return {
    token: TokenReference,
    returns: Punctuated<Expression>,
}

impl Return {
    /// Creates a new empty Return
    /// Default return token is followed by a single space
    pub fn new() -> Self {
        Self { token: TokenReference::basic_symbol("return "), returns: Punctuated::new() }
    }

    /// The `return` token
    pub fn token(&self) -> &TokenReference {
        &self.token
    }

    /// The values being returned
    pub fn returns(&self) -> &Punctuated<Expression> {
        &self.returns
    }

    /// Returns a new Return with the given `return` token
    pub fn with_token(self, token: TokenReference) -> Self {
        Self { token, ..self }
    }

    /// Returns a new Return with the given punctuated sequence
    pub fn with_returns(self, returns: Punctuated<Expression>) -> Self {
        Self { returns, ..self }
    }
}

impl Default for Return {
    fn default() -> Self {
        Self::new()
    }
}

/// Fields of a [`TableConstructor`]
#[derive(Clone, Debug, Display, PartialEq, Node, Deserialize, Serialize)]
#[non_exhaustive]
pub enum Field {
    /// A key in the format of `[expression] = value`
    #[display("{}{key}{}{equal}{value}", brackets.tokens().0, brackets.tokens().1)]
    ExpressionKey {
        /// The `[...]` part of `[expression] = value`
        brackets: ContainedSpan,

        /// The `expression` part of `[expression] = value`
        key: Expression,

        /// The `=` part of `[expression] = value`
        equal: TokenReference,

        /// The `value` part of `[expression] = value`
        value: Expression,
    },

    /// A key in the format of `name = value`
    #[display("{key}{equal}{value}")]
    NameKey {
        /// The `name` part of `name = value`
        key: TokenReference,

        /// The `=` part of `name = value`
        equal: TokenReference,

        /// The `value` part of `name = value`
        value: Expression,
    },

    /// A field with no key, just a value (such as `"a"` in `{ "a" }`)
    #[display("{_0}")]
    NoKey(Expression),
}

impl From<Field> for Expression {
    fn from(value: Field) -> Self {
        match value {
            Field::ExpressionKey { value, .. } | Field::NameKey { value, .. } | Field::NoKey(value) => value,
        }
    }
}

impl<'a> From<&'a Field> for &'a Expression {
    fn from(value: &'a Field) -> Self {
        match value {
            Field::ExpressionKey { value, .. } | Field::NameKey { value, .. } | Field::NoKey(value) => value,
        }
    }
}

/// A table being constructed, such as `{ 1, 2, 3 }` or `{ a = 1 }`
#[derive(Clone, Debug, Display, PartialEq, Node, Visit, Deserialize, Serialize)]
#[display("{}{fields}{}", braces.tokens().0, braces.tokens().1)]
pub struct TableConstructor {
    #[node(full_range)]
    #[visit(contains = "fields")]
    braces: ContainedSpan,
    fields: Punctuated<Field>,
}

impl TableConstructor {
    /// Creates a new empty TableConstructor
    /// Brace tokens are followed by spaces, such that { `fields` }
    pub fn new() -> Self {
        Self {
            braces: ContainedSpan::new(TokenReference::basic_symbol("{ "), TokenReference::basic_symbol(" }")),
            fields: Punctuated::new(),
        }
    }

    /// The braces of the constructor
    pub fn braces(&self) -> &ContainedSpan {
        &self.braces
    }

    /// Returns the [`Punctuated`] sequence of the fields used to create the table
    pub fn fields(&self) -> &Punctuated<Field> {
        &self.fields
    }

    /// Get the number of fields in the table construct
    pub fn fields_len(&self) -> usize {
        self.fields().len()
    }

    /// Returns a new TableConstructor with the given braces
    pub fn with_braces(self, braces: ContainedSpan) -> Self {
        Self { braces, ..self }
    }

    /// Returns a new TableConstructor with the given fields
    pub fn with_fields(self, fields: Punctuated<Field>) -> Self {
        Self { fields, ..self }
    }
}

impl Default for TableConstructor {
    fn default() -> Self {
        Self::new()
    }
}

/// An expression, mostly useful for getting values
#[derive(Clone, Debug, Display, PartialEq, Node, Deserialize, Serialize)]
#[non_exhaustive]
pub enum Expression {
    /// A binary operation, such as `1 + 3`
    #[display("{lhs}{binop}{rhs}")]
    BinaryOperator {
        /// The left hand side of the binary operation, the `1` part of `1 + 3`
        lhs: Box<Expression>,
        /// The binary operation used, the `+` part of `1 + 3`
        binop: BinOp,
        /// The right hand side of the binary operation, the `3` part of `1 + 3`
        rhs: Box<Expression>,
    },

    /// A statement in parentheses, such as `(#list)`
    #[display("{}{expression}{}", contained.tokens().0, contained.tokens().1)]
    Parentheses {
        /// The parentheses of the expression
        #[node(full_range)]
        contained: ContainedSpan,
        /// The expression inside the parentheses
        expression: Box<Expression>,
    },

    /// A unary operation, such as `#list`
    #[display("{unop}{expression}")]
    UnaryOperator {
        /// The unary operation, the `#` part of `#list`
        unop: UnOp,
        /// The expression the operation is being done on, the `list` part of `#list`
        expression: Box<Expression>,
    },

    /// An anonymous function, such as `function() end)`
    #[display("{}{}", _0.0, _0.1)]
    Function(Box<(TokenReference, FunctionBody)>),

    /// A call of a function, such as `call()`
    #[display("{_0}")]
    FunctionCall(FunctionCall),

    /// An if expression, such as `if foo then true else false`.
    #[display("{_0}")]
    IfExpression(IfExpression),

    /// A table constructor, such as `{ 1, 2, 3 }`
    #[display("{_0}")]
    TableConstructor(TableConstructor),

    /// A number token, such as `3.3`
    #[display("{_0}")]
    Number(TokenReference),

    /// A string token, such as `"hello"`
    #[display("{_0}")]
    String(TokenReference),

    /// A symbol, such as `true`
    #[display("{_0}")]
    Symbol(TokenReference),

    /// A value that has been asserted for a particular type, for use in Luau.
    #[display("{expression}{type_assertion}")]
    TypeAssertion {
        /// The expression being asserted
        expression: Box<Expression>,

        /// The type assertion
        type_assertion: TypeAssertion,
    },

    /// A more complex value, such as `call().x`
    #[display("{_0}")]
    Var(Var),
}

/// A statement that stands alone
#[derive(Clone, Debug, Display, PartialEq, Node, Visit, Deserialize, Serialize)]
#[non_exhaustive]
pub enum Stmt {
    /// An assignment, such as `x = 1`
    #[display("{_0}")]
    Assignment(Assignment),

    /// A do block, `do end`
    #[display("{_0}")]
    Do(Do),

    /// A function call on its own, such as `call()`
    #[display("{_0}")]
    FunctionCall(FunctionCall),

    /// A function declaration, such as `function x() end`
    #[display("{_0}")]
    FunctionDeclaration(FunctionDeclaration),

    /// A job declaration, such as `job x(l: lines): lines return l end`
    #[display("{_0}")]
    JobDeclaration(FunctionDeclaration),

    /// A generic for loop, such as `for index, value in pairs(list) do end`
    #[display("{_0}")]
    GenericFor(GenericFor),

    /// An if statement
    #[display("{_0}")]
    If(If),

    /// A local assignment, such as `local x = 1`
    #[display("{_0}")]
    LocalAssignment(LocalAssignment),

    /// A local function declaration, such as `local function x() end`
    #[display("{_0}")]
    LocalFunction(LocalFunction),

    /// An import statement, such as `import "math"`
    #[display("{_0}")]
    Import(Import),

    /// A numeric for loop, such as `for index = 1, 10 do end`
    #[display("{_0}")]
    NumericFor(NumericFor),

    /// A repeat loop
    #[display("{_0}")]
    Repeat(Repeat),

    /// A while loop
    #[display("{_0}")]
    While(While),

    /// The `option` directive, to declare an option
    #[display("{_0}")]
    OptionDecl(OptionDecl),

    /// The `write` directive, to write variables to the final file.
    ///
    /// Can be used like `write "init"` or `write { "init", "last", "etc" }`
    #[display("{_0}")]
    Write(Write),

    /// The `main` directive. See [Main] for more informations.
    #[display("{_0}")]
    Main(Main),

    /// The `yield` statement for jobs to be able to build automatically things in a job
    #[display("{_0}")]
    Yield(Yield),

    /// A compound assignment, such as `+=`
    #[display("{_0}")]
    CompoundAssignment(CompoundAssignment),

    /// A type declaration, such as `type Meters = number`
    TypeDeclaration(TypeDeclaration),
}

/// A `yield` statement
#[derive(Clone, Debug, Display, PartialEq, Node, Visit, Deserialize, Serialize)]
#[display("{token}{yielded}")]
pub struct Yield {
    token: TokenReference,
    yielded: Punctuated<Expression>,
}

impl Yield {
    /// Creates a new empty Yield
    /// Default yield token is followed by a single space
    pub fn new() -> Self {
        Self { token: TokenReference::basic_symbol("yield "), yielded: Punctuated::new() }
    }

    /// The `yield` token
    pub fn token(&self) -> &TokenReference {
        &self.token
    }

    /// The values being returned
    pub fn yields(&self) -> &Punctuated<Expression> {
        &self.yielded
    }

    /// Returns a new Return with the given `yield` token
    pub fn with_token(self, token: TokenReference) -> Self {
        Self { token, ..self }
    }

    /// Returns a new Yield with the given punctuated sequence
    pub fn with_yields(self, yielded: Punctuated<Expression>) -> Self {
        Self { yielded, ..self }
    }
}

impl Default for Yield {
    fn default() -> Self {
        Self::new()
    }
}

/// A node used before another in cases such as function calling
/// The `("foo")` part of `("foo"):upper()`
#[derive(Clone, Debug, Display, PartialEq, Node, Visit, Deserialize, Serialize)]
#[non_exhaustive]
pub enum Prefix {
    #[display("{_0}")]
    /// A complicated expression, such as `("foo")`
    Expression(Box<Expression>),

    #[display("{_0}")]
    /// Just a name, such as `foo`
    Name(TokenReference),
}

/// The indexing of something, such as `x.y` or `x["y"]`
/// Values of variants are the keys, such as `"y"`
#[derive(Clone, Debug, Display, PartialEq, Node, Deserialize, Serialize)]
#[non_exhaustive]
pub enum Index {
    /// Indexing in the form of `x["y"]`
    #[display("{}{expression}{}", brackets.tokens().0, brackets.tokens().1)]
    Brackets {
        /// The `[...]` part of `["y"]`
        brackets: ContainedSpan,
        /// The `"y"` part of `["y"]`
        expression: Expression,
    },

    /// Indexing in the form of `x.y`
    #[display("{dot}{name}")]
    Dot {
        /// The `.` part of `.y`
        dot: TokenReference,
        /// The `y` part of `.y`
        name: TokenReference,
    },
}

/// Arguments used for a function
#[derive(Clone, Debug, Display, PartialEq, Node, Deserialize, Serialize)]
#[non_exhaustive]
pub enum FunctionArgs {
    /// Used when a function is called in the form of `call(1, 2, 3)`
    #[display("{}{arguments}{}", parentheses.tokens().0, parentheses.tokens().1)]
    Parentheses {
        /// The `(...) part of (1, 2, 3)`
        #[node(full_range)]
        parentheses: ContainedSpan,
        /// The `1, 2, 3` part of `1, 2, 3`
        arguments: Punctuated<Expression>,
    },

    /// Used when a function is called in the form of `call "foobar"`
    #[display("{_0}")]
    String(TokenReference),

    /// Used when a function is called in the form of `call { 1, 2, 3 }`
    #[display("{_0}")]
    TableConstructor(TableConstructor),
}

impl FunctionArgs {
    pub(crate) fn empty() -> Self {
        FunctionArgs::Parentheses {
            parentheses: ContainedSpan::new(TokenReference::basic_symbol("("), TokenReference::basic_symbol(")")),
            arguments: Punctuated::new(),
        }
    }

    /// Like [FunctionArgs::into_arguments], but without consuming the [FunctionArgs].
    pub fn as_arguments(&self) -> Vec<Cow<'_, Field>> {
        match self {
            FunctionArgs::String(string) => vec![Cow::Owned(Field::NoKey(Expression::String(string.clone())))],
            FunctionArgs::TableConstructor(table) => table.fields().iter().map(Cow::Borrowed).collect(),
            FunctionArgs::Parentheses { arguments, .. } => {
                arguments.iter().cloned().map(Field::NoKey).map(Cow::Owned).collect()
            }
        }
    }

    /// Turn the arguments from a function call into a list of fields that are used to call the
    /// function…
    pub fn into_arguments(self) -> Vec<Field> {
        match self {
            FunctionArgs::Parentheses { arguments, .. } => arguments.into_iter().map(Field::NoKey).collect(),
            FunctionArgs::TableConstructor(table) => table.fields.into_iter().collect(),
            FunctionArgs::String(string) => vec![Field::NoKey(Expression::String(string))],
        }
    }
}

/// A numeric for loop, such as `for index = 1, 10 do end`
#[derive(Clone, Debug, PartialEq, Display, Node, Deserialize, Serialize)]
#[display("{for_token}{index_variable}{}{equal_token}{start}{start_end_comma}{end}{}{}{do_token}{block}{end_token}",
    display_option(self.type_specifier()),
    display_option(self.end_step_comma()),
    display_option(self.step()),
)]
pub struct NumericFor {
    for_token: TokenReference,
    index_variable: TokenReference,
    equal_token: TokenReference,
    start: Expression,
    start_end_comma: TokenReference,
    end: Expression,
    end_step_comma: Option<TokenReference>,
    step: Option<Expression>,
    do_token: TokenReference,
    block: Block,
    end_token: TokenReference,
    #[serde(skip_serializing_if = "Option::is_none")]
    type_specifier: Option<TypeSpecifier>,
}

impl NumericFor {
    /// Creates a new NumericFor from the given index variable, start, and end expressions
    pub fn new(index_variable: TokenReference, start: Expression, end: Expression) -> Self {
        Self {
            for_token: TokenReference::basic_symbol("for "),
            index_variable,
            equal_token: TokenReference::basic_symbol(" = "),
            start,
            start_end_comma: TokenReference::basic_symbol(", "),
            end,
            end_step_comma: None,
            step: None,
            do_token: TokenReference::basic_symbol(" do\n"),
            block: Block::new(),
            end_token: TokenReference::basic_symbol("\nend"),
            type_specifier: None,
        }
    }

    /// The `for` token
    pub fn for_token(&self) -> &TokenReference {
        &self.for_token
    }

    /// The index identity, `index` in the initial example
    pub fn index_variable(&self) -> &TokenReference {
        &self.index_variable
    }

    /// The `=` token
    pub fn equal_token(&self) -> &TokenReference {
        &self.equal_token
    }

    /// The starting point, `1` in the initial example
    pub fn start(&self) -> &Expression {
        &self.start
    }

    /// The comma in between the starting point and end point
    /// for _ = 1, 10 do
    ///          ^
    pub fn start_end_comma(&self) -> &TokenReference {
        &self.start_end_comma
    }

    /// The ending point, `10` in the initial example
    pub fn end(&self) -> &Expression {
        &self.end
    }

    /// The comma in between the ending point and limit, if one exists
    /// for _ = 0, 10, 2 do
    ///              ^
    pub fn end_step_comma(&self) -> Option<&TokenReference> {
        self.end_step_comma.as_ref()
    }

    /// The step if one exists, `2` in `for index = 0, 10, 2 do end`
    pub fn step(&self) -> Option<&Expression> {
        self.step.as_ref()
    }

    /// The `do` token
    pub fn do_token(&self) -> &TokenReference {
        &self.do_token
    }

    /// The code inside the for loop
    pub fn block(&self) -> &Block {
        &self.block
    }

    /// The `end` token
    pub fn end_token(&self) -> &TokenReference {
        &self.end_token
    }

    /// The type specifiers of the index variable
    /// `for i: number = 1, 10 do` returns:
    /// `Some(TypeSpecifier(number))`
    pub fn type_specifier(&self) -> Option<&TypeSpecifier> {
        self.type_specifier.as_ref()
    }

    /// Returns a new NumericFor with the given for token
    pub fn with_for_token(self, for_token: TokenReference) -> Self {
        Self { for_token, ..self }
    }

    /// Returns a new NumericFor with the given index variable
    pub fn with_index_variable(self, index_variable: TokenReference) -> Self {
        Self { index_variable, ..self }
    }

    /// Returns a new NumericFor with the given `=` token
    pub fn with_equal_token(self, equal_token: TokenReference) -> Self {
        Self { equal_token, ..self }
    }

    /// Returns a new NumericFor with the given start expression
    pub fn with_start(self, start: Expression) -> Self {
        Self { start, ..self }
    }

    /// Returns a new NumericFor with the given comma between the start and end expressions
    pub fn with_start_end_comma(self, start_end_comma: TokenReference) -> Self {
        Self { start_end_comma, ..self }
    }

    /// Returns a new NumericFor with the given end expression
    pub fn with_end(self, end: Expression) -> Self {
        Self { end, ..self }
    }

    /// Returns a new NumericFor with the given comma between the end and the step expressions
    pub fn with_end_step_comma(self, end_step_comma: Option<TokenReference>) -> Self {
        Self { end_step_comma, ..self }
    }

    /// Returns a new NumericFor with the given step expression
    pub fn with_step(self, step: Option<Expression>) -> Self {
        Self { step, ..self }
    }

    /// Returns a new NumericFor with the given `do` token
    pub fn with_do_token(self, do_token: TokenReference) -> Self {
        Self { do_token, ..self }
    }

    /// Returns a new NumericFor with the given block
    pub fn with_block(self, block: Block) -> Self {
        Self { block, ..self }
    }

    /// Returns a new NumericFor with the given `end` token
    pub fn with_end_token(self, end_token: TokenReference) -> Self {
        Self { end_token, ..self }
    }

    /// Returns a new NumericFor with the given type specifiers
    pub fn with_type_specifier(self, type_specifier: Option<TypeSpecifier>) -> Self {
        Self { type_specifier, ..self }
    }
}

/// A generic for loop, such as `for index, value in pairs(list) do end`
#[derive(Clone, Debug, PartialEq, Display, Node, Deserialize, Serialize)]
#[display("{for_token}{}{in_token}{expr_list}{do_token}{block}{end_token}",
    join_type_specifiers(&self.names, self.type_specifiers()),
)]
pub struct GenericFor {
    for_token: TokenReference,
    names: Punctuated<TokenReference>,
    in_token: TokenReference,
    expr_list: Punctuated<Expression>,
    do_token: TokenReference,
    block: Block,
    end_token: TokenReference,
    type_specifiers: Vec<Option<TypeSpecifier>>,
}

impl GenericFor {
    /// Creates a new GenericFor from the given names and expressions
    pub fn new(names: Punctuated<TokenReference>, expr_list: Punctuated<Expression>) -> Self {
        Self {
            for_token: TokenReference::basic_symbol("for "),
            names,
            in_token: TokenReference::basic_symbol(" in "),
            expr_list,
            do_token: TokenReference::basic_symbol(" do\n"),
            block: Block::new(),
            end_token: TokenReference::basic_symbol("\nend"),
            type_specifiers: Vec::new(),
        }
    }

    /// The `for` token
    pub fn for_token(&self) -> &TokenReference {
        &self.for_token
    }

    /// Returns the punctuated sequence of names
    /// In `for index, value in pairs(list) do`, iterates over `index` and `value`
    pub fn names(&self) -> &Punctuated<TokenReference> {
        &self.names
    }

    /// The `in` token
    pub fn in_token(&self) -> &TokenReference {
        &self.in_token
    }

    /// Returns the punctuated sequence of the expressions looped over
    /// In `for index, value in pairs(list) do`, iterates over `pairs(list)`
    pub fn expressions(&self) -> &Punctuated<Expression> {
        &self.expr_list
    }

    /// The `do` token
    pub fn do_token(&self) -> &TokenReference {
        &self.do_token
    }

    /// The code inside the for loop
    pub fn block(&self) -> &Block {
        &self.block
    }

    /// The `end` token
    pub fn end_token(&self) -> &TokenReference {
        &self.end_token
    }

    /// The type specifiers of the named variables, in the order that they were assigned.
    /// `for i, v: string in pairs() do` returns an iterator containing:
    /// `None, Some(TypeSpecifier(string))`
    pub fn type_specifiers(&self) -> impl Iterator<Item = Option<&TypeSpecifier>> {
        self.type_specifiers.iter().map(Option::as_ref)
    }

    /// Returns a new GenericFor with the given `for` token
    pub fn with_for_token(self, for_token: TokenReference) -> Self {
        Self { for_token, ..self }
    }

    /// Returns a new GenericFor with the given names
    pub fn with_names(self, names: Punctuated<TokenReference>) -> Self {
        Self { names, ..self }
    }

    /// Returns a new GenericFor with the given `in` token
    pub fn with_in_token(self, in_token: TokenReference) -> Self {
        Self { in_token, ..self }
    }

    /// Returns a new GenericFor with the given expression list
    pub fn with_expressions(self, expr_list: Punctuated<Expression>) -> Self {
        Self { expr_list, ..self }
    }

    /// Returns a new GenericFor with the given `do` token
    pub fn with_do_token(self, do_token: TokenReference) -> Self {
        Self { do_token, ..self }
    }

    /// Returns a new GenericFor with the given block
    pub fn with_block(self, block: Block) -> Self {
        Self { block, ..self }
    }

    /// Returns a new GenericFor with the given `end` token
    pub fn with_end_token(self, end_token: TokenReference) -> Self {
        Self { end_token, ..self }
    }

    /// Returns a new GenericFor with the given type specifiers
    pub fn with_type_specifiers(self, type_specifiers: Vec<Option<TypeSpecifier>>) -> Self {
        Self { type_specifiers, ..self }
    }
}

/// An if statement
#[derive(Clone, Debug, Display, PartialEq, Node, Visit, Deserialize, Serialize)]
#[display("{if_token}{condition}{then_token}{block}{}{}{}{end_token}",
    display_option(else_if.as_ref().map(join_vec)),
    display_option(else_token),
    display_option(r#else),
)]
pub struct If {
    if_token: TokenReference,
    condition: Expression,
    then_token: TokenReference,
    block: Block,
    else_if: Option<Vec<ElseIf>>,
    else_token: Option<TokenReference>,

    #[serde(rename = "else")]
    r#else: Option<Block>,
    end_token: TokenReference,
}

impl If {
    /// Creates a new If from the given condition
    pub fn new(condition: Expression) -> Self {
        Self {
            if_token: TokenReference::basic_symbol("if "),
            condition,
            then_token: TokenReference::basic_symbol(" then"),
            block: Block::new(),
            else_if: None,
            else_token: None,
            r#else: None,
            end_token: TokenReference::basic_symbol("\nend"),
        }
    }

    /// The `if` token
    pub fn if_token(&self) -> &TokenReference {
        &self.if_token
    }

    /// The condition of the if statement, `condition` in `if condition then`
    pub fn condition(&self) -> &Expression {
        &self.condition
    }

    /// The `then` token
    pub fn then_token(&self) -> &TokenReference {
        &self.then_token
    }

    /// The block inside the initial if statement
    pub fn block(&self) -> &Block {
        &self.block
    }

    /// The `else` token if one exists
    pub fn else_token(&self) -> Option<&TokenReference> {
        self.else_token.as_ref()
    }

    /// If there are `elseif` conditions, returns a vector of them
    /// Expression is the condition, block is the code if the condition is true
    // TODO: Make this return an iterator, and remove Option part entirely?
    pub fn else_if(&self) -> Option<&Vec<ElseIf>> {
        self.else_if.as_ref()
    }

    /// The code inside an `else` block if one exists
    pub fn else_block(&self) -> Option<&Block> {
        self.r#else.as_ref()
    }

    /// The `end` token
    pub fn end_token(&self) -> &TokenReference {
        &self.end_token
    }

    /// Returns a new If with the given `if` token
    pub fn with_if_token(self, if_token: TokenReference) -> Self {
        Self { if_token, ..self }
    }

    /// Returns a new If with the given condition
    pub fn with_condition(self, condition: Expression) -> Self {
        Self { condition, ..self }
    }

    /// Returns a new If with the given `then` token
    pub fn with_then_token(self, then_token: TokenReference) -> Self {
        Self { then_token, ..self }
    }

    /// Returns a new If with the given block
    pub fn with_block(self, block: Block) -> Self {
        Self { block, ..self }
    }

    /// Returns a new If with the given list of `elseif` blocks
    pub fn with_else_if(self, else_if: Option<Vec<ElseIf>>) -> Self {
        Self { else_if, ..self }
    }

    /// Returns a new If with the given `else` token
    pub fn with_else_token(self, else_token: Option<TokenReference>) -> Self {
        Self { else_token, ..self }
    }

    /// Returns a new If with the given `else` body
    pub fn with_else(self, r#else: Option<Block>) -> Self {
        Self { r#else, ..self }
    }

    /// Returns a new If with the given `end` token
    pub fn with_end_token(self, end_token: TokenReference) -> Self {
        Self { end_token, ..self }
    }
}

/// An elseif block in a bigger [`If`] statement
#[derive(Clone, Debug, Display, PartialEq, Node, Visit, Deserialize, Serialize)]
#[display("{else_if_token}{condition}{then_token}{block}")]
pub struct ElseIf {
    else_if_token: TokenReference,
    condition: Expression,
    then_token: TokenReference,
    block: Block,
}

impl ElseIf {
    /// Creates a new ElseIf from the given condition
    pub fn new(condition: Expression) -> Self {
        Self {
            else_if_token: TokenReference::basic_symbol("elseif "),
            condition,
            then_token: TokenReference::basic_symbol(" then\n"),
            block: Block::new(),
        }
    }

    /// The `elseif` token
    pub fn else_if_token(&self) -> &TokenReference {
        &self.else_if_token
    }

    /// The condition of the `elseif`, `condition` in `elseif condition then`
    pub fn condition(&self) -> &Expression {
        &self.condition
    }

    /// The `then` token
    pub fn then_token(&self) -> &TokenReference {
        &self.then_token
    }

    /// The body of the `elseif`
    pub fn block(&self) -> &Block {
        &self.block
    }

    /// Returns a new ElseIf with the given `elseif` token
    pub fn with_else_if_token(self, else_if_token: TokenReference) -> Self {
        Self { else_if_token, ..self }
    }

    /// Returns a new ElseIf with the given condition
    pub fn with_condition(self, condition: Expression) -> Self {
        Self { condition, ..self }
    }

    /// Returns a new ElseIf with the given `then` token
    pub fn with_then_token(self, then_token: TokenReference) -> Self {
        Self { then_token, ..self }
    }

    /// Returns a new ElseIf with the given block
    pub fn with_block(self, block: Block) -> Self {
        Self { block, ..self }
    }
}

/// A while loop
#[derive(Clone, Debug, Display, PartialEq, Node, Visit, Deserialize, Serialize)]
#[display("{while_token}{condition}{do_token}{block}{end_token}")]
pub struct While {
    while_token: TokenReference,
    condition: Expression,
    do_token: TokenReference,
    block: Block,
    end_token: TokenReference,
}

impl While {
    /// Creates a new While from the given condition
    pub fn new(condition: Expression) -> Self {
        Self {
            while_token: TokenReference::basic_symbol("while "),
            condition,
            do_token: TokenReference::basic_symbol(" do\n"),
            block: Block::new(),
            end_token: TokenReference::basic_symbol("end\n"),
        }
    }

    /// The `while` token
    pub fn while_token(&self) -> &TokenReference {
        &self.while_token
    }

    /// The `condition` part of `while condition do`
    pub fn condition(&self) -> &Expression {
        &self.condition
    }

    /// The `do` token
    pub fn do_token(&self) -> &TokenReference {
        &self.do_token
    }

    /// The code inside the while loop
    pub fn block(&self) -> &Block {
        &self.block
    }

    /// The `end` token
    pub fn end_token(&self) -> &TokenReference {
        &self.end_token
    }

    /// Returns a new While with the given `while` token
    pub fn with_while_token(self, while_token: TokenReference) -> Self {
        Self { while_token, ..self }
    }

    /// Returns a new While with the given condition
    pub fn with_condition(self, condition: Expression) -> Self {
        Self { condition, ..self }
    }

    /// Returns a new While with the given `do` token
    pub fn with_do_token(self, do_token: TokenReference) -> Self {
        Self { do_token, ..self }
    }

    /// Returns a new While with the given block
    pub fn with_block(self, block: Block) -> Self {
        Self { block, ..self }
    }

    /// Returns a new While with the given `end` token
    pub fn with_end_token(self, end_token: TokenReference) -> Self {
        Self { end_token, ..self }
    }
}

/// A repeat loop
#[derive(Clone, Debug, Display, PartialEq, Node, Visit, Deserialize, Serialize)]
#[display("{repeat_token}{block}{until_token}{until}")]
pub struct Repeat {
    repeat_token: TokenReference,
    block: Block,
    until_token: TokenReference,
    until: Expression,
}

impl Repeat {
    /// Creates a new Repeat from the given expression to repeat until
    pub fn new(until: Expression) -> Self {
        Self {
            repeat_token: TokenReference::basic_symbol("repeat\n"),
            block: Block::new(),
            until_token: TokenReference::basic_symbol("\nuntil "),
            until,
        }
    }

    /// The `repeat` token
    pub fn repeat_token(&self) -> &TokenReference {
        &self.repeat_token
    }

    /// The code inside the `repeat` block
    pub fn block(&self) -> &Block {
        &self.block
    }

    /// The `until` token
    pub fn until_token(&self) -> &TokenReference {
        &self.until_token
    }

    /// The condition for the `until` part
    pub fn until(&self) -> &Expression {
        &self.until
    }

    /// Returns a new Repeat with the given `repeat` token
    pub fn with_repeat_token(self, repeat_token: TokenReference) -> Self {
        Self { repeat_token, ..self }
    }

    /// Returns a new Repeat with the given block
    pub fn with_block(self, block: Block) -> Self {
        Self { block, ..self }
    }

    /// Returns a new Repeat with the given `until` token
    pub fn with_until_token(self, until_token: TokenReference) -> Self {
        Self { until_token, ..self }
    }

    /// Returns a new Repeat with the given `until` block
    pub fn with_until(self, until: Expression) -> Self {
        Self { until, ..self }
    }
}

/// An attribute on a local variable, `<const>` in `local x <const>`
#[derive(Clone, Debug, Display, PartialEq, Eq, Node, Visit, Deserialize, Serialize)]
#[display("{}{name}{}", brackets.tokens().0, brackets.tokens().1)]
pub struct Attribute {
    #[node(full_range)]
    #[visit(contains = "name")]
    pub(crate) brackets: ContainedSpan,
    pub(crate) name: TokenReference,
}

impl Attribute {
    /// Creates a new Label with the given name
    pub fn new(name: TokenReference) -> Self {
        Self {
            brackets: ContainedSpan::new(TokenReference::symbol("<").unwrap(), TokenReference::symbol(">").unwrap()),
            name,
        }
    }

    /// The name used for the attribute, the `const` part of `<const>`
    pub fn name(&self) -> &TokenReference {
        &self.name
    }

    /// The angle brackets (`<` and `>`) surrounding the attribute
    pub fn brackets(&self) -> &ContainedSpan {
        &self.brackets
    }

    /// Returns a new Attribute with the given attribute name
    pub fn with_name(self, name: TokenReference) -> Self {
        Self { name, ..self }
    }

    /// Returns a new Attribute with the given angle brackets
    pub fn with_brackets(self, brackets: ContainedSpan) -> Self {
        Self { brackets, ..self }
    }
}

/// A method call, such as `x:y()`
#[derive(Clone, Debug, Display, PartialEq, Node, Visit, Deserialize, Serialize)]
#[display("{colon_token}{name}{args}")]
pub struct MethodCall {
    colon_token: TokenReference,
    name: TokenReference,
    args: FunctionArgs,
}

impl MethodCall {
    /// Returns a new MethodCall from the given name and args
    pub fn new(name: TokenReference, args: FunctionArgs) -> Self {
        Self { colon_token: TokenReference::basic_symbol(":"), name, args }
    }

    /// The `:` in `x:y()`
    pub fn colon_token(&self) -> &TokenReference {
        &self.colon_token
    }

    /// The arguments of a method call, the `x, y, z` part of `method:call(x, y, z)`
    pub fn args(&self) -> &FunctionArgs {
        &self.args
    }

    /// The method being called, the `call` part of `method:call()`
    pub fn name(&self) -> &TokenReference {
        &self.name
    }

    /// Returns a new MethodCall with the given `:` token
    pub fn with_colon_token(self, colon_token: TokenReference) -> Self {
        Self { colon_token, ..self }
    }

    /// Returns a new MethodCall with the given name
    pub fn with_name(self, name: TokenReference) -> Self {
        Self { name, ..self }
    }

    /// Returns a new MethodCall with the given args
    pub fn with_args(self, args: FunctionArgs) -> Self {
        Self { args, ..self }
    }
}

/// Something being called
#[derive(Clone, Debug, Display, PartialEq, Node, Visit, Deserialize, Serialize)]
#[non_exhaustive]
pub enum Call {
    #[display("{_0}")]
    /// A function being called directly, such as `x(1)`
    AnonymousCall(FunctionArgs),

    #[display("{_0}")]
    /// A method call, such as `x:y()`
    MethodCall(MethodCall),
}

/// A function body, everything except `function x` in `function x(a, b, c) call() end`
#[derive(Clone, Debug, PartialEq, Display, Node, Deserialize, Serialize)]
#[display("{}{}{}{}{block}{end_token}",
    parameters_parentheses.tokens().0,
    join_type_specifiers(parameters, self.type_specifiers()),
    parameters_parentheses.tokens().1,
    display_option(return_type.as_ref()),
)]
pub struct FunctionBody {
    parameters_parentheses: ContainedSpan,
    parameters: Punctuated<Parameter>,

    type_specifiers: Vec<Option<TypeSpecifier>>,

    #[serde(skip_serializing_if = "Option::is_none")]
    return_type: Option<TypeSpecifier>,

    block: Block,
    end_token: TokenReference,
}

impl FunctionBody {
    /// Returns a new empty FunctionBody
    pub fn new() -> Self {
        Self {
            parameters_parentheses: ContainedSpan::new(
                TokenReference::basic_symbol("("),
                TokenReference::basic_symbol(")"),
            ),
            parameters: Punctuated::new(),
            type_specifiers: Vec::new(),
            return_type: None,
            block: Block::new(),
            end_token: TokenReference::basic_symbol("\nend"),
        }
    }

    /// The parentheses of the parameters
    pub fn parameters_parentheses(&self) -> &ContainedSpan {
        &self.parameters_parentheses
    }

    /// Returns the [`Punctuated`] sequence of the parameters for the function declaration
    pub fn parameters(&self) -> &Punctuated<Parameter> {
        &self.parameters
    }

    /// The code of a function body
    pub fn block(&self) -> &Block {
        &self.block
    }

    /// The mutable code of a function body
    pub fn block_mut(&mut self) -> &mut Block {
        &mut self.block
    }

    /// The `end` token
    pub fn end_token(&self) -> &TokenReference {
        &self.end_token
    }

    /// The type specifiers of the variables, in the order that they were assigned.
    /// `(foo: number, bar, baz: boolean)` returns an iterator containing:
    /// `Some(TypeSpecifier(number)), None, Some(TypeSpecifier(boolean))`
    pub fn type_specifiers(&self) -> impl Iterator<Item = Option<&TypeSpecifier>> {
        self.type_specifiers.iter().map(Option::as_ref)
    }

    /// The return type of the function, if one exists.
    pub fn return_type(&self) -> Option<&TypeSpecifier> {
        self.return_type.as_ref()
    }

    /// Returns a new FunctionBody with the given parentheses for the parameters
    pub fn with_parameters_parentheses(self, parameters_parentheses: ContainedSpan) -> Self {
        Self { parameters_parentheses, ..self }
    }

    /// Returns a new FunctionBody with the given parameters
    pub fn with_parameters(self, parameters: Punctuated<Parameter>) -> Self {
        Self { parameters, ..self }
    }

    /// Returns a new FunctionBody with the given type specifiers
    pub fn with_type_specifiers(self, type_specifiers: Vec<Option<TypeSpecifier>>) -> Self {
        Self { type_specifiers, ..self }
    }

    /// Returns a new FunctionBody with the given return type
    pub fn with_return_type(self, return_type: Option<TypeSpecifier>) -> Self {
        Self { return_type, ..self }
    }

    /// Returns a new FunctionBody with the given block
    pub fn with_block(self, block: Block) -> Self {
        Self { block, ..self }
    }

    /// Returns a new FunctionBody with the given `end` token
    pub fn with_end_token(self, end_token: TokenReference) -> Self {
        Self { end_token, ..self }
    }
}

impl Default for FunctionBody {
    fn default() -> Self {
        Self::new()
    }
}

/// A name parameter in a function declaration, such as `function x(a, b, c)`
#[derive(Clone, Debug, Display, Deref, PartialEq, Eq, Node, Visit, Deserialize, Serialize)]
#[display("{name}")]
pub struct Parameter {
    name: TokenReference,
}

impl Parameter {
    /// Get the name of the parameter.
    pub fn name(&self) -> &TokenReference {
        &self.name
    }

    /// Get the name of the parameter. Consumes the [Parameter].
    pub fn into_name(self) -> TokenReference {
        self.name
    }
}

/// A suffix in certain cases, such as `:y()` in `x:y()`
/// Can be stacked on top of each other, such as in `x()()()`
#[derive(Clone, Debug, Display, PartialEq, Node, Visit, Deserialize, Serialize)]
#[non_exhaustive]
pub enum Suffix {
    #[display("{_0}")]
    /// A call, including method calls and direct calls
    Call(Call),

    #[display("{_0}")]
    /// An index, such as `x.y`
    Index(Index),
}

/// A complex expression used by [`Var`], consisting of both a prefix and suffixes
#[derive(Clone, Debug, Display, PartialEq, Node, Visit, Deserialize, Serialize)]
#[display("{prefix}{}", join_vec(suffixes))]
pub struct VarExpression {
    prefix: Prefix,
    suffixes: Vec<Suffix>,
}

impl VarExpression {
    /// Returns a new VarExpression from the given prefix
    pub fn new(prefix: Prefix) -> Self {
        Self { prefix, suffixes: Vec::new() }
    }

    /// The prefix of the expression, such as a name
    pub fn prefix(&self) -> &Prefix {
        &self.prefix
    }

    /// An iter over the suffixes, such as indexing or calling
    pub fn suffixes(&self) -> impl Iterator<Item = &Suffix> {
        self.suffixes.iter()
    }

    /// Returns a new VarExpression with the given prefix
    pub fn with_prefix(self, prefix: Prefix) -> Self {
        Self { prefix, ..self }
    }

    /// Returns a new VarExpression with the given suffixes
    pub fn with_suffixes(self, suffixes: Vec<Suffix>) -> Self {
        Self { suffixes, ..self }
    }
}

/// Used in [`Assignment`s](Assignment) and [`Value`s](Value)
#[derive(Clone, Debug, Display, PartialEq, Node, Visit, Deserialize, Serialize)]
#[non_exhaustive]
pub enum Var {
    /// An expression, such as `x.y.z` or `x()`
    #[display("{_0}")]
    Expression(Box<VarExpression>),

    /// A literal identifier, such as `x`
    #[display("{_0}")]
    Name(TokenReference),
}

/// An assignment, such as `x = y`. Not used for [`LocalAssignment`s](LocalAssignment)
#[derive(Clone, Debug, Display, PartialEq, Node, Visit, Deserialize, Serialize)]
#[display("{var_list}{equal_token}{expr_list}")]
pub struct Assignment {
    var_list: Punctuated<Var>,
    equal_token: TokenReference,
    expr_list: Punctuated<Expression>,
}

impl Assignment {
    /// Returns a new Assignment from the given variable and expression list
    pub fn new(var_list: Punctuated<Var>, expr_list: Punctuated<Expression>) -> Self {
        Self { var_list, equal_token: TokenReference::basic_symbol(" = "), expr_list }
    }

    /// Returns the punctuated sequence over the expressions being assigned.
    /// This is the the `1, 2` part of `x, y["a"] = 1, 2`
    pub fn expressions(&self) -> &Punctuated<Expression> {
        &self.expr_list
    }

    /// The `=` token in between `x = y`
    pub fn equal_token(&self) -> &TokenReference {
        &self.equal_token
    }

    /// Returns the punctuated sequence over the variables being assigned to.
    /// This is the `x, y["a"]` part of `x, y["a"] = 1, 2`
    pub fn variables(&self) -> &Punctuated<Var> {
        &self.var_list
    }

    /// Returns a new Assignment with the given variables
    pub fn with_variables(self, var_list: Punctuated<Var>) -> Self {
        Self { var_list, ..self }
    }

    /// Returns a new Assignment with the given `=` token
    pub fn with_equal_token(self, equal_token: TokenReference) -> Self {
        Self { equal_token, ..self }
    }

    /// Returns a new Assignment with the given expressions
    pub fn with_expressions(self, expr_list: Punctuated<Expression>) -> Self {
        Self { expr_list, ..self }
    }
}

/// An import directive, such as `import "math"`
#[derive(Clone, Debug, Display, PartialEq, Node, Visit, Deserialize, Serialize)]
#[display("{import_token}{name}")]
pub struct Import {
    import_token: TokenReference,
    name: TokenReference,
}

impl Import {
    /// Returns a new LocalFunction from the given name
    pub fn new(name: TokenReference) -> Self {
        Import { import_token: TokenReference::basic_symbol("import "), name }
    }

    /// The `import` token
    pub fn import_token(&self) -> &TokenReference {
        &self.import_token
    }

    /// The name of the module to import, the `x` part of `import "x"`
    pub fn name(&self) -> &TokenReference {
        &self.name
    }

    /// Returns a new Import with the given `import` token
    pub fn with_import_token(self, import_token: TokenReference) -> Self {
        Self { import_token, ..self }
    }

    /// Returns a new Import with the given name
    pub fn with_name(self, name: TokenReference) -> Self {
        Self { name, ..self }
    }
}

/// A declaration of a local function, such as `local function x() end`
#[derive(Clone, Debug, Display, PartialEq, Node, Visit, Deserialize, Serialize)]
#[display("{local_token}{function_token}{name}{body}")]
pub struct LocalFunction {
    local_token: TokenReference,
    function_token: TokenReference,
    name: TokenReference,
    body: FunctionBody,
}

impl LocalFunction {
    /// Returns a new LocalFunction from the given name
    pub fn new(name: TokenReference) -> Self {
        LocalFunction {
            local_token: TokenReference::basic_symbol("local "),
            function_token: TokenReference::basic_symbol("function "),
            name,
            body: FunctionBody::new(),
        }
    }

    /// The `local` token
    pub fn local_token(&self) -> &TokenReference {
        &self.local_token
    }

    /// The `function` token
    pub fn function_token(&self) -> &TokenReference {
        &self.function_token
    }

    /// The function body, everything except `local function x` in `local function x(a, b, c) call() end`
    pub fn body(&self) -> &FunctionBody {
        &self.body
    }

    /// The name of the function, the `x` part of `local function x() end`
    pub fn name(&self) -> &TokenReference {
        &self.name
    }

    /// Returns a new LocalFunction with the given `local` token
    pub fn with_local_token(self, local_token: TokenReference) -> Self {
        Self { local_token, ..self }
    }

    /// Returns a new LocalFunction with the given `function` token
    pub fn with_function_token(self, function_token: TokenReference) -> Self {
        Self { function_token, ..self }
    }

    /// Returns a new LocalFunction with the given name
    pub fn with_name(self, name: TokenReference) -> Self {
        Self { name, ..self }
    }

    /// Returns a new LocalFunction with the given function body
    pub fn with_body(self, body: FunctionBody) -> Self {
        Self { body, ..self }
    }
}

/// An assignment to a local variable, such as `local x = 1`
#[derive(Clone, Debug, PartialEq, Display, Node, Deserialize, Serialize)]
#[display("{local_token}{}{}{expr_list}",
    join_iterators(name_list,
        self.attributes().chain(std::iter::repeat(None)),
        self.type_specifiers().chain(std::iter::repeat(None))
    ),
    display_option(equal_token),
)]
pub struct LocalAssignment {
    local_token: TokenReference,

    #[serde(skip_serializing_if = "empty_optional_vector")]
    type_specifiers: Vec<Option<TypeSpecifier>>,
    name_list: Punctuated<TokenReference>,

    #[serde(skip_serializing_if = "empty_optional_vector")]
    attributes: Vec<Option<Attribute>>,
    equal_token: Option<TokenReference>,
    expr_list: Punctuated<Expression>,
}

impl LocalAssignment {
    /// Returns a new LocalAssignment from the given name list
    pub fn new(name_list: Punctuated<TokenReference>) -> Self {
        Self {
            local_token: TokenReference::basic_symbol("local "),
            type_specifiers: Vec::new(),
            name_list,
            attributes: Vec::new(),
            equal_token: None,
            expr_list: Punctuated::new(),
        }
    }

    /// The `local` token
    pub fn local_token(&self) -> &TokenReference {
        &self.local_token
    }

    /// The `=` token in between `local x = y`, if one exists
    pub fn equal_token(&self) -> Option<&TokenReference> {
        self.equal_token.as_ref()
    }

    /// Returns the punctuated sequence of the expressions being assigned.
    /// This is the `1, 2` part of `local x, y = 1, 2`
    pub fn expressions(&self) -> &Punctuated<Expression> {
        &self.expr_list
    }

    /// Returns the punctuated sequence of names being assigned to.
    /// This is the `x, y` part of `local x, y = 1, 2`
    pub fn names(&self) -> &Punctuated<TokenReference> {
        &self.name_list
    }

    /// The type specifiers of the variables, in the order that they were assigned.
    /// `local foo: number, bar, baz: boolean` returns an iterator containing:
    /// `Some(TypeSpecifier(number)), None, Some(TypeSpecifier(boolean))`
    pub fn type_specifiers(&self) -> impl Iterator<Item = Option<&TypeSpecifier>> {
        self.type_specifiers.iter().map(Option::as_ref)
    }

    /// The attributes specified for the variables, in the order that they were assigned.
    /// `local foo <const>, bar, baz <close>` returns an iterator containing:
    /// `Some(Attribute("const")), None, Some(Attribute("close"))`
    pub fn attributes(&self) -> impl Iterator<Item = Option<&Attribute>> {
        self.attributes.iter().map(Option::as_ref)
    }

    /// Returns a new LocalAssignment with the given `local` token
    pub fn with_local_token(self, local_token: TokenReference) -> Self {
        Self { local_token, ..self }
    }

    /// Returns a new LocalAssignment with the given type specifiers
    pub fn with_type_specifiers(self, type_specifiers: Vec<Option<TypeSpecifier>>) -> Self {
        Self { type_specifiers, ..self }
    }

    /// Returns a new LocalAssignment with the given attributes
    pub fn with_attributes(self, attributes: Vec<Option<Attribute>>) -> Self {
        Self { attributes, ..self }
    }

    /// Returns a new LocalAssignment with the given name list
    pub fn with_names(self, name_list: Punctuated<TokenReference>) -> Self {
        Self { name_list, ..self }
    }

    /// Returns a new LocalAssignment with the given `=` token
    pub fn with_equal_token(self, equal_token: Option<TokenReference>) -> Self {
        Self { equal_token, ..self }
    }

    /// Returns a new LocalAssignment with the given expression list
    pub fn with_expressions(self, expr_list: Punctuated<Expression>) -> Self {
        Self { expr_list, ..self }
    }
}

/// A `do` block, such as `do ... end`
/// This is not used for things like `while true do end`, only those on their own
#[derive(Clone, Debug, Display, PartialEq, Node, Visit, Deserialize, Serialize)]
#[display("{do_token}{block}{end_token}")]
pub struct Do {
    do_token: TokenReference,
    block: Block,
    end_token: TokenReference,
}

impl Do {
    /// Creates an empty Do
    pub fn new() -> Self {
        Self {
            do_token: TokenReference::basic_symbol("do\n"),
            block: Block::new(),
            end_token: TokenReference::basic_symbol("\nend"),
        }
    }

    /// The `do` token
    pub fn do_token(&self) -> &TokenReference {
        &self.do_token
    }

    /// The code inside the `do ... end`
    pub fn block(&self) -> &Block {
        &self.block
    }

    /// The `end` token
    pub fn end_token(&self) -> &TokenReference {
        &self.end_token
    }

    /// Returns a new Do with the given `do` token
    pub fn with_do_token(self, do_token: TokenReference) -> Self {
        Self { do_token, ..self }
    }

    /// Returns a new Do with the given block
    pub fn with_block(self, block: Block) -> Self {
        Self { block, ..self }
    }

    /// Returns a new Do with the given `end` token
    pub fn with_end_token(self, end_token: TokenReference) -> Self {
        Self { end_token, ..self }
    }
}

impl Default for Do {
    fn default() -> Self {
        Self::new()
    }
}

/// A function being called, such as `call()`
#[derive(Clone, Debug, Display, PartialEq, Node, Visit, Deserialize, Serialize)]
#[display("{prefix}{}", join_vec(suffixes))]
pub struct FunctionCall {
    prefix: Prefix,
    suffixes: Vec<Suffix>,
}

impl FunctionCall {
    /// Creates a new FunctionCall from the given prefix
    /// Sets the suffixes such that the return is `prefixes()`
    pub fn new(prefix: Prefix) -> Self {
        FunctionCall {
            prefix,
            suffixes: vec![Suffix::Call(Call::AnonymousCall(FunctionArgs::Parentheses {
                arguments: Punctuated::new(),
                parentheses: ContainedSpan::new(TokenReference::basic_symbol("("), TokenReference::basic_symbol(")")),
            }))],
        }
    }

    /// The prefix of a function call, the `call` part of `call()`
    pub fn prefix(&self) -> &Prefix {
        &self.prefix
    }

    /// The suffix of a function call, the `()` part of `call()`
    pub fn suffixes(&self) -> impl Iterator<Item = &Suffix> {
        self.suffixes.iter()
    }

    /// Returns a new FunctionCall with the given prefix
    pub fn with_prefix(self, prefix: Prefix) -> Self {
        Self { prefix, ..self }
    }

    /// Returns a new FunctionCall with the given suffixes
    pub fn with_suffixes(self, suffixes: Vec<Suffix>) -> Self {
        Self { suffixes, ..self }
    }
}

/// A function name when being declared as [`FunctionDeclaration`]
#[derive(Clone, Debug, Display, PartialEq, Eq, Node, Visit, Deserialize, Serialize)]
#[display("{names}{}{}",
    display_option(self.method_colon()),
    display_option(self.method_name())
)]
pub struct FunctionName {
    names: Punctuated<TokenReference>,
    colon_name: Option<(TokenReference, TokenReference)>,
}

impl FunctionName {
    /// Creates a new FunctionName from the given list of names
    pub fn new(names: Punctuated<TokenReference>) -> Self {
        Self { names, colon_name: None }
    }

    /// The colon between the name and the method, the `:` part of `function x:y() end`
    pub fn method_colon(&self) -> Option<&TokenReference> {
        Some(&self.colon_name.as_ref()?.0)
    }

    /// A method name if one exists, the `y` part of `function x:y() end`
    pub fn method_name(&self) -> Option<&TokenReference> {
        Some(&self.colon_name.as_ref()?.1)
    }

    /// Returns the punctuated sequence over the names used when defining the function.
    /// This is the `x.y.z` part of `function x.y.z() end`
    pub fn names(&self) -> &Punctuated<TokenReference> {
        &self.names
    }

    /// Returns a new FunctionName with the given names
    pub fn with_names(self, names: Punctuated<TokenReference>) -> Self {
        Self { names, ..self }
    }

    /// Returns a new FunctionName with the given method name
    /// The first token is the colon, and the second token is the method name itself
    pub fn with_method(self, method: Option<(TokenReference, TokenReference)>) -> Self {
        Self { colon_name: method, ..self }
    }
}

/// A normal function declaration, supports simple declarations like `function x() end`
/// as well as complicated declarations such as `function x.y.z:a() end`
#[derive(Clone, Debug, Display, PartialEq, Node, Visit, Deserialize, Serialize)]
#[display("{function_token}{name}{body}")]
pub struct FunctionDeclaration {
    function_token: TokenReference,
    name: FunctionName,
    body: FunctionBody,
}

impl FunctionDeclaration {
    /// Creates a new FunctionDeclaration from the given name
    pub fn new(name: FunctionName) -> Self {
        Self { function_token: TokenReference::basic_symbol("function "), name, body: FunctionBody::new() }
    }

    /// The `function` token
    pub fn function_token(&self) -> &TokenReference {
        &self.function_token
    }

    /// The body of the function
    pub fn body(&self) -> &FunctionBody {
        &self.body
    }

    /// The name of the function
    pub fn name(&self) -> &FunctionName {
        &self.name
    }

    /// Returns a new FunctionDeclaration with the given `function` token
    pub fn with_function_token(self, function_token: TokenReference) -> Self {
        Self { function_token, ..self }
    }

    /// Returns a new FunctionDeclaration with the given function name
    pub fn with_name(self, name: FunctionName) -> Self {
        Self { name, ..self }
    }

    /// Returns a new FunctionDeclaration with the given function body
    pub fn with_body(self, body: FunctionBody) -> Self {
        Self { body, ..self }
    }
}

/// An option declaration, such as `option x : number = 1`
#[derive(Clone, Debug, Display, PartialEq, Node, Visit, Deserialize, Serialize)]
#[display("{option_token}{}{equal_token}{expr_list}",
    join_type_specifiers(&self.name_list, self.type_specifiers())
)]
pub struct OptionDecl {
    option_token: TokenReference,

    name_list: Punctuated<TokenReference>,

    #[serde(skip_serializing_if = "empty_optional_vector")]
    type_specifiers: Vec<Option<TypeSpecifier>>,

    equal_token: TokenReference,
    expr_list: Punctuated<Expression>,
}

impl OptionDecl {
    /// Get the declarations of this option declaration statement.
    pub fn declarations(&self) -> impl Iterator<Item = (&TokenReference, Option<&TypeSpecifier>, &Expression)> {
        self.name_list
            .iter()
            .zip(&self.type_specifiers)
            .zip(&self.expr_list)
            .map(|((name, type_specifier), expression)| (name, type_specifier.as_ref(), expression))
    }
}

/// A call list item. We need a call list to know how to a job from the...
#[derive(Clone, Debug, PartialEq, Display, Node, Visit, Deserialize, Serialize)]
pub enum CallListItem {
    /// A reference to a variable, e.g. `"init"`
    #[display("{_0}")]
    Variable(TokenReference),

    /// Setting of a parameter for the job, e.g. `pre_line = 900`
    #[display("{parameter}{equal_token}{expression}")]
    SetParameter {
        /// The name of the parameter to set for a given job, must be an argument of said job
        parameter: TokenReference,

        /// The `=` token of the assignation
        equal_token: TokenReference,

        /// The value to set for the given parameter in the given job.
        expression: Expression,
    },
}

impl CallListItem {
    /// Returns the last token, for any variant of this enum.
    pub fn first_token(&self) -> &TokenReference {
        match self {
            CallListItem::SetParameter { parameter: first, .. } | CallListItem::Variable(first) => first,
        }
    }
}

/// The call list for a job, e.g. `{ param = 1, "init" }`
#[derive(Clone, Debug, Display, PartialEq, Node, Visit, Deserialize, Serialize, Default)]
#[display("{}{items}{}", display_option(begin_token), display_option(end_token))]
pub struct CallList {
    begin_token: Option<TokenReference>,
    items: Punctuated<CallListItem>,
    end_token: Option<TokenReference>,
}

impl CallList {
    /// Create a new empty call list
    pub fn new() -> Self {
        Self::default()
    }

    /// Set the `{` token for the call list
    pub fn with_begin_token(self, begin_token: Option<TokenReference>) -> Self {
        Self { begin_token, ..self }
    }

    /// Set the `}` token for the call list
    pub fn with_end_token(self, end_token: Option<TokenReference>) -> Self {
        Self { end_token, ..self }
    }

    /// Set the items in this list
    pub fn with_items(self, items: Punctuated<CallListItem>) -> Self {
        Self { items, ..self }
    }

    /// Get the items out of the call list.
    pub fn items(&self) -> &Punctuated<CallListItem> {
        &self.items
    }
}

/// The main assignement statement, only inside the `main` statement.
///
/// ```vivy
/// "OUTLINED" = utils:outline { "BEFORE", "INIT", "AFTER" }
/// "TAGGED"   = tag:syl_modulo { every = 3, disp = 1, "OUTLINED" }
/// ```
#[derive(Clone, Debug, Display, PartialEq, Node, Visit, Deserialize, Serialize)]
#[display("{variable}{assign_token}{module_name}{dot_token}{job_name}{call_list}")]
pub struct MainAssignement {
    variable: TokenReference,
    assign_token: TokenReference,

    module_name: TokenReference,
    dot_token: TokenReference,
    job_name: TokenReference,

    call_list: CallList,
}

impl MainAssignement {
    /// Create a new assign statement, only inside the `main` statement.
    pub fn new(
        assign_token: TokenReference,
        variable: TokenReference,
        module_name: TokenReference,
        dot_token: TokenReference,
        job_name: TokenReference,
    ) -> Self {
        Self { assign_token, variable, module_name, dot_token, job_name, call_list: Default::default() }
    }

    /// Set the call-list for this invocation of a job.
    pub fn with_call_list(self, call_list: CallList) -> Self {
        Self { call_list, ..self }
    }

    /// Get the called job out of the assignation
    pub fn called_job(&self) -> (&TokenReference, &TokenReference) {
        (&self.module_name, &self.job_name)
    }

    /// Get the called job out of the assignation as identifiers, otherwise panic. If the program
    /// is correct and the parser is Ok, this should never panic
    pub fn called_job_identifiers(&self) -> (&ShortString, &ShortString) {
        let (module, job) = self.called_job();
        match (module.token_type(), job.token_type()) {
            (TokenType::Identifier { identifier: m }, TokenType::Identifier { identifier: j }) => (m, j),
            _ => unreachable!("module and job names should be identifiers"),
        }
    }

    /// Get the call list to know how to call the job.
    pub fn call_list(&self) -> &CallList {
        &self.call_list
    }

    /// Get the items of the call list to know how to call the job.
    pub fn call_list_items(&self) -> impl Iterator<Item = &CallListItem> {
        self.call_list.items().into_iter()
    }

    /// Get the name of the written variable
    pub fn destination(&self) -> &TokenReference {
        &self.variable
    }
}

/// The main statement: `main "INIT" { ... }`
#[derive(Clone, Debug, Display, PartialEq, Node, Visit, Deserialize, Serialize)]
#[display(
    "{main_token}{initial_variable}{begin_token}{}{}{end_token}",
    join_vec(assignements),
    display_option(write_stmt)
)]
pub struct Main {
    main_token: TokenReference,
    initial_variable: TokenReference,
    begin_token: TokenReference,
    assignements: Vec<MainAssignement>,
    write_stmt: Option<Write>,
    end_token: TokenReference,
}

impl Main {
    /// Create a new empty main statement.
    pub fn new(
        main_token: TokenReference,
        initial_variable: TokenReference,
        begin_token: TokenReference,
        end_token: TokenReference,
    ) -> Self {
        Self { main_token, begin_token, end_token, initial_variable, assignements: Vec::new(), write_stmt: None }
    }

    /// Get the `main` token.
    pub fn main_token(&self) -> &TokenReference {
        &self.main_token
    }

    /// Get the `do` token in `main do ... end`.
    pub fn begin_list_token(&self) -> &TokenReference {
        &self.begin_token
    }

    /// Get the `end` token in `main do ... end`.
    pub fn end_list_token(&self) -> &TokenReference {
        &self.end_token
    }

    /// Set the `return` variables to write, the token here
    pub fn with_returns(self, write_stmt: Option<Write>) -> Self {
        Self { write_stmt, ..self }
    }

    /// Set the assignements for the 'main' statement.
    pub fn with_assignements(self, assignements: Vec<MainAssignement>) -> Self {
        Self { assignements, ..self }
    }

    /// Get the assignations, i.e. the compute steps
    pub fn assignements(&self) -> &[MainAssignement] {
        &self.assignements
    }

    /// Get the written variables
    pub fn returns(&self) -> Option<&Write> {
        self.write_stmt.as_ref()
    }

    /// Get the name of the initial variable
    pub fn initial_variable(&self) -> &TokenReference {
        &self.initial_variable
    }
}

/// The write directive, with the variables to write into the final file: `write "INIT"`.
#[derive(Clone, Debug, Display, PartialEq, Node, Visit, Deserialize, Serialize)]
#[display("{write_token}{}{variable_list}{}",
    display_option(self.begin_list_token()),
    display_option(self.end_list_token())
)]
pub struct Write {
    write_token: TokenReference,
    begin_token: Option<TokenReference>,
    variable_list: Punctuated<TokenReference>,
    end_token: Option<TokenReference>,
}

impl Write {
    /// Create a new empty write statement.
    pub fn new(write_token: TokenReference) -> Self {
        Self { write_token, begin_token: None, variable_list: Punctuated::new(), end_token: None }
    }

    /// Get the `write` token for this statement.
    pub fn write_token(&self) -> &TokenReference {
        &self.write_token
    }

    /// Get the list of variables to write
    pub fn variables(&self) -> &Punctuated<TokenReference> {
        &self.variable_list
    }

    /// Get the '{' token if present in the write statement.
    pub fn begin_list_token(&self) -> Option<&TokenReference> {
        self.begin_token.as_ref()
    }

    /// Get the '}' token if present in the write statement.
    pub fn end_list_token(&self) -> Option<&TokenReference> {
        self.end_token.as_ref()
    }

    /// Set the `}` token.
    pub fn with_end_list_token(self, end_token: Option<TokenReference>) -> Self {
        Self { end_token, ..self }
    }

    /// Set the `{` token.
    pub fn with_begin_list_token(self, begin_token: Option<TokenReference>) -> Self {
        Self { begin_token, ..self }
    }

    /// Set the list of variables to write.
    pub fn with_variables(self, variable_list: Punctuated<TokenReference>) -> Self {
        Self { variable_list, ..self }
    }
}

#[doc(hidden)]
#[macro_export]
macro_rules! make_bin_op {
    ($(#[$outer:meta])* { $(
        $operator:ident = $precedence:expr,
    )+ }) => {
        paste::paste! {
            #[derive(Clone, Debug, Display, PartialEq, Eq, Node, Visit, Deserialize, Serialize)]
            #[non_exhaustive]
            $(#[$outer])*
            pub enum BinOp {
                $(#[display("{_0}")] #[allow(missing_docs)] $operator(TokenReference),)+
            }

            impl BinOp {
                /// The precedence of non-unary operator. The larger the number, the higher the precedence.
                /// Shares the same precedence table as unary operators.
                pub fn precedence_of_token(token: &TokenReference) -> Option<u8> {
                    match token.token_type() {
                        TokenType::Symbol { symbol } => match symbol {
                            $(Symbol::$operator => Some($precedence),)+ _ => None,
                        },
                        _ => None
                    }
                }

                /// The token associated with this operator
                pub fn token(&self) -> &TokenReference {
                    match self {
                        $(BinOp::$operator(token) => token,)+
                    }
                }

                pub(crate) fn consume(state: &mut ParserState) -> Option<Self> {
                    match state.current().unwrap().token_type() {
                        TokenType::Symbol { symbol } => match symbol {
                            $(Symbol::$operator => { Some(BinOp::$operator(state.consume().unwrap())) },)+
                            _ => None,
                        },
                        _ => None,
                    }
                }
            }
        }
    };
}

make_bin_op!(
    #[doc = "Operators that require two operands, such as X + Y or X - Y"]
    #[visit(skip_visit_self)]
    {
        Caret = 12,

        // At the 11° position we have the unary operations.

        Percent = 10,
        Slash = 10,
        Star = 10,
        DoubleSlash = 10,

        Minus = 9,
        Plus = 9,

        TwoDots = 8,

        DoubleGreaterThan = 7,
        DoubleLesserThan = 7,

        Ampersand = 6,

        Tilde = 5,

        Pipe = 4,

        GreaterThan = 3,
        GreaterThanEqual = 3,
        LessThan = 3,
        LessThanEqual = 3,
        TildeEqual = 3,
        TwoEqual = 3,

        And = 2,

        Or = 1,
    }
);

impl BinOp {
    /// The precedence of the operator. The larger the number, the higher the precedence.
    /// See more at <http://www.lua.org/manual/5.1/manual.html#2.5.6>
    pub fn precedence(&self) -> u8 {
        BinOp::precedence_of_token(self.token()).expect("invalid token")
    }

    /// Whether the operator is right associative. If not, it is left associative.
    /// See more at <https://www.lua.org/pil/3.5.html>
    pub fn is_right_associative(&self) -> bool {
        matches!(*self, BinOp::Caret(_) | BinOp::TwoDots(_))
    }

    /// Given a token, returns whether it is a right associative binary operator.
    pub fn is_right_associative_token(token: &TokenReference) -> bool {
        matches!(
            token.token_type(),
            TokenType::Symbol { symbol: Symbol::Caret } | TokenType::Symbol { symbol: Symbol::TwoDots }
        )
    }
}

/// Operators that require just one operand, such as #X
#[derive(Clone, Debug, Display, PartialEq, Eq, Node, Visit, Deserialize, Serialize)]
#[allow(missing_docs)]
#[non_exhaustive]
pub enum UnOp {
    Minus(TokenReference),
    Not(TokenReference),
    Hash(TokenReference),
    Tilde(TokenReference),
}

impl UnOp {
    /// The token associated with the operator
    pub fn token(&self) -> &TokenReference {
        match self {
            UnOp::Minus(token) | UnOp::Not(token) | UnOp::Hash(token) => token,
            UnOp::Tilde(token) => token,
        }
    }

    /// The precedence of unary operator. The larger the number, the higher the precedence.
    /// Shares the same precedence table as binary operators.
    pub fn precedence() -> u8 {
        11
    }
}

impl OptionDecl {
    /// Returns a new OptionSet from the given name list
    pub fn new(name_list: Punctuated<TokenReference>) -> Self {
        Self {
            option_token: TokenReference::basic_symbol("option "),
            type_specifiers: Vec::new(),
            name_list,
            equal_token: TokenReference::basic_symbol("= "),
            expr_list: Punctuated::new(),
        }
    }

    /// Get the `option` token
    pub fn option_token(&self) -> &TokenReference {
        &self.option_token
    }

    /// Get the `=` token
    pub fn equal_token(&self) -> &TokenReference {
        &self.equal_token
    }

    /// Get the expressions
    pub fn expressions(&self) -> &Punctuated<Expression> {
        &self.expr_list
    }

    /// Get the names of the options to set
    pub fn names(&self) -> &Punctuated<TokenReference> {
        &self.name_list
    }

    /// Get the type specifiers
    pub fn type_specifiers(&self) -> impl Iterator<Item = Option<&TypeSpecifier>> {
        self.type_specifiers.iter().map(Option::as_ref)
    }

    /// Set the type specifiers
    pub fn with_type_specifiers(self, type_specifiers: Vec<Option<TypeSpecifier>>) -> Self {
        Self { type_specifiers, ..self }
    }

    /// Set the `option` token
    pub fn with_option_token(self, option_token: TokenReference) -> Self {
        Self { option_token, ..self }
    }

    /// Set the `=` token
    pub fn with_equal_token(self, equal_token: TokenReference) -> Self {
        Self { equal_token, ..self }
    }

    /// Set the expressions
    pub fn with_expressions(self, expr_list: Punctuated<Expression>) -> Self {
        Self { expr_list, ..self }
    }
}

/// An error that occurs when creating the AST.
#[derive(Clone, Debug, PartialEq, Eq, Display, Deserialize, Serialize)]
#[display(
    "unexpected token `{token}`. (starting from line {}, character {} and ending on line {}, character {})\nadditional information: {additional}",
    self.range().0.line(),
    self.range().0.character(),
    self.range().1.line(),
    self.range().1.character(),
)]
pub struct AstError {
    /// The token that caused the error
    token: Token,

    /// Any additional information that could be provided for debugging
    additional: Cow<'static, str>,

    /// If set, this is the complete range of the error
    #[serde(skip_serializing_if = "Option::is_none")]
    range: Option<(Position, Position)>,
}

impl AstError {
    /// Returns the token that caused the error
    pub fn token(&self) -> &Token {
        &self.token
    }

    /// Returns a human readable error message
    pub fn error_message(&self) -> Cow<'static, str> {
        self.additional.clone()
    }

    /// Returns the range of the error
    pub fn range(&self) -> (Position, Position) {
        self.range
            .or_else(|| Some((self.token.start_position(), self.token.end_position())))
            .unwrap()
    }

    /// Create an error from its parts, not really something we want to expose, but errors may come
    /// from the passes and not the parse stage...
    pub fn from_parts(token_reference: TokenReference, error: impl Into<Cow<'static, str>>) -> Self {
        Self { token: token_reference.token().clone(), additional: error.into(), range: token_reference.range() }
    }
}

impl std::error::Error for AstError {}

/// An abstract syntax tree, contains all the nodes used in the code
#[derive(Clone, Debug, Display, Deserialize, Serialize)]
#[display("{nodes}{eof}")]
pub struct Ast {
    pub(crate) nodes: Block,
    pub(crate) eof: TokenReference,
}

impl Ast {
    /// Returns a new Ast with the given nodes
    pub fn with_nodes(self, nodes: Block) -> Self {
        Self { nodes, ..self }
    }

    /// Returns a new Ast with the given EOF token
    pub fn with_eof(self, eof: TokenReference) -> Self {
        Self { eof, ..self }
    }

    /// The entire code of the function
    pub fn nodes(&self) -> &Block {
        &self.nodes
    }

    /// The entire code of the function, but mutable
    pub fn nodes_mut(&mut self) -> &mut Block {
        &mut self.nodes
    }

    /// The EOF token at the end of every Ast
    pub fn eof(&self) -> &TokenReference {
        &self.eof
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use crate::{prelude::parser::parse_lua_tree, visitors::VisitorMut};

    #[test]
    fn test_with_eof_safety() {
        print!("{}", {
            let ast = parse_lua_tree("local foo = 1").unwrap();
            let eof = ast.eof().clone();
            ast.with_eof(eof)
        });
    }

    #[test]
    fn test_with_nodes_safety() {
        print!("{}", {
            let ast = parse_lua_tree("local foo = 1").unwrap();
            let nodes = ast.nodes().clone();
            ast.with_nodes(nodes)
        });
    }

    #[test]
    fn test_with_visitor_safety() {
        print!("{}", {
            let ast = parse_lua_tree("local foo = 1").unwrap();
            struct SyntaxRewriter;
            impl VisitorMut for SyntaxRewriter {
                fn visit_token(&mut self, token: Token) -> Token {
                    token
                }
            }
            SyntaxRewriter.visit_ast(ast)
        });
    }

    // Tests AST nodes with new methods that call unwrap
    #[test]
    fn test_new_validity() {
        let token: TokenReference =
            TokenReference::new(Vec::new(), Token::new(TokenType::Identifier { identifier: "foo".into() }), Vec::new());

        let expression = Expression::Var(Var::Name(token.clone()));

        Assignment::new(Punctuated::new(), Punctuated::new());
        Do::new();
        ElseIf::new(expression.clone());
        FunctionBody::new();
        FunctionCall::new(Prefix::Name(token.clone()));
        FunctionDeclaration::new(FunctionName::new(Punctuated::new()));
        GenericFor::new(Punctuated::new(), Punctuated::new());
        If::new(expression.clone());
        LocalAssignment::new(Punctuated::new());
        LocalFunction::new(token.clone());
        MethodCall::new(
            token.clone(),
            FunctionArgs::Parentheses {
                arguments: Punctuated::new(),
                parentheses: ContainedSpan::new(token.clone(), token.clone()),
            },
        );
        NumericFor::new(token, expression.clone(), expression.clone());
        Repeat::new(expression.clone());
        Return::new();
        TableConstructor::new();
        While::new(expression);
    }

    #[test]
    fn test_local_assignment_print() {
        let block = Block::new().with_stmts(vec![(
            Stmt::LocalAssignment(
                LocalAssignment::new(
                    std::iter::once(Pair::End(TokenReference::new(
                        vec![],
                        Token::new(TokenType::Identifier { identifier: "variable".into() }),
                        vec![],
                    )))
                    .collect(),
                )
                .with_equal_token(Some(TokenReference::symbol(" = ").unwrap()))
                .with_expressions(
                    std::iter::once(Pair::End(Expression::Number(TokenReference::new(
                        vec![],
                        Token::new(TokenType::Number { text: "1".into() }),
                        vec![],
                    ))))
                    .collect(),
                ),
            ),
            None,
        )]);

        let ast = parse_lua_tree("").unwrap().with_nodes(block);
        assert_eq!(format!("{ast}"), "local variable = 1");
    }
}

/// Any type, such as `string`, `boolean?`, etc.
#[derive(Clone, Debug, Display, IsVariant, Node, Deserialize, Serialize)]
#[non_exhaustive]
pub enum TypeInfo {
    /// A shorthand type annotating the structure of an array: { number }
    #[display("{}{type_info}{}", braces.tokens().0, braces.tokens().1)]
    Array {
        /// The braces (`{}`) containing the type info.
        braces: ContainedSpan,

        /// The type info for the values in the Array
        type_info: Box<TypeInfo>,
    },

    /// A standalone type, such as `string` or `Foo`.
    #[display("{_0}")]
    Basic(TokenReference),

    /// An imported module, such as `string` after an `import "string"`, or `self` outside of
    /// method definitions.
    #[display("{_0}")]
    Module(TokenReference),

    /// A callback type, such as `(string, number) => boolean`.
    #[display("{}{arguments}{}{arrow}{return_type}", parentheses.tokens().0, parentheses.tokens().1)]
    Callback {
        /// The parentheses for the arguments.
        parentheses: ContainedSpan,

        /// The argument types: `(string, number)`.
        arguments: Punctuated<TypeArgument>,

        /// The "thin arrow" (`->`) in between the arguments and the return type.
        arrow: TokenReference,

        /// The return type: `boolean`.
        return_type: Box<TypeInfo>,
    },

    /// An optional type, such as `string?`.
    #[display("{base}{question_mark}")]
    Optional {
        /// The type that is optional: `string`.
        base: Box<TypeInfo>,

        /// The question mark: `?`.
        question_mark: TokenReference,
    },

    /// A type annotating the structure of a table: { foo: number, bar: string }
    #[display("{}{fields}{}", braces.tokens().0, braces.tokens().1)]
    Table {
        /// The braces (`{}`) containing the fields.
        braces: ContainedSpan,

        /// The fields: `foo: number, bar: string`.
        fields: Punctuated<TypeField>,
    },

    /// A tuple expression: `(string, number)`.
    #[display("{}{types}{}", parentheses.tokens().0, parentheses.tokens().1)]
    Tuple {
        /// The parentheses used to contain the types
        parentheses: ContainedSpan,

        /// The types: `(string, number)`.
        types: Punctuated<TypeInfo>,
    },
}

macro_rules! basic_type_info {
    ($type: ident, $($types: ident),+ $(,)?) => {
        basic_type_info! { $type }
        basic_type_info! { $($types),+ }
    };

    ($type: ident) => {
        #[doc = concat!("Get the type for '", stringify!($type), "'")]
        pub fn $type() -> &'static Self {
            const _: () = assert!(stringify!($type).len() <= 23, "the type name can't be inline");
            static TYPE_INFO: TypeInfo = TypeInfo::Basic(TokenReference::new(
                vec![],
                Token::new(TokenType::Identifier { identifier: ShortString::new_inline(stringify!($type)) }),
                vec![])
            );
            &TYPE_INFO
        }
    };
}

impl TypeInfo {
    basic_type_info! {
        string, number, char, nil, table, aux, any,
        line, lines, syllabe, syllabes,
    }
}

impl Default for TypeInfo {
    fn default() -> Self {
        TypeInfo::nil().clone()
    }
}

impl DefaultRef for TypeInfo {
    fn default_ref() -> &'static Self {
        TypeInfo::nil()
    }
}

impl PartialEq for TypeInfo {
    fn eq(&self, other: &Self) -> bool {
        match (self, other) {
            (Self::Basic(l0), Self::Basic(r0)) => l0.token_type() == r0.token_type(),

            (Self::Array { type_info: lty, .. }, Self::Array { type_info: rty, .. })
            | (Self::Optional { base: lty, .. }, Self::Optional { base: rty, .. }) => lty == rty,

            (Self::Table { fields: l, .. }, Self::Table { fields: r, .. }) => {
                l.iter().zip(r.iter()).all(|(l, r)| PartialEq::eq(l, r))
            }

            (Self::Tuple { types: l, .. }, Self::Tuple { types: r, .. }) => {
                l.iter().zip(r.iter()).all(|(l, r)| PartialEq::eq(l, r))
            }

            (Self::Callback { .. }, Self::Callback { .. }) => unimplemented!("need to code the callback type check"),

            _ => false,
        }
    }
}

/// A type field used within table types.
/// The `foo: number` in `{ foo: number }`.
#[derive(Clone, Debug, Display, PartialEq, Node, Visit, Deserialize, Serialize)]
#[display("{key}{colon}{value}")]
pub struct TypeField {
    pub(crate) key: TypeFieldKey,
    pub(crate) colon: TokenReference,
    pub(crate) value: TypeInfo,
}

impl TypeField {
    /// Creates a new TypeField from the given key and value
    pub fn new(key: TypeFieldKey, value: TypeInfo) -> Self {
        Self { key, colon: TokenReference::symbol(": ").unwrap(), value }
    }

    /// The key of the field, `foo` in `foo: number`.
    pub fn key(&self) -> &TypeFieldKey {
        &self.key
    }

    /// The colon in between the key name and the value type.
    pub fn colon_token(&self) -> &TokenReference {
        &self.colon
    }

    /// The type for the field, `number` in `foo: number`.
    pub fn value(&self) -> &TypeInfo {
        &self.value
    }

    /// Returns a new TypeField with the given key
    pub fn with_key(self, key: TypeFieldKey) -> Self {
        Self { key, ..self }
    }

    /// Returns a new TypeField with the `:` token
    pub fn with_colon_token(self, colon_token: TokenReference) -> Self {
        Self { colon: colon_token, ..self }
    }

    /// Returns a new TypeField with the `:` token
    pub fn with_value(self, value: TypeInfo) -> Self {
        Self { value, ..self }
    }
}

/// A key in a [`TypeField`]. Can either be a name or an index signature.
#[derive(Clone, Debug, Display, PartialEq, Node, Deserialize, Serialize)]
#[non_exhaustive]
pub enum TypeFieldKey {
    /// A name, such as `foo`.
    #[display("{_0}")]
    Name(TokenReference),

    /// An index signature, such as `[number]`.
    #[display("{}{inner}{}", brackets.tokens().0, brackets.tokens().1)]
    IndexSignature {
        /// The brackets (`[]`) used to contain the type.
        brackets: ContainedSpan,

        /// The type for the index signature, `number` in `[number]`.
        inner: TypeInfo,
    },
}

/// A type assertion using `::`, such as `:: number`.
#[derive(Clone, Debug, Display, PartialEq, Node, Visit, Deserialize, Serialize)]
#[display("{assertion_op}{cast_to}")]
pub struct TypeAssertion {
    pub(crate) assertion_op: TokenReference,
    pub(crate) cast_to: TypeInfo,
}

impl TypeAssertion {
    /// Creates a new TypeAssertion from the given cast to TypeInfo
    pub fn new(cast_to: TypeInfo) -> Self {
        Self { assertion_op: TokenReference::symbol("::").unwrap(), cast_to }
    }

    /// The token `::`.
    pub fn assertion_op(&self) -> &TokenReference {
        &self.assertion_op
    }

    /// The type to cast the expression into, `number` in `:: number`.
    pub fn cast_to(&self) -> &TypeInfo {
        &self.cast_to
    }

    /// Returns a new TypeAssertion with the given `::` token
    pub fn with_assertion_op(self, assertion_op: TokenReference) -> Self {
        Self { assertion_op, ..self }
    }

    /// Returns a new TypeAssertion with the given TypeInfo to cast to
    pub fn with_cast_to(self, cast_to: TypeInfo) -> Self {
        Self { cast_to, ..self }
    }
}

/// A type declaration, such as `type Meters = number`
#[derive(Clone, Debug, Display, PartialEq, Node, Visit, Deserialize, Serialize)]
#[display("{type_token}{base}{equal_token}{declare_as}")]
pub struct TypeDeclaration {
    pub(crate) type_token: TokenReference,
    pub(crate) base: TokenReference,
    pub(crate) equal_token: TokenReference,
    pub(crate) declare_as: TypeInfo,
}

impl TypeDeclaration {
    /// Creates a new TypeDeclaration from the given type name and type declaration
    pub fn new(type_name: TokenReference, type_definition: TypeInfo) -> Self {
        Self {
            type_token: TokenReference::new(
                Vec::new(),
                Token::new(TokenType::Identifier { identifier: "type".into() }),
                vec![Token::new(TokenType::spaces(1))],
            ),
            base: type_name,
            equal_token: TokenReference::symbol(" = ").unwrap(),
            declare_as: type_definition,
        }
    }

    /// The token `type`.
    pub fn type_token(&self) -> &TokenReference {
        &self.type_token
    }

    /// The name of the type, `Meters` in `type Meters = number`.
    pub fn type_name(&self) -> &TokenReference {
        &self.base
    }

    /// The `=` token in between the type name and the definition.
    pub fn equal_token(&self) -> &TokenReference {
        &self.equal_token
    }

    /// The definition of the type, `number` in `type Meters = number`.
    pub fn type_definition(&self) -> &TypeInfo {
        &self.declare_as
    }

    /// Returns a new TypeDeclaration with the given `type` token
    pub fn with_type_token(self, type_token: TokenReference) -> Self {
        Self { type_token, ..self }
    }

    /// Returns a new TypeDeclaration with the given type name
    pub fn with_type_name(self, type_name: TokenReference) -> Self {
        Self { base: type_name, ..self }
    }

    /// Returns a new TypeDeclaration with the given generics of the type
    pub fn with_equal_token(self, equal_token: TokenReference) -> Self {
        Self { equal_token, ..self }
    }

    /// Returns a new TypeDeclaration with the given generics of the type
    pub fn with_type_definition(self, type_definition: TypeInfo) -> Self {
        Self { declare_as: type_definition, ..self }
    }
}

/// A type specifier, the `: number` in `local foo: number`
#[derive(Clone, Debug, Display, PartialEq, Node, Visit, Deserialize, Serialize)]
#[display("{punctuation}{type_info}")]
pub struct TypeSpecifier {
    pub(crate) punctuation: TokenReference,
    pub(crate) type_info: TypeInfo,
}

impl TypeSpecifier {
    /// Creates a new TypeSpecifier with the given type info
    pub fn new(type_info: TypeInfo) -> Self {
        Self { punctuation: TokenReference::symbol(": ").unwrap(), type_info }
    }

    /// The punctuation being used.
    /// `:` for `local foo: number`.
    pub fn punctuation(&self) -> &TokenReference {
        &self.punctuation
    }

    /// The type being specified: `number` in `local foo: number`.
    pub fn type_info(&self) -> &TypeInfo {
        &self.type_info
    }

    /// Returns a new TypeSpecifier with the given punctuation
    pub fn with_punctuation(self, punctuation: TokenReference) -> Self {
        Self { punctuation, ..self }
    }

    /// Returns a new TypeSpecifier with the given type being specified
    pub fn with_type_info(self, type_info: TypeInfo) -> Self {
        Self { type_info, ..self }
    }
}

/// A type argument specified in a callback type, the `count: number` in `(count: number) -> ()`
#[derive(Clone, Debug, PartialEq, Node, Visit, Deserialize, Serialize)]
pub struct TypeArgument {
    pub(crate) name: Option<(TokenReference, TokenReference)>,
    pub(crate) type_info: TypeInfo,
}

impl TypeArgument {
    /// Creates a new TypeArgument with the given type info
    pub fn new(type_info: TypeInfo) -> Self {
        Self { name: None, type_info }
    }

    /// The name of the argument split into identifier and punctuation: `count:` in `count: number`.
    pub fn name(&self) -> Option<&(TokenReference, TokenReference)> {
        self.name.as_ref()
    }

    /// The type info for the argument: `number` in `count: number`.
    pub fn type_info(&self) -> &TypeInfo {
        &self.type_info
    }

    /// The type info for the argument: `number` in `count: number`. Consumes the [TypeArgument]
    /// structure.
    pub fn into_type_info(self) -> TypeInfo {
        self.type_info
    }

    /// Returns a new TypeArgument with the given punctuation
    pub fn with_name(self, name: Option<(TokenReference, TokenReference)>) -> Self {
        Self { name, ..self }
    }

    /// Returns a new TypeArgument with the given type info
    pub fn with_type_info(self, type_info: TypeInfo) -> Self {
        Self { type_info, ..self }
    }
}

impl fmt::Display for TypeArgument {
    fn fmt(&self, formatter: &mut fmt::Formatter) -> fmt::Result {
        if let Some((identifier, punctuation)) = self.name() {
            write!(formatter, "{}{}{}", identifier, punctuation, self.type_info)
        } else {
            write!(formatter, "{}", self.type_info)
        }
    }
}

/// Compound operators, such as X += Y or X -= Y
#[derive(Clone, Debug, Display, PartialEq, Eq, Node, Visit, Deserialize, Serialize)]
#[non_exhaustive]
#[allow(missing_docs)]
pub enum CompoundOp {
    PlusEqual(TokenReference),
    MinusEqual(TokenReference),
    StarEqual(TokenReference),
    SlashEqual(TokenReference),
    DoubleSlashEqual(TokenReference),
    PercentEqual(TokenReference),
    CaretEqual(TokenReference),
    TwoDotsEqual(TokenReference),
}

impl CompoundOp {
    /// The token associated with the operator
    pub fn token(&self) -> &TokenReference {
        match self {
            Self::PlusEqual(token)
            | Self::MinusEqual(token)
            | Self::StarEqual(token)
            | Self::SlashEqual(token)
            | Self::DoubleSlashEqual(token)
            | Self::PercentEqual(token)
            | Self::CaretEqual(token)
            | Self::TwoDotsEqual(token) => token,
        }
    }

    pub(crate) fn from_token(token: TokenReference) -> Self {
        if token.is_symbol(Symbol::PlusEqual) {
            Self::PlusEqual(token)
        } else if token.is_symbol(Symbol::MinusEqual) {
            Self::MinusEqual(token)
        } else if token.is_symbol(Symbol::StarEqual) {
            Self::StarEqual(token)
        } else if token.is_symbol(Symbol::SlashEqual) {
            Self::SlashEqual(token)
        } else if token.is_symbol(Symbol::DoubleSlashEqual) {
            Self::DoubleSlashEqual(token)
        } else if token.is_symbol(Symbol::PercentEqual) {
            Self::PercentEqual(token)
        } else if token.is_symbol(Symbol::CaretEqual) {
            Self::CaretEqual(token)
        } else if token.is_symbol(Symbol::TwoDotsEqual) {
            Self::TwoDotsEqual(token)
        } else {
            unreachable!("converting an unknown token into a compound operator")
        }
    }

    /// The operator must be applied to numerical values.
    pub fn is_numerical_op(&self) -> bool {
        matches!(
            self,
            Self::PlusEqual(_)
                | Self::MinusEqual(_)
                | Self::StarEqual(_)
                | Self::SlashEqual(_)
                | Self::DoubleSlashEqual(_)
                | Self::PercentEqual(_)
                | Self::CaretEqual(_)
        )
    }
}

/// A Compound Assignment statement, such as `x += 1` or `x -= 1`
#[derive(Clone, Debug, Display, PartialEq, Node, Visit, Deserialize, Serialize)]
#[display("{lhs}{compound_operator}{rhs}")]
pub struct CompoundAssignment {
    pub(crate) lhs: Var,
    pub(crate) compound_operator: CompoundOp,
    pub(crate) rhs: Expression,
}

impl CompoundAssignment {
    /// Creates a new CompoundAssignment from the left and right hand side
    pub fn new(lhs: Var, compound_operator: CompoundOp, rhs: Expression) -> Self {
        Self { lhs, compound_operator, rhs }
    }

    /// The variable assigned to, the `x` part of `x += 1`
    pub fn lhs(&self) -> &Var {
        &self.lhs
    }

    /// The operator used, the `+=` part of `x += 1`
    pub fn compound_operator(&self) -> &CompoundOp {
        &self.compound_operator
    }

    /// The value being assigned, the `1` part of `x += 1`
    pub fn rhs(&self) -> &Expression {
        &self.rhs
    }

    /// Returns a new CompoundAssignment with the given variable being assigned to
    pub fn with_lhs(self, lhs: Var) -> Self {
        Self { lhs, ..self }
    }

    /// Returns a new CompoundAssignment with the given operator used
    pub fn with_compound_operator(self, compound_operator: CompoundOp) -> Self {
        Self { compound_operator, ..self }
    }

    /// Returns a new CompoundAssignment with the given value being assigned
    pub fn with_rhs(self, rhs: Expression) -> Self {
        Self { rhs, ..self }
    }
}

/// An if statement
#[derive(Clone, Debug, Display, PartialEq, Node, Visit, Deserialize, Serialize)]
#[display("{if_token}{condition}{then_token}{if_expression}{}{else_token}{else_expression}",
    display_option(else_if_expressions.as_ref().map(join_vec)),
)]
pub struct IfExpression {
    pub(crate) if_token: TokenReference,
    pub(crate) condition: Box<Expression>,
    pub(crate) then_token: TokenReference,
    pub(crate) if_expression: Box<Expression>,
    pub(crate) else_if_expressions: Option<Vec<ElseIfExpression>>,
    pub(crate) else_token: TokenReference,
    pub(crate) else_expression: Box<Expression>,
}

impl IfExpression {
    /// Creates a new If from the given condition
    pub fn new(condition: Expression, if_expression: Expression, else_expression: Expression) -> Self {
        Self {
            if_token: TokenReference::symbol("if ").unwrap(),
            condition: Box::new(condition),
            then_token: TokenReference::symbol(" then").unwrap(),
            if_expression: Box::new(if_expression),
            else_if_expressions: None,
            else_token: TokenReference::symbol(" else ").unwrap(),
            else_expression: Box::new(else_expression),
        }
    }

    /// The `if` token
    pub fn if_token(&self) -> &TokenReference {
        &self.if_token
    }

    /// The condition of the if expression, `condition` in `if condition then`
    pub fn condition(&self) -> &Expression {
        &self.condition
    }

    /// The `then` token
    pub fn then_token(&self) -> &TokenReference {
        &self.then_token
    }

    /// The expression evaluated if the initial if condition holds
    pub fn if_expression(&self) -> &Expression {
        &self.if_expression
    }

    /// The `else` token
    pub fn else_token(&self) -> &TokenReference {
        &self.else_token
    }

    /// If there are `elseif` conditions, returns a vector of them
    pub fn else_if_expressions(&self) -> impl Iterator<Item = &ElseIfExpression> {
        self.else_if_expressions.as_ref().into_iter().flatten()
    }

    /// The else expression if all other conditions do not hold
    pub fn else_expression(&self) -> &Expression {
        &self.else_expression
    }

    /// Returns a new IfExpression with the given `if` token
    pub fn with_if_token(self, if_token: TokenReference) -> Self {
        Self { if_token, ..self }
    }

    /// Returns a new IfExpression with the given condition
    pub fn with_condition(self, condition: Expression) -> Self {
        Self { condition: Box::new(condition), ..self }
    }

    /// Returns a new IfExpression with the given `then` token
    pub fn with_then_token(self, then_token: TokenReference) -> Self {
        Self { then_token, ..self }
    }

    /// Returns a new IfExpression with the given if expression
    pub fn with_if_expression(self, if_expression: Expression) -> Self {
        Self { if_expression: Box::new(if_expression), ..self }
    }

    /// Returns a new If with the given list of `elseif` expressions
    pub fn with_else_if(self, else_if_expressions: Option<Vec<ElseIfExpression>>) -> Self {
        Self { else_if_expressions, ..self }
    }

    /// Returns a new IfExpression with the given `else` token
    pub fn with_else_token(self, else_token: TokenReference) -> Self {
        Self { else_token, ..self }
    }

    /// Returns a new IfExpression with the given `else` expression
    pub fn with_else(self, else_expression: Expression) -> Self {
        Self { else_expression: Box::new(else_expression), ..self }
    }
}

/// An elseif expression in a bigger [`IfExpression`] expression
#[derive(Clone, Debug, Display, PartialEq, Node, Visit, Deserialize, Serialize)]
#[display("{else_if_token}{condition}{then_token}{expression}")]
pub struct ElseIfExpression {
    pub(crate) else_if_token: TokenReference,
    pub(crate) condition: Expression,
    pub(crate) then_token: TokenReference,
    pub(crate) expression: Expression,
}

impl ElseIfExpression {
    /// Creates a new ElseIf from the given condition
    pub fn new(condition: Expression, expression: Expression) -> Self {
        Self {
            else_if_token: TokenReference::symbol(" elseif ").unwrap(),
            condition,
            then_token: TokenReference::symbol(" then ").unwrap(),
            expression,
        }
    }

    /// The `elseif` token
    pub fn else_if_token(&self) -> &TokenReference {
        &self.else_if_token
    }

    /// The condition of the `elseif`, `condition` in `elseif condition then`
    pub fn condition(&self) -> &Expression {
        &self.condition
    }

    /// The `then` token
    pub fn then_token(&self) -> &TokenReference {
        &self.then_token
    }

    /// The evaluated expression of the `elseif` when condition is true
    pub fn expression(&self) -> &Expression {
        &self.expression
    }

    /// Returns a new ElseIfExpression with the given `elseif` token
    pub fn with_else_if_token(self, else_if_token: TokenReference) -> Self {
        Self { else_if_token, ..self }
    }

    /// Returns a new ElseIfExpression with the given condition
    pub fn with_condition(self, condition: Expression) -> Self {
        Self { condition, ..self }
    }

    /// Returns a new ElseIfExpression with the given `then` token
    pub fn with_then_token(self, then_token: TokenReference) -> Self {
        Self { then_token, ..self }
    }

    /// Returns a new ElseIfExpression with the given expression
    pub fn with_block(self, expression: Expression) -> Self {
        Self { expression, ..self }
    }
}
