use crate::{
    ast::{
        parsers::{parse_block, ParserResult, ParserState},
        Ast,
    },
    tokenizer::{Lexer, LexerResult, TokenKind},
};

/// A produced [`Ast`](crate::ast::Ast), along with any errors found during parsing.
/// This Ast may not be exactly the same as the input code, as reconstruction may have occurred.
/// For more information, read the documentation for [`parse_fallible`](crate::parse_fallible).
pub struct AstResult {
    ast: Ast,
    errors: Vec<crate::Error>,
}

impl AstResult {
    /// Returns a reference to the [`Ast`](crate::ast::Ast) that was parsed.
    /// If there were any errors, this will not be exactly the same,
    /// as reconstruction will have occurred.
    /// For more information, read the documentation for [`parse_fallible`](crate::parse_fallible).
    pub fn ast(&self) -> &Ast {
        &self.ast
    }

    /// Consumes the [`Ast`](crate::ast::Ast) that was parsed.
    /// If there were any errors, this will not be exactly the same,
    /// as reconstruction will have occurred.
    /// For more information, read the documentation for [`parse_fallible`](crate::parse_fallible).
    pub fn into_ast(self) -> Ast {
        self.ast
    }

    /// Returns all errors that occurred during parsing.
    pub fn errors(&self) -> &[crate::Error] {
        &self.errors
    }

    /// Parse the AST, but can fail...
    pub fn parse_fallible(code: &str) -> Self {
        use crate::{ast::AstError, Error};
        const UNEXPECTED_TOKEN_ERROR: &str = "unexpected token, this needs to be a statement";

        let lexer = Lexer::new(code);
        let mut parser_state = ParserState::new(lexer);
        let mut block = parse_block(&mut parser_state).unwrap_or_defaut();

        loop {
            match parser_state.lexer.current() {
                Some(LexerResult::Ok(token)) if token.token_kind() == TokenKind::Eof => break,
                Some(LexerResult::Ok(_) | LexerResult::Recovered(_, _)) => {
                    let ParserResult::Value(new_block) = parse_block(&mut parser_state) else {
                        continue;
                    };
                    if new_block.stmts.is_empty() {
                        match parser_state.current() {
                            Ok(token) if token.token_kind() == TokenKind::Eof => break,
                            _ => {}
                        }
                        match parser_state.consume() {
                            ParserResult::Value(token) => match parser_state.errors.last() {
                                Some(Error::AstError(AstError { additional, .. }))
                                    if additional == UNEXPECTED_TOKEN_ERROR =>
                                {
                                    continue
                                }
                                Some(Error::AstError(AstError { .. })) => {
                                    parser_state.token_error(token, UNEXPECTED_TOKEN_ERROR)
                                }
                                _ => parser_state.token_error(token, UNEXPECTED_TOKEN_ERROR),
                            },
                            ParserResult::LexerMoved => {}
                            ParserResult::NotFound => unreachable!(),
                        }
                        continue;
                    }
                    block.merge_blocks(new_block);
                }
                Some(LexerResult::Fatal(_)) => parser_state.errors.extend(
                    Option::unwrap(parser_state.lexer.consume())
                        .unwrap_errors()
                        .into_iter()
                        .map(Error::TokenizerError),
                ),
                None => break,
            }
        }

        let eof = match parser_state.lexer.consume().unwrap() {
            LexerResult::Ok(token) => token,
            LexerResult::Recovered(token, errors) => {
                parser_state
                    .errors
                    .extend(errors.into_iter().map(Error::TokenizerError));
                token
            }
            LexerResult::Fatal(error) => unreachable!("error: {error:?}"),
        };

        debug_assert_eq!(eof.token_kind(), TokenKind::Eof);
        Self { ast: Ast { nodes: block, eof }, errors: parser_state.errors }
    }

    /// Consumes this AstResult, returning the [`Ast`](crate::ast::Ast) that was parsed.
    pub fn into_result(self) -> Result<Ast, Vec<crate::Error>> {
        self.into()
    }
}

impl From<AstResult> for Result<Ast, Vec<crate::Error>> {
    fn from(ast_result: AstResult) -> Self {
        if ast_result.errors.is_empty() {
            Ok(ast_result.ast)
        } else {
            Err(ast_result.errors)
        }
    }
}
