use crate::{
    ast::AstResult,
    node::Node,
    tests::common::run_test_folder,
    tokenizer::{self, Token, TokenReference},
};
use insta::assert_yaml_snapshot;
use std::{fmt, fs, path::Path};

#[derive(PartialEq, Eq)]
struct PrettyString<'a>(pub &'a str);

// Make diff to display string as multi-line string
impl<'a> fmt::Debug for PrettyString<'a> {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        f.write_str(self.0)
    }
}

fn unpack_token_reference(token: &TokenReference) -> Vec<Token> {
    token
        .leading_trivia()
        .chain(std::iter::once(token.token()))
        .chain(token.trailing_trivia())
        .cloned()
        .collect()
}

fn test_option_case(path: &Path) {
    let source = fs::read_to_string(path.join("source.ini")).expect("couldn't read source.ini");
    let tokens = tokenizer::Lexer::new(&source).collect().unwrap();
    assert_yaml_snapshot!("tokens", tokens);

    let options = crate::ast::OptionTableResult::parse_fallible(&source)
        .into_result()
        .unwrap_or_else(|error| panic!("couldn't make option table for {path:?} - {error:#?}"));

    assert_yaml_snapshot!("options", options);
}

fn test_pass_case(path: &Path) {
    let source = fs::read_to_string(path.join("source.lua")).expect("couldn't read source.lua");
    let tokens = tokenizer::Lexer::new(&source).collect().unwrap();
    assert_yaml_snapshot!("tokens", tokens);

    let ast = AstResult::parse_fallible(&source)
        .into_result()
        .unwrap_or_else(|error| panic!("couldn't make ast for {path:?} - {error:#?}"));

    let old_positions: Vec<_> = ast.tokens().flat_map(unpack_token_reference).collect();

    assert_yaml_snapshot!("ast", ast.nodes());
    assert_eq!(PrettyString(&format!("{ast}")), PrettyString(&source));

    let ast = ast.update_positions();
    assert_eq!(old_positions, ast.tokens().flat_map(unpack_token_reference).collect::<Vec<_>>(),);
}

#[test]
fn test_roblox_pass_cases() {
    run_test_folder("./src/tests/roblox_cases/pass", test_pass_case);
}

#[test]
fn test_lua52_pass_cases() {
    run_test_folder("./src/tests/lua52_cases/pass", test_pass_case);
}

#[test]
fn test_lua53_pass_cases() {
    run_test_folder("./src/tests/lua53_cases/pass", test_pass_case);
}

#[test]
fn test_lua54_pass_cases() {
    run_test_folder("./src/tests/lua54_cases/pass", test_pass_case);
}

#[test]
fn test_vivy_pass_cases() {
    run_test_folder("./src/tests/vivy_cases/pass", test_pass_case);
}

#[test]
fn test_option_pass_cases() {
    run_test_folder("./src/tests/option_cases/pass", test_option_case);
}
