//! Utility macros used in the Vivy Script Runtime.

use anyhow::{anyhow, bail};
use proc_macro2::{Ident, TokenStream};
use quote::quote;
use syn::{
    token::{Pub, Unsafe},
    Generics, Visibility,
};

pub(crate) type MethodInfo = (String, Option<Role>);

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub(crate) enum Role {
    Clone,
    Eq,
    Drop,
    Len,
}

fn handle_function(name: &str, func: &mut syn::ItemFn) -> anyhow::Result<MethodInfo> {
    func.vis = Visibility::Public(Pub::default());
    func.sig.asyncness = None;
    func.sig.unsafety = Some(Unsafe::default());
    func.sig.constness = None;
    func.sig.variadic = None;
    func.sig.generics = Generics::default();

    let func_name = func.sig.ident.to_string().to_lowercase();
    let role = Role::try_from(func_name.as_str()).ok();
    let func_name = format!("{name}_{}", func_name);
    func.sig.ident = Ident::new(&func_name, func.sig.ident.span());

    Ok((func_name, role))
}

pub(crate) fn impl_vvrt_bridge(name: String, mut module: syn::ItemMod) -> anyhow::Result<TokenStream> {
    let (module, funcs) = if let Some((_, items)) = module.content.as_mut() {
        let funcs = items.iter_mut().map(|item| match item {
            syn::Item::Fn(item) => handle_function(&name, item).map(Some),
            _ => Ok(None),
        });
        let funcs = funcs
            .collect::<Result<Vec<_>, _>>()?
            .into_iter()
            .flatten()
            .collect::<Vec<_>>();
        items.push(syn::parse2(quote! { use super::*; })?);
        (module, funcs)
    } else {
        bail!("expected a module")
    };

    let module_name = &module.ident;
    let has_clone = funcs.iter().any(|(_, role, ..)| {
        role.as_ref()
            .map(|role| matches!(role, Role::Clone))
            .unwrap_or_default()
    });
    let has_drop = funcs
        .iter()
        .any(|(_, role, ..)| role.as_ref().map(|role| matches!(role, Role::Drop)).unwrap_or_default());
    if has_clone ^ has_drop {
        panic!("you can't have a drop implementation without the clone one, or the other way around")
    }

    let name = Ident::new(&name, module.ident.span());
    Ok(quote! {
        #module
        pub(super) use self::#module_name::*;
        impl LLVMExported for #name {
            unsafe fn llvm_type(c: LLVMContextRef) -> LLVMTypeRef {
                #[allow(dead_code)]
                static SIZE_IS_VALID: () = {
                    assert!(std::mem::size_of::<#name>() == 8);
                    assert!(std::mem::align_of::<#name>() == 8);
                };
                LLVMPointerTypeInContext(c, 0)
            }
        }
    })
}

impl TryFrom<&str> for Role {
    type Error = anyhow::Error;
    fn try_from(value: &str) -> Result<Self, Self::Error> {
        match value.to_lowercase().as_ref() {
            "clone" => Ok(Self::Clone),
            "eq" => Ok(Self::Eq),
            "drop" => Ok(Self::Drop),
            "len" => Ok(Self::Len),
            str => Err(anyhow!("unknown role `{str}`")),
        }
    }
}
