use crate::workers::Workers;
use anyhow::{bail, Context as _};
use std::{fs, path::Path, sync::atomic::Ordering, thread};
use vvs_ass::{ass_container_from_file, ASSContainer};
use vvs_llvm::prelude::*;
use vvs_runtime_types::types::VVRTTable;

use ratatui::{
    crossterm::event::{self, Event, KeyCode, KeyEventKind},
    widgets::{Block, Paragraph},
};

#[allow(dead_code)]
pub struct Runtime<'a> {
    jit: JIT,
    workers: Workers,
    output: Option<&'a mut fs::File>,
    terminal: Option<ratatui::DefaultTerminal>,
}

impl<'a> Runtime<'a> {
    pub fn new<'ctx>(init: impl FnOnce(Context<'ctx>) -> anyhow::Result<Module<'ctx>>) -> anyhow::Result<Self> {
        Ok(Self {
            jit: JIT::init(init)?,
            output: None,
            terminal: None,
            workers: Workers::new(|flag| loop {
                match flag.load(Ordering::SeqCst) {
                    0 => thread::yield_now(),
                    u64::MAX => return Ok(()),
                    x => bail!("won't handle work package n°{x}"),
                }
            })?,
        })
    }

    pub fn with_output_file(self, output: &'a mut fs::File) -> Self {
        Self { output: Some(output), ..self }
    }

    pub fn with_terminal(self, terminal: ratatui::DefaultTerminal) -> Self {
        Self { terminal: Some(terminal), ..self }
    }

    pub fn run_on_file(self, file: impl AsRef<Path>) -> anyhow::Result<()> {
        self.run_on_container(
            ass_container_from_file::<VVRTTable>(file.as_ref()).context("failed to parse the subtitle file")?,
        )
    }

    pub fn run_on_container(self, ass: ASSContainer<VVRTTable>) -> anyhow::Result<()> {
        let Self { output, terminal, .. } = self;
        log::error!("run on {ass:?}");

        match terminal {
            None => todo!("handle the case where we don't have a terminal"),
            Some(mut terminal) => loop {
                terminal.draw(|frame| {
                    frame.render_widget(
                        Paragraph::new("Hello World!").block(Block::bordered().title("Greeting")),
                        frame.area(),
                    );
                })?;
                if let Event::Key(key) = event::read()? {
                    if key.kind == KeyEventKind::Press && key.code == KeyCode::Char('q') {
                        break;
                    }
                }
            },
        }

        if let Some(output) = output {
            ass.write_ass_to_file(output)?;
        }

        Ok(())
    }
}
