//! Contains definitions for the [VVRTAny] runtime object and its associated methods.

/// Implement [TryFrom<VVRTAny>] and [TryFrom<&VVRTAny>] for a given object that can be contained
/// inside [VVRTAny].
macro_rules! impl_try_from_any {
    ($vvrtty: ident => $struct: ident) => {
        impl TryFrom<crate::types::VVRTAny> for $struct {
            #[inline]
            fn try_from(value: crate::types::VVRTAny) -> Result<Self, Self::Error> {
                unsafe { VVRTAny_is_ty(value, VVRTType::$vvrtty) }
                    .then_some(Self(value.ptr()))
                    .ok_or(value.ty())
            }
            type Error = crate::types::VVRTType;
        }

        impl TryFrom<&crate::types::VVRTAny> for $struct {
            #[inline]
            fn try_from(value: &crate::types::VVRTAny) -> Result<Self, Self::Error> {
                (*value).try_into()
            }
            type Error = crate::types::VVRTType;
        }
    };
}
pub(super) use impl_try_from_any;

use crate::{types::*, TaggedPtr};
use std::{
    mem::{align_of, transmute},
    ptr::NonNull,
};

/// We represent the *any* Vivy Script type by an enum. We try to fit this bad boi into 8 bytes
/// with a tagged pointer. With 3 bits for the tag we have values from 0 to 7, is enaugh to store
/// the [VVRTType]. We store the [VVRTType::Nil] as [None] so we can gain one slot.
///
/// Nil is all zero, which is Ok even if we have a non-null thingy (at least I hope).
#[derive(Debug, Clone, Copy)]
#[repr(transparent)]
pub struct VVRTAny(Option<TaggedPtr<u64, 3>>);

impl VVRTAny {
    /// Get the real type of the object inside the `any` container.
    #[inline]
    pub fn ty(&self) -> VVRTType {
        self.map_or_default(|this| unsafe { transmute(this.tag() as u8 + 1) })
    }

    /// Map on the inner tagged pointer, or return the specified default.
    #[inline]
    pub fn map_or<U, F: FnOnce(TaggedPtr<u64, 3>) -> U>(self, default: U, f: F) -> U {
        self.0.map_or(default, f)
    }

    /// Map on the inner tagger pointer, or return the default value for the returned type.
    #[inline]
    pub fn map_or_default<U: Default, F: FnOnce(TaggedPtr<u64, 3>) -> U>(self, f: F) -> U {
        self.0.map_or(U::default(), f)
    }

    /// Get the pointer out of the tagged pointer, or a dangling pointer in the case where we store
    /// a value of type [VVRTType::Nil].
    #[inline]
    pub(crate) fn ptr<T: Sized>(&self) -> NonNull<T> {
        debug_assert_eq!(align_of::<T>(), 8);
        self.map_or(NonNull::dangling(), |this| unsafe { NonNull::new_unchecked(this.ptr().as_ptr() as *mut T) })
    }

    /// Build the [VVRTAny] from a pointer assumed to be not null and the type of such pointer.
    #[inline]
    unsafe fn from_parts(ptr: *mut u64, ty: VVRTType) -> Self {
        Self((ty != VVRTType::Nil).then(|| TaggedPtr::new(NonNull::new_unchecked(ptr), ty as usize - 1)))
    }

    /// Tells whever the contained type is a floating number or not.
    #[inline]
    pub fn number_is_f32(&self) -> bool {
        (self.ty() == VVRTType::Number) && VVRTNumber::from(*self).is_f32()
    }

    /// Tells whever the contained type is an integer number or not.
    #[inline]
    pub fn number_is_i32(&self) -> bool {
        (self.ty() == VVRTType::Number) && VVRTNumber::from(*self).is_i32()
    }
}

impl Default for VVRTAny {
    /// By default, [VVRTAny] contains an object of type [VVRTType::Nil].
    #[inline]
    fn default() -> Self {
        Self(None)
    }
}

impl From<VVRTAny> for VVRTNumber {
    fn from(value: VVRTAny) -> Self {
        value.map_or_default(|ptr| VVRTNumber::from_number_be_bytes(ptr.addr().to_be_bytes()))
    }
}

impl From<VVRTNumber> for VVRTAny {
    fn from(value: VVRTNumber) -> Self {
        unsafe { Self::from_parts(value.as_ptr(), VVRTType::Number) }
    }
}

impl From<VVRTAny> for bool {
    #[inline]
    fn from(this: VVRTAny) -> Self {
        Into::<i32>::into(this) != 0
    }
}

impl From<VVRTAny> for i32 {
    #[inline]
    fn from(this: VVRTAny) -> Self {
        VVRTNumber::from(this).into()
    }
}

impl From<VVRTAny> for f32 {
    #[inline]
    fn from(this: VVRTAny) -> Self {
        VVRTNumber::from(this).into()
    }
}

impl From<VVRTVariant> for VVRTAny {
    #[inline]
    fn from(ptr: VVRTVariant) -> Self {
        unsafe { Self::from_parts(ptr.as_ptr(), VVRTType::Variant) }
    }
}

impl From<VVRTString> for VVRTAny {
    #[inline]
    fn from(ptr: VVRTString) -> Self {
        unsafe { Self::from_parts(ptr.as_ptr(), VVRTType::String) }
    }
}

impl From<VVRTTable> for VVRTAny {
    #[inline]
    fn from(ptr: VVRTTable) -> Self {
        unsafe { Self::from_parts(ptr.as_ptr(), VVRTType::Table) }
    }
}

impl From<VVRTLine> for VVRTAny {
    #[inline]
    fn from(ptr: VVRTLine) -> Self {
        unsafe { Self::from_parts(ptr.as_ptr(), VVRTType::ASSLine) }
    }
}

impl From<VVRTSyllabe> for VVRTAny {
    #[inline]
    fn from(ptr: VVRTSyllabe) -> Self {
        unsafe { Self::from_parts(ptr.as_ptr(), VVRTType::ASSSyllabe) }
    }
}

impl From<i32> for VVRTAny {
    #[inline]
    fn from(value: i32) -> Self {
        unsafe { Self::from_parts(VVRTNumber::from(value).as_ptr(), VVRTType::Number) }
    }
}

impl From<bool> for VVRTAny {
    #[inline]
    fn from(value: bool) -> Self {
        (value as i32).into()
    }
}

impl From<f32> for VVRTAny {
    #[inline]
    fn from(value: f32) -> Self {
        unsafe { Self::from_parts(VVRTNumber::from(value).as_ptr(), VVRTType::Number) }
    }
}

impl Eq for VVRTAny {}
impl PartialEq for VVRTAny {
    #[inline]
    fn eq(&self, other: &Self) -> bool {
        unsafe {
            macro_rules! into {
                ($what: expr) => {
                    self.try_into().unwrap_unchecked()
                };
            }
            match self.ty() {
                ty if ty != other.ty() => false,
                VVRTType::Nil => true,
                VVRTType::Number => self.0.unwrap_unchecked().addr() == other.0.unwrap_unchecked().addr(),
                VVRTType::ASSSyllabe => VVRTSyllabe_eq(into!(self), into!(other)),
                VVRTType::ASSLine => VVRTLine_eq(into!(self), into!(other)),
                VVRTType::Variant => VVRTVariant_eq(into!(self), into!(other)),
                VVRTType::String => VVRTString_eq(into!(self), into!(other)),
                VVRTType::Table => VVRTTable_eq(into!(self), into!(other)),
            }
        }
    }
}

/// Module used to define functions to be exported for the JIT.
#[allow(clippy::missing_safety_doc)]
#[vvs_procmacro::vvrt_bridge(VVRTAny)]
mod ffi {
    #[no_mangle]
    pub unsafe extern "C" fn nil() -> VVRTAny {
        Default::default()
    }

    #[no_mangle]
    pub unsafe extern "C" fn ty(any: VVRTAny) -> VVRTType {
        any.ty()
    }

    #[no_mangle]
    pub unsafe extern "C" fn is_ty(any: VVRTAny, ty: VVRTType) -> bool {
        any.ty() == ty
    }

    #[no_mangle]
    pub unsafe extern "C" fn is_nil(any: VVRTAny) -> bool {
        any.0.is_none()
    }

    #[no_mangle]
    pub unsafe extern "C" fn deep_clone(any: VVRTAny) -> VVRTAny {
        match any.ty() {
            // Shallow copy
            VVRTType::Nil | VVRTType::Number => any,
            VVRTType::String => VVRTString_clone(any.try_into().unwrap_unchecked()).into(),
            VVRTType::Variant => VVRTVariant_clone(any.try_into().unwrap_unchecked()).into(),

            // Deep copy
            VVRTType::Table => VVRTTable_deep_clone(any.try_into().unwrap_unchecked()).into(),
            VVRTType::ASSLine => VVRTLine_deep_clone(any.try_into().unwrap_unchecked()).into(),
            VVRTType::ASSSyllabe => VVRTSyllabe_deep_clone(any.try_into().unwrap_unchecked()).into(),
        }
    }

    #[no_mangle]
    pub unsafe extern "C" fn clone(any: VVRTAny) -> VVRTAny {
        match any.ty() {
            VVRTType::Nil | VVRTType::Number => any,
            VVRTType::Table => VVRTTable_clone(any.try_into().unwrap_unchecked()).into(),
            VVRTType::String => VVRTString_clone(any.try_into().unwrap_unchecked()).into(),
            VVRTType::Variant => VVRTVariant_clone(any.try_into().unwrap_unchecked()).into(),
            VVRTType::ASSLine => VVRTLine_clone(any.try_into().unwrap_unchecked()).into(),
            VVRTType::ASSSyllabe => VVRTSyllabe_clone(any.try_into().unwrap_unchecked()).into(),
        }
    }

    #[no_mangle]
    pub unsafe extern "C" fn drop(any: VVRTAny) {
        match any.ty() {
            VVRTType::Nil | VVRTType::Number => {}
            VVRTType::Table => VVRTTable_drop(any.try_into().unwrap_unchecked()),
            VVRTType::String => VVRTString_drop(any.try_into().unwrap_unchecked()),
            VVRTType::Variant => VVRTVariant_drop(any.try_into().unwrap_unchecked()),
            VVRTType::ASSLine => VVRTLine_drop(any.try_into().unwrap_unchecked()),
            VVRTType::ASSSyllabe => VVRTSyllabe_drop(any.try_into().unwrap_unchecked()),
        }
    }

    #[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn eq(this: VVRTAny, other: VVRTAny) -> bool { this == other }
    #[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn ne(this: VVRTAny, other: VVRTAny) -> bool { this != other }

    #[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn from_boolean (o: bool)        -> VVRTAny { o.into() }
    #[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn from_integer (o: i32)         -> VVRTAny { o.into() }
    #[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn from_floating(o: f32)         -> VVRTAny { o.into() }
    #[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn from_string  (o: VVRTString)  -> VVRTAny { o.into() }
    #[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn from_table   (o: VVRTTable)   -> VVRTAny { o.into() }
    #[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn from_variant (o: VVRTVariant) -> VVRTAny { o.into() }
    #[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn from_line    (o: VVRTLine)    -> VVRTAny { o.into() }
    #[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn from_syllabe (o: VVRTSyllabe) -> VVRTAny { o.into() }

    #[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn as_boolean (any: VVRTAny) -> bool        { any.into() }
    #[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn as_integer (any: VVRTAny) -> i32         { any.into() }
    #[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn as_floating(any: VVRTAny) -> f32         { any.into() }
    #[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn as_string  (any: VVRTAny) -> VVRTString  { any.try_into().unwrap_unchecked() }
    #[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn as_table   (any: VVRTAny) -> VVRTTable   { any.try_into().unwrap_unchecked() }
    #[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn as_variant (any: VVRTAny) -> VVRTVariant { any.try_into().unwrap_unchecked() }
    #[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn as_line    (any: VVRTAny) -> VVRTLine    { any.try_into().unwrap_unchecked() }
    #[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn as_syllabe (any: VVRTAny) -> VVRTSyllabe { any.try_into().unwrap_unchecked() }

    #[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn maybe_as_boolean (any: VVRTAny) -> bool        { VVRTAny_as_boolean(any) }
    #[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn maybe_as_integer (any: VVRTAny) -> i32         { VVRTAny_as_integer(any) }
    #[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn maybe_as_floating(any: VVRTAny) -> f32         { VVRTAny_as_floating(any) }
    #[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn maybe_as_string  (any: VVRTAny) -> VVRTString  { any.try_into().unwrap_or(VVRTString ::dangling()) }
    #[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn maybe_as_table   (any: VVRTAny) -> VVRTTable   { any.try_into().unwrap_or(VVRTTable  ::dangling()) }
    #[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn maybe_as_variant (any: VVRTAny) -> VVRTVariant { any.try_into().unwrap_or(VVRTVariant::dangling()) }
    #[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn maybe_as_line    (any: VVRTAny) -> VVRTLine    { any.try_into().unwrap_or(VVRTLine   ::dangling()) }
    #[rustfmt::skip] #[no_mangle] pub unsafe extern "C" fn maybe_as_syllabe (any: VVRTAny) -> VVRTSyllabe { any.try_into().unwrap_or(VVRTSyllabe::dangling()) }
}
