//! Here we have the base types for the runtime.

/// Get the list of all the builtin types.
macro_rules! builtin_types {
    ($macro: ident) => {
        $macro!(any, string, table, variant, line, syllabe);
    };
}

/// Declare the modules associated with the builtin types.
macro_rules! decl_modules {
    ($mod: ident $(,)?) => {
        mod $mod;
    };
    ($mod: ident, $($mods: ident),+ $(,)?) => {
        decl_modules!($mod);
        decl_modules!($($mods),+);
    };
}
builtin_types!(decl_modules);

mod number;

use self::{any::*, line::*, number::*, string::*, syllabe::*, table::*, variant::*};
use crate::vvll::*;
use std::sync::OnceLock;
use vvs_llvm::sys::*;

#[cfg(test)]
mod test;

pub use self::{
    any::VVRTAny, line::VVRTLine, string::VVRTString, syllabe::VVRTSyllabe, table::VVRTTable, variant::VVRTVariant,
};

/// Get the builtin types
pub fn builtin_types() -> &'static [&'static str] {
    #[allow(dead_code)]
    static NUMBERS_ARE_COHERENT: () = {
        assert!(
            std::mem::size_of::<f32>() == std::mem::size_of::<i32>(),
            "we need for the layout of the numbers to be the same to do our shenanigans"
        );
        assert!(
            std::mem::align_of::<f32>() == std::mem::align_of::<i32>(),
            "we need for the layout of the numbers to be the same to do our shenanigans"
        );
    };

    static TYPES: OnceLock<Vec<&'static str>> = OnceLock::new();
    TYPES
        .get_or_init(|| {
            let mut ret = vec![];
            macro_rules! decl_hashmap {
                ($mod: ident $(,)?) => {{
                    ret.push(stringify!($mod));
                }};
                ($mod: ident, $($mods: ident),+ $(,)?) => {{
                    decl_hashmap!($mod);
                    decl_hashmap!($($mods),+);
                }};
            }
            builtin_types!(decl_hashmap);
            ret
        })
        .as_slice()
}

/// Represent every types.
#[derive(Debug, Clone, Copy, PartialEq, Eq)]
#[repr(u8)]
pub enum VVRTType {
    /// Null, we ensure that [Option<VVRTAny>::None] is 0.
    Nil = 0,

    /// Numbers are floats or integers
    Number = 1,

    /// An immutable string
    String = 2,

    /// Table, a container indexed by string.
    Table = 3,

    /// A variant.
    Variant = 4,

    /// An ASS Line
    ASSLine = 5,

    /// An ASS Syllabe
    ASSSyllabe = 6,
}

impl Default for VVRTType {
    fn default() -> Self {
        use VVRTType::*;

        #[allow(dead_code)]
        const SMUGGLE_TYPE: () = {
            assert!(Nil as usize <= 8);
            assert!(String as usize <= 8);
            assert!(Number as usize <= 8);
            assert!(Table as usize <= 8);
            assert!(ASSLine as usize <= 8);
            assert!(ASSSyllabe as usize <= 8);
            assert!(Variant as usize <= 8);
        };

        #[rustfmt::skip] #[allow(dead_code)]
        const TYPE_ABI: () = {
            // Non-managed
            assert!((VVRTType::Number     as usize - 1) == 0b000);

            // Managed
            assert!((VVRTType::String     as usize - 1) == 0b001);
            assert!((VVRTType::Table      as usize - 1) == 0b010);
            assert!((VVRTType::Variant    as usize - 1) == 0b011);
            assert!((VVRTType::ASSLine    as usize - 1) == 0b100);
            assert!((VVRTType::ASSSyllabe as usize - 1) == 0b101);

            // We want to use 0b111 for user types in the future...
        };
        VVRTType::Nil
    }
}

impl VVRTType {
    /// Get the type as a constant to be used in LLVM codegen.
    ///
    /// # Safety
    /// The [LLVMContextRef] must be valid.
    #[inline]
    pub unsafe fn as_llvm_const(&self, c: LLVMContextRef) -> LLVMValueRef {
        LLVMConstInt(LLVMInt8TypeInContext(c), *self as u64, 0)
    }
}

impl LLVMExported for VVRTType {
    unsafe fn llvm_type(c: LLVMContextRef) -> LLVMTypeRef {
        #[allow(dead_code)]
        const TYPE_IS_VALID: () = {
            assert!(std::mem::size_of::<VVRTType>() == 1);
            assert!(std::mem::align_of::<VVRTType>() == 1);
        };
        LLVMInt8TypeInContext(c)
    }
}

impl AsRef<str> for VVRTType {
    #[inline]
    fn as_ref(&self) -> &str {
        match self {
            VVRTType::Nil => "nil",
            VVRTType::String => "string",
            VVRTType::Number => "number",
            VVRTType::Table => "table",
            VVRTType::Variant => "variant",
            VVRTType::ASSLine => "line",
            VVRTType::ASSSyllabe => "line",
        }
    }
}

impl std::fmt::Display for VVRTType {
    #[inline]
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_str(self.as_ref())
    }
}
