//! Contains utility to manipulate numbers from the runtime.

use std::{fmt, ops};
use vvs_utils::either;

/// Wrapper type for numbers that is intended to be used only for the [crate::types::VVRTAny]
/// and [crate::types::VVRTVariant] structs.
#[derive(Clone, Copy)]
#[repr(transparent)]
pub struct VVRTNumber(u64);

impl VVRTNumber {
    const FLT_TAG: u8 = 1;
    const INT_TAG: u8 = !Self::FLT_TAG;

    /// Cast the number as a pointer. We ensure by construction that this pointer is never null.
    /// You may never deref this pointer!
    #[inline]
    pub fn as_ptr(&self) -> *mut u64 {
        self.0 as *mut _
    }

    /// In debug mode, assert that the passed bytes are a valid number representation. In case of
    /// success re-return those bytes.
    ///
    /// In release build this function is a NoOp.
    #[inline]
    fn assert_is_number(bytes: [u8; 8]) -> [u8; 8] {
        let [.., flag, z1, z2, z3] = bytes;
        debug_assert!(z1 == z2 && z2 == z3 && z3 == 0, "invalid number");
        debug_assert!(flag == Self::INT_TAG || flag == Self::FLT_TAG, "invalid tag");
        bytes
    }

    /// Get the byte representation of the number we store.
    #[inline]
    fn as_bytes(&self) -> (u8, [u8; 4]) {
        let [b1, b2, b3, b4, flag, _, _, _] = Self::assert_is_number(self.0.to_be_bytes());
        (flag, [b1, b2, b3, b4])
    }

    /// Create [VVRTAny] from bytes assumed to be a valid number representation.
    #[inline]
    pub fn from_number_be_bytes(bytes: [u8; 8]) -> Self {
        Self(u64::from_be_bytes(Self::assert_is_number(bytes)))
    }

    /// Tells whever the contained type is a floating number or not.
    #[inline]
    pub fn is_f32(&self) -> bool {
        self.as_bytes().0 == Self::FLT_TAG
    }

    /// Tells whever the contained type is an integer number or not.
    #[inline]
    pub fn is_i32(&self) -> bool {
        !self.is_f32()
    }

    /// Convert the number into an integer.
    #[inline]
    pub fn as_i32(self) -> i32 {
        self.into()
    }

    /// Convert the number into a float.
    #[inline]
    pub fn as_f32(self) -> f32 {
        self.into()
    }

    /// Tells wether the number is not a number or is a number.
    #[inline]
    pub fn is_nan(&self) -> bool {
        self.as_f32().is_nan()
    }

    /// Tells wether the number is finite or not.
    #[inline]
    pub fn is_finite(&self) -> bool {
        self.is_i32() || self.as_f32().is_finite()
    }

    /// Tells wether the number is `true`
    #[inline]
    pub fn is_true(&self) -> bool {
        self.as_i32() != 0
    }

    /// Tells wether the number is `false`
    #[inline]
    pub fn is_false(&self) -> bool {
        self.as_i32() == 0
    }

    /// Tells wether the number is 0, or 0.0 with precision of [f32::EPSILON], or not.
    #[inline]
    pub fn is_zero(&self) -> bool {
        (self.is_i32() && self.as_i32() == 0) || (self.as_f32().abs() <= f32::EPSILON)
    }
}

impl Default for VVRTNumber {
    #[inline]
    fn default() -> Self {
        0_i32.into()
    }
}

impl From<VVRTNumber> for i32 {
    #[inline]
    fn from(value: VVRTNumber) -> Self {
        let (flag, bytes) = value.as_bytes();
        either!(flag == VVRTNumber::INT_TAG
            => i32::from_be_bytes(bytes)
            ;  f32::from_be_bytes(bytes).trunc() as i32
        )
    }
}

impl From<VVRTNumber> for f32 {
    #[inline]
    fn from(value: VVRTNumber) -> Self {
        let (flag, bytes) = value.as_bytes();
        either!(flag == VVRTNumber::INT_TAG
            => i32::from_be_bytes(bytes) as f32
            ;  f32::from_be_bytes(bytes)
        )
    }
}

impl From<i32> for VVRTNumber {
    #[inline]
    fn from(value: i32) -> Self {
        let [b1, b2, b3, b4] = value.to_be_bytes();
        VVRTNumber::from_number_be_bytes([b1, b2, b3, b4, VVRTNumber::INT_TAG, 0, 0, 0])
    }
}

impl From<f32> for VVRTNumber {
    #[inline]
    fn from(value: f32) -> Self {
        let [b1, b2, b3, b4] = value.to_be_bytes();
        VVRTNumber::from_number_be_bytes([b1, b2, b3, b4, VVRTNumber::FLT_TAG, 0, 0, 0])
    }
}

impl fmt::Debug for VVRTNumber {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let (flag, bytes) = self.as_bytes();
        f.debug_struct("VVRTNumber")
            .field("flag", &flag)
            .field("bytes", &bytes)
            .finish()
    }
}

impl fmt::Display for VVRTNumber {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self.is_f32() {
            true => write!(f, "{}", Into::<f32>::into(*self)),
            false => write!(f, "{}", Into::<i32>::into(*self)),
        }
    }
}

macro_rules! impl_arithmetic_ops {
    () => {};

    ( $op:ident  on $ty:ident  : $func:ident  ($lh:ident,  $rh:ident ) -> $expr:expr;
    $($ops:ident on $tys:ident : $funcs:ident ($lhs:ident, $rhs:ident) -> $exprs:expr;)*) => {
        impl_arithmetic_ops! { @@ $op, $ty, $func($lh, $rh) -> $expr }
        impl_arithmetic_ops! { $($ops on $tys : $funcs($lhs, $rhs) -> $exprs;)* }
    };

    (@@ $ops:ident, both, $func:ident ($lhs:ident, $rhs:ident) -> $expr:expr) => {
        impl_arithmetic_ops! { @@ $ops, $func(self, rhs) -> {
            if self.is_f32() || rhs.is_f32() {
                let ($lhs, $rhs) = (self.as_f32(), rhs.as_f32());
                VVRTNumber::from($expr)
            } else {
                let ($lhs, $rhs) = (self.as_i32(), rhs.as_i32());
                VVRTNumber::from($expr)
            }
        } }
    };

    (@@ $ops:ident, $ty:ident, $func:ident ($lhs:ident, $rhs:ident) -> $expr:expr) => {
        impl_arithmetic_ops! { @@ $ops, $func(self, rhs) -> {
            let ($lhs, $rhs) = paste::paste! { (self.[< as_$ty >](), rhs.[< as_$ty >]()) };
            VVRTNumber::from($expr)
        } }
    };

    (@@ $ops: ident, $func:ident ($lhs:ident, $rhs:ident) -> $expr:expr) => {
        impl ops::$ops for VVRTNumber {
            type Output = VVRTNumber;
            #[inline] fn $func($lhs, $rhs: Self) -> Self::Output { $expr }
        }
    }
}

impl_arithmetic_ops! {
    Add    on both : add   (lhs, rhs) -> lhs + rhs;
    Sub    on both : sub   (lhs, rhs) -> lhs - rhs;
    Mul    on both : mul   (lhs, rhs) -> lhs * rhs;
    BitOr  on i32  : bitor (lhs, rhs) -> lhs | rhs;
    BitAnd on i32  : bitand(lhs, rhs) -> lhs & rhs;
    BitXor on i32  : bitxor(lhs, rhs) -> lhs ^ rhs;

    Shl on i32 : shl(lhs, rhs) -> {
        let (res, overflow) = match rhs {
            rhs if rhs < 0 => lhs.overflowing_shr((-rhs) as u32),
            rhs => lhs.overflowing_shl(rhs as u32)
        };
        match overflow {
            true => VVRTNumber::from(f32::NAN),
            false => VVRTNumber::from(res),
        }
    };

    Shr on i32 : shr(lhs, rhs) -> {
        let (res, overflow) = match rhs {
            rhs if rhs < 0 => lhs.overflowing_shl((-rhs) as u32),
            rhs => lhs.overflowing_shr(rhs as u32)
        };
        match overflow {
            true => VVRTNumber::from(f32::NAN),
            false => VVRTNumber::from(res),
        }
    };

    Div on both : div(lhs, rhs) -> match rhs as f32 == 0.0 {
        true => VVRTNumber::from(f32::NAN),
        false => VVRTNumber::from(lhs / rhs),
    };

    Rem on i32 : rem(lhs, rhs) -> match rhs {
        0 => VVRTNumber::from(f32::NAN),
        rhs => VVRTNumber::from(match lhs % rhs {
            res if res < 0 => res + rhs,
            res => res,
        })
    };
}

impl ops::Neg for VVRTNumber {
    type Output = VVRTNumber;

    #[inline]
    fn neg(self) -> Self::Output {
        match self.is_i32() {
            true => (-self.as_i32()).into(),
            false => (-self.as_f32()).into(),
        }
    }
}

impl ops::Not for VVRTNumber {
    type Output = VVRTNumber;

    #[inline]
    fn not(self) -> Self::Output {
        (!self.as_i32()).into()
    }
}

#[cfg(any(test, doctest))]
mod test {
    use super::*;

    #[test]
    fn pack_unpack() {
        let one = VVRTNumber::from(1);
        assert!(one.is_i32());
        assert!(!one.is_f32());
        assert_eq!(Into::<i32>::into(one), 1);

        let pi = VVRTNumber::from(std::f32::consts::PI);
        assert!(!pi.is_i32());
        assert!(pi.is_f32());
        assert_eq!(Into::<f32>::into(pi), std::f32::consts::PI);
    }

    #[test]
    fn arithmetic() {
        assert_eq!((VVRTNumber::from(1) * VVRTNumber::from(std::f32::consts::PI)).as_f32(), std::f32::consts::PI);
        assert_eq!((VVRTNumber::from(3) % VVRTNumber::from(2)).as_i32(), 1);
        assert_eq!((VVRTNumber::from(-3) % VVRTNumber::from(2)).as_i32(), 1);
        assert!((VVRTNumber::from(1.0) / VVRTNumber::from(10000000.0)).is_zero());
    }

    #[test]
    fn divide_by_zero() {
        assert!((VVRTNumber::from(1.0) / VVRTNumber::from(0.0)).is_nan());
        assert!((VVRTNumber::from(1) / VVRTNumber::from(0)).is_nan());
        assert!((VVRTNumber::from(3) % VVRTNumber::from(0)).is_nan());
    }
}
