use crate::types::*;
use std::{hash::Hash, ptr::NonNull, rc::Rc};

/// We represent a string, in an FFI-Safe way because slices are not FFI-Safe...
#[derive(Debug, Clone, Copy)]
#[repr(transparent)]
pub struct VVRTString(NonNull<Vec<u8>>);
crate::types::any::impl_try_from_any! { String => VVRTString }

impl VVRTString {
    #[inline]
    pub fn dangling() -> Self {
        Self(NonNull::dangling())
    }

    #[inline]
    pub fn as_ptr(self) -> *mut u64 {
        self.0.as_ptr() as *mut _
    }

    #[inline]
    pub fn as_str(&self) -> &str {
        unsafe { std::str::from_utf8_unchecked(self.as_bytes()) }
    }

    #[inline]
    pub fn as_bytes(&self) -> &[u8] {
        unsafe { self.0.as_ref().as_slice() }
    }

    #[inline]
    pub fn from_bytes(bytes: &[u8]) -> Self {
        Self::from_vec(bytes.to_vec())
    }

    #[inline]
    pub fn from_vec(bytes: Vec<u8>) -> Self {
        unsafe { VVRTString(NonNull::new_unchecked(Rc::into_raw(Rc::new(bytes)) as *mut _)) }
    }
}

impl Default for VVRTString {
    fn default() -> Self {
        Self::from_vec(Default::default())
    }
}

impl AsRef<str> for VVRTString {
    #[inline]
    fn as_ref(&self) -> &str {
        self.as_str()
    }
}

impl Eq for VVRTString {}
impl PartialEq for VVRTString {
    #[inline]
    fn eq(&self, other: &Self) -> bool {
        self.as_str() == other.as_str()
    }
}

impl Hash for VVRTString {
    #[inline]
    fn hash<H: std::hash::Hasher>(&self, state: &mut H) {
        self.as_str().hash(state);
    }
}

impl PartialOrd for VVRTString {
    #[inline]
    fn partial_cmp(&self, other: &Self) -> Option<std::cmp::Ordering> {
        PartialOrd::partial_cmp(self.as_str(), other.as_str())
    }
}

/// Module used to defined functions to be exported to the JIT.
#[allow(clippy::missing_safety_doc)]
#[vvs_procmacro::vvrt_bridge(VVRTString)]
mod ffi {
    #[no_mangle]
    unsafe extern "C" fn new_from_raw(ptr: *const u8, len: i32) -> VVRTString {
        let len: usize = len.try_into().expect("string too big");
        VVRTString::from_bytes(std::slice::from_raw_parts(ptr, len))
    }

    #[no_mangle]
    unsafe extern "C" fn cat(left: VVRTString, right: VVRTString) -> VVRTString {
        let mut res = left.as_bytes().to_vec();
        res.extend_from_slice(right.as_bytes());
        VVRTString::from_vec(res)
    }

    #[no_mangle]
    pub unsafe extern "C" fn deep_clone(this: VVRTString) -> VVRTString {
        VVRTString_clone(this)
    }

    #[no_mangle]
    unsafe extern "C" fn clone(VVRTString(ptr): VVRTString) -> VVRTString {
        Rc::increment_strong_count(ptr.as_ptr());
        VVRTString(ptr)
    }

    #[no_mangle]
    unsafe extern "C" fn drop(VVRTString(ptr): VVRTString) {
        Rc::decrement_strong_count(ptr.as_ptr())
    }

    #[no_mangle]
    unsafe extern "C" fn len(str: VVRTString) -> i32 {
        str.as_str().len().try_into().expect("string too long")
    }

    use std::cmp::Ordering::*;
    #[rustfmt::skip] #[no_mangle] unsafe extern "C" fn eq(a: VVRTString, b: VVRTString) -> bool { PartialEq::eq(&a, &b) }
    #[rustfmt::skip] #[no_mangle] unsafe extern "C" fn ne(a: VVRTString, b: VVRTString) -> bool { PartialEq::ne(&a, &b) }
    #[rustfmt::skip] #[no_mangle] unsafe extern "C" fn lt(a: VVRTString, b: VVRTString) -> bool { matches!(a.partial_cmp(&b), Some(Less           )) }
    #[rustfmt::skip] #[no_mangle] unsafe extern "C" fn gt(a: VVRTString, b: VVRTString) -> bool { matches!(a.partial_cmp(&b), Some(Greater        )) }
    #[rustfmt::skip] #[no_mangle] unsafe extern "C" fn le(a: VVRTString, b: VVRTString) -> bool { matches!(a.partial_cmp(&b), Some(Less    | Equal)) }
    #[rustfmt::skip] #[no_mangle] unsafe extern "C" fn ge(a: VVRTString, b: VVRTString) -> bool { matches!(a.partial_cmp(&b), Some(Greater | Equal)) }
}
