#![allow(non_snake_case)]

use hashbrown::HashSet;
use std::{
    borrow::Borrow,
    sync::{OnceLock, RwLock},
};
use vvs_llvm::sys::*;

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub struct ManageFunctions {
    pub clone: &'static str,
    pub drop: &'static str,
}

/// Trait used to give to the codegen the layout and functions associated with a type that is
/// present in the runtime and that must be handled in the codegen.
pub trait LLVMExported {
    /// Get the layout of the type.
    ///
    /// # Safety
    /// The [LLVMContextRef] must be valid.
    unsafe fn llvm_type(_: LLVMContextRef) -> LLVMTypeRef;
}

/// Tells if a type is managed or not.
///
/// # Safety
/// - All the managed types must have been exported into at least one module
/// - This function is not thread-safe
pub unsafe fn LLVMTypeIsManaged(ty: LLVMTypeRef) -> bool {
    /// Don't use this, just to allow for [LLVMTypeIsManaged] to exists…
    #[derive(Eq, PartialEq, Hash)]
    struct LLVMTypeRefWrapper(LLVMTypeRef);

    unsafe impl Send for LLVMTypeRefWrapper {}
    unsafe impl Sync for LLVMTypeRefWrapper {}
    impl Borrow<LLVMTypeRef> for LLVMTypeRefWrapper {
        fn borrow(&self) -> &LLVMTypeRef {
            &self.0
        }
    }

    static MANAGED_TYPES: OnceLock<RwLock<HashSet<LLVMTypeRefWrapper>>> = OnceLock::new();

    MANAGED_TYPES
        .get_or_init(RwLock::default)
        .read()
        .map_or(false, |types| types.contains(&ty))
}
